

#define _USE_MATH_DEFINES
#include <math.h>
#define RAND1 ((double)(rand()&0x7FFF)/32768.0)

//#include "tnt.h"
//#include "tnt_linalg.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "SegRegions.h"
#include "DensityMap.h"


DensityMap::DensityMap ()
{
	nstart[0] = 0; nstart[1] = 0; nstart[2] = 0;
	n[0] = 0; n[1] = 0; n[2] = 0;
}


void DensityMap::ReadText (const char *fname)
{
	name = fname;
	string map_file = name + string(".txt");

	printf ( "Getting map from: %s\n", map_file.c_str() );
	FILE *fp = fopen ( map_file.c_str(), "r" );

	fscanf (fp, "%d %d %d", &n[0], &n[1], &n[2]);
	printf ("Map size: %d %d %d\n", n[0], n[1], n[2]);

	float f[3];
	fscanf (fp, "%f %f %f", &f[0], &f[1], &f[2]);
	O = Vector3d ( f[0], f[1], f[2] );
	printf (" - origin: %f %f %f\n", O[0], O[1], O[2]);

	fscanf (fp, "%f %f %f", &f[0], &f[1], &f[2]);
	step = Vector3d ( f[0], f[1], f[2] );
	printf (" - step: %f %f %f\n", step[0], step[1], step[2]);

	// M = malloc ( n[0] * n[1] * n[2] * sizeof(float) );
	dMaxVal = -1e99; dMinVal = 1e99;

	vSize = Vector3d ( ((double) n[0]) * step[0], ((double) n[1]) * step[1], ((double) n[2]) * step[2] );

	Allocate ( n[0], n[1], n[2] );

	for (int i=0; i < n[0]; i++) {
		for (int j=0; j < n[1]; j++) {
			for (int k=0; k < n[2]; k++) {
				float val;
				
				if ( fscanf ( fp, "%f", &val ) != 1 ) {
					printf (" -- error reading value %d %d %d\n", i, j, k);
					return;
				}

				M[i][j][k] = val;
				if ( val > dMaxVal ) dMaxVal = val;
				if ( val < dMinVal ) dMinVal = val;

				if ( val < 0.5 ) continue;

				Vector3d p = O + Vector3d ( ((double)i)*step[0], ((double)j)*step[1], ((double)k)*step[2] );
			}
		}
	}
	fclose (fp);
	printf ( " - max %f, min %f\n", dMaxVal, dMinVal );
}


void DensityMap::ReadCCP4 (const char *fname)
{
	name = fname;

	int i = (int) name.rfind ( '.' );
	name = name.substr ( 0, i );

	printf ( "Getting map from: %s\n", fname );
	FILE *fp = fopen ( fname, "rb" );

	if ( fp == NULL ) {
		printf (" - could not open\n");
		exit (0);
	}

	// format from http://www.ccp4.ac.uk/html/maplib.html#description

	fread ( &n, sizeof(int), 3, fp );
	printf (" - map size: %d %d %d\n", n[0], n[1], n[2]);

	int type;
	fread ( &type, sizeof(int), 1, fp );
	char types[10][100] = { "envelope stored as signed bytes (from -128 lowest to 127 highest)",
						"Image     stored as Integer*2",
						"Image     stored as Reals",
						"Transform stored as Complex Integer*2",
						"Transform stored as Complex Reals",
						"0" };
	printf ( " - type %d: %s\n", type, types[type] );

	fread ( &nstart, sizeof(int), 3, fp );
	printf (" - map start: %d %d %d\n", nstart[0], nstart[1], nstart[2]);
	// printf (" - map start: %f %f %f\n", (float)nstart[0], (float)nstart[1], (float)nstart[2]);

	int num_intervals[3];
	fread ( &num_intervals, sizeof(int), 3, fp );
	printf (" - # intervals: %d %d %d\n", num_intervals[0], num_intervals[1], num_intervals[2]);

	float cell_dim[3];
	fread ( &cell_dim, sizeof(float), 3, fp );
	printf (" - cell dimensions (angstroms): %f %f %f\n", cell_dim[0], cell_dim[1], cell_dim[2]);

	float cell_angles[3];
	fread ( &cell_angles, sizeof(float), 3, fp );
	printf (" - cell angles: %f %f %f\n", cell_angles[0], cell_angles[1], cell_angles[2]);

	int axes[3];
	fread ( &axes, sizeof(int), 3, fp );
	printf (" - axes: %d %d %d\n", axes[0], axes[1], axes[2]);

	float dvals[3];
	fread ( &dvals, sizeof(float), 3, fp );
	printf (" - min density %f, max density %f, mean density %f\n", dvals[0], dvals[1], dvals[2]);

	int Space_group_number;
	fread ( &Space_group_number, sizeof(int), 1, fp );
	printf (" - space group number: %d\n", Space_group_number);

	int nb_sym_ops;
	fread ( &nb_sym_ops, sizeof(int), 1, fp );
	printf (" - Number of bytes used for storing symmetry operators: %d\n", nb_sym_ops);
	if ( nb_sym_ops != 0 ) exit ( 0 );

	int skew_flag;
	fread ( &skew_flag, sizeof(int), 1, fp );
	printf (" - Flag for skew transformation, =0 none, =1 if foll: %d\n", skew_flag);

	printf (" - Skew matrix S (S11, S12, S13, S21 etc):");
	for (int i=0; i < 9; i++) {
		float me;
		fread ( &me, sizeof(float), 1, fp );
		printf (" %.2f%s", me, i==8?"\n":"");
	}

	float skew_translation[3];
	fread ( &skew_translation, sizeof(float), 3, fp );
	printf (" - skew translation: %f %f %f\n", skew_translation[0], skew_translation[1], skew_translation[2]);
		// Skew transformation is from standard orthogonal
		// coordinate frame (as used for atoms) to orthogonal
		// map frame, as Xo(map) = S * (Xo(atoms) - t)

	for (int i=38; i < 53; i++) {
		// for future/ad-hoc use
		float me; fread ( &me, sizeof(float), 1, fp );
	}

	char map_c[5]; map_c[4] = '\0';
	fread ( map_c, sizeof(char), 4, fp );
	printf (" - string: [%s] %d %d %d %d\n", map_c, map_c[0], map_c[1], map_c[2], map_c[3]);

	int machine_stamp;
	fread ( &machine_stamp, sizeof(int), 1, fp );
	printf (" - machine stamp: %d\n", machine_stamp);

	float rms_from_mean;
	fread ( &rms_from_mean, sizeof(float), 1, fp );
	printf (" - Rms deviation of map from mean density: %f\n", rms_from_mean);

	int num_labels;
	fread ( &num_labels, sizeof(int), 1, fp );
	printf (" - Number of labels being used: %d\n", num_labels);

	for (int i=0; i < 10; i++) {
		char label[81]; fread ( label, sizeof(char), 80, fp ); label[80] = '\0';
		printf (" - label %d: %s\n", i, label);
	}

	printf ("\n");

	//if ( num_intervals[0] != n[0]-1 || num_intervals[1] != n[1]-1 || num_intervals[2] != n[2]-1 ) {
	//	printf ("Map dimensions and # intervals are inconsistent\n");
		//exit (0);
	//}

	step = Vector3d ( ( (double) cell_dim[0] )/( (double) num_intervals[0] ),
					  ( (double) cell_dim[1] )/( (double) num_intervals[1] ),
					  ( (double) cell_dim[2] )/( (double) num_intervals[2] )  );
	printf (" - step: %f %f %f\n", step[0], step[1], step[2]);

	vSize = Vector3d ( ((double) n[0]) * step[0], ((double) n[1]) * step[1], ((double) n[2]) * step[2] );
	printf (" - map size: %f %f %f\n", vSize[0], vSize[1], vSize[2]);

	O = Vector3d ( ( (double) nstart[0] ) * step[0] + step[0]/2.0,
				   ( (double) nstart[1] ) * step[1] + step[1]/2.0,
				   ( (double) nstart[2] ) * step[2] + step[2]/2.0   );

	O = Vector3d ( ( (double) nstart[0] ) * step[0],
				   ( (double) nstart[1] ) * step[1],
				   ( (double) nstart[2] ) * step[2]   );

	printf (" - origin: %f %f %f\n", O[0], O[1], O[2]);

	dCellVolume = step[0]*step[1]*step[2];
	printf (" - cell volume: %.4f\n", dCellVolume);

	Allocate ( n[0], n[1], n[2] );
	dMaxVal = -1e99; dMinVal = 1e99;
	dAvgDensity = 0.0; double nn = 0.0;

	for (int k=0; k < n[2]; k++)
	{
		for (int j=0; j < n[1]; j++)
		{
			for (int i=0; i < n[0]; i++)
			{
				float val;
				if ( fread ( &val, sizeof(float), 1, fp ) == 0 ) {
					printf (" -- error reading value %d %d %d\n", i, j, k);
					return;
				}
				M[k][i][j] = val;
				
				dAvgDensity += val;
				nn += 1.0;
				
				if ( val > dMaxVal ) dMaxVal = val;
				if ( val < dMinVal ) dMinVal = val;
			}
		}
	}

	fclose (fp);
	dAvgDensity /= nn;
	printf ( "Max %f, min %f, avg %f\n", dMaxVal, dMinVal, dAvgDensity );
}




void DensityMap::ReadMRC (const char *fname, bool bPrint)
{
	name = fname;

	int i = (int) name.rfind ( '.' );
	name = name.substr ( 0, i );

	printf ( "Getting map from: %s\n", fname );
	FILE *fp = fopen ( fname, "rb" );

	if ( fp == NULL ) {
		printf (" - could not open\n");
		exit (0);
	}

	// mrc 2000 : http://www2.mrc-lmb.cam.ac.uk/image2000.html

	int old_n[3]; old_n[0] = n[0]; old_n[1] = n[1]; old_n[2] = n[2];
	fread ( &n, sizeof(int), 3, fp );
	printf (" - map size: %d %d %d\n", n[0], n[1], n[2]);

	int type;
	fread ( &type, sizeof(int), 1, fp );
	char types[10][100] = { "envelope stored as signed bytes (from -128 lowest to 127 highest)",
						"Image     stored as Integer*2",
						"Image     stored as Reals",
						"Transform stored as Complex Integer*2",
						"Transform stored as Complex Reals",
						"0" };
	if (bPrint) printf ( " - type %d: %s\n", type, types[type] );

	fread ( &nstart, sizeof(int), 3, fp );
	if (bPrint) printf (" - map start: %d %d %d\n", nstart[0], nstart[1], nstart[2]);
	// printf (" - map start: %f %f %f\n", (float)nstart[0], (float)nstart[1], (float)nstart[2]);

	// 8,9,10
	int num_intervals[3];
	fread ( &num_intervals, sizeof(int), 3, fp );
	if (bPrint) printf (" - # intervals: %d %d %d\n", num_intervals[0], num_intervals[1], num_intervals[2]);

	// 11-13 cell dimensions in angstroms
	float cell_dim[3];
	fread ( &cell_dim, sizeof(float), 3, fp );
	if (bPrint) printf (" - cell dimensions (angstroms): %f %f %f\n", cell_dim[0], cell_dim[1], cell_dim[2]);

	// 14-16 cell angles in degress
	float cell_angles[3];
	fread ( &cell_angles, sizeof(float), 3, fp );
	if (bPrint) printf (" - cell angles: %f %f %f\n", cell_angles[0], cell_angles[1], cell_angles[2]);

	// 17, 18, 19 - axis for cols, axis for rows, axis for sections
	int axes[3];
	fread ( &axes, sizeof(int), 3, fp );
	if (bPrint) printf (" - axes: %d %d %d\n", axes[0], axes[1], axes[2]);

	// 20, 21, 22 - min val, max val, mean val
	float dvals[3];
	fread ( &dvals, sizeof(float), 3, fp );
	if (bPrint) printf (" - min density %f, max density %f, mean density %f\n", dvals[0], dvals[1], dvals[2]);

	// 23 ispg - space group number (0 or 1)
	int Space_group_number;
	fread ( &Space_group_number, sizeof(int), 1, fp );
	if (bPrint) printf (" - space group number: %d\n", Space_group_number);

	// 24 nsymbt - number of bytes used for symmetry data (0 or 80)
	int nb_sym_ops;
	fread ( &nb_sym_ops, sizeof(int), 1, fp );
	if (bPrint) printf (" - Number of bytes used for storing symmetry operators: %d\n", nb_sym_ops);
	if ( nb_sym_ops != 0 ) exit ( 0 );

	// 25-49 extra - extra space for anything
	for (int i=25; i <= 49; i++) {
		// for future/ad-hoc use
		int ii; fread ( &ii, sizeof(int), 1, fp );
		// printf ("%d: %d\n", i, ii);
	}

	// 50-52 origin - x,y,z
	float origin[3];
	fread ( &origin, sizeof(float), 3, fp );
	O = Vector3d ( origin[0], origin[1], origin[2] );
	printf (" - origin: %f %f %f\n", O[0], O[1], O[2]);

	// 53 map - character string 'MAP'
	char map_c[5]; map_c[4] = '\0';
	fread ( map_c, sizeof(char), 4, fp );
	if (bPrint) printf (" - string: [%s] %d %d %d %d\n", map_c, map_c[0], map_c[1], map_c[2], map_c[3]);

	int machine_stamp;
	fread ( &machine_stamp, sizeof(int), 1, fp );
	if (bPrint) printf (" - machine stamp: %d\n", machine_stamp);

	float rms_from_mean;
	fread ( &rms_from_mean, sizeof(float), 1, fp );
	if (bPrint) printf (" - Rms deviation of map from mean density: %f\n", rms_from_mean);

	int num_labels;
	fread ( &num_labels, sizeof(int), 1, fp );
	if (bPrint) printf (" - Number of labels being used: %d\n", num_labels);

	for (int i=0; i < 10; i++) {
		char label[81]; fread ( label, sizeof(char), 80, fp ); label[80] = '\0';
		if (bPrint) printf (" - label %d: %s\n", i, label);
	}

	if (bPrint) printf ("\n");

	//if ( num_intervals[0] != n[0]-1 || num_intervals[1] != n[1]-1 || num_intervals[2] != n[2]-1 ) {
	//	printf ("Map dimensions and # intervals are inconsistent\n");
		//exit (0);
	//}

	step = Vector3d ( ( (double) cell_dim[0] )/( (double) num_intervals[0] ),
					  ( (double) cell_dim[1] )/( (double) num_intervals[1] ),
					  ( (double) cell_dim[2] )/( (double) num_intervals[2] )  );
	printf (" - step: %f %f %f\n", step[0], step[1], step[2]);

	vSize = Vector3d ( ((double) n[0]) * step[0], ((double) n[1]) * step[1], ((double) n[2]) * step[2] );
	if (bPrint) printf (" - map size: %f %f %f\n", vSize[0], vSize[1], vSize[2]);

	dCellVolume = step[0]*step[1]*step[2];
	if (bPrint) printf (" - cell volume: %.4f\n", dCellVolume);

	// if ( old_n[0] != n[0] || old_n[1] != n[1] || old_n[2] != n[2] )
	Allocate ( n[0], n[1], n[2] );

	dMaxVal = -1e99; dMinVal = 1e99;
	dAvgDensity = 0.0; double nn = 0.0;

	for (int k=0; k < n[2]; k++)
	{
		for (int j=0; j < n[1]; j++)
		{
			for (int i=0; i < n[0]; i++)
			{
				float val;
				if ( type == 2 ) {
					if ( fread ( &val, sizeof(float), 1, fp ) == 0 ) {
						printf (" -- error reading value %d %d %d\n", i, j, k);
						return;
					}
				}
				else if ( type == 0 ) {
					signed char c;
					if ( fread ( &c, sizeof(signed char), 1, fp ) == 0 ) {
						printf (" -- error reading value %d %d %d\n", i, j, k);
						return;
					}
					val = - (float) c;
				}
				else if ( type == 1 ) {
					short int i;
					if ( fread ( &i, sizeof(short int), 1, fp ) == 0 ) {
						printf (" -- error reading value %d %d %d\n", i, j, k);
						return;
					}
					val = (float) i;
				}
				else {
					printf ("don't know how to handle type %d yet\n", type);
					return;
				}

				M[k][j][i] = (double) val;
				
				dAvgDensity += val;
				nn += 1.0;
				
				if ( val > dMaxVal ) dMaxVal = val;
				if ( val < dMinVal ) dMinVal = val;
			}
		}
	}

	fclose (fp);
	dAvgDensity /= nn;
	
	printf ( "- max %f, min %f, avg %f\n", dMaxVal, dMinVal, dAvgDensity );
}





void DensityMap::WriteCCP4 (const char *fname)
{
	name = fname;
	string map_file = name + string(".map");

	printf ( "Writing map to: %s\n", map_file.c_str() );
	FILE *fp = fopen ( map_file.c_str(), "wb" );

	if ( fp == NULL ) {
		printf (" - could not open\n");
		exit (0);
	}

	// format from http://www.ccp4.ac.uk/html/maplib.html#description

	fwrite ( &n, sizeof(int), 3, fp );
	printf (" - map size: %d %d %d\n", n[0], n[1], n[2]);

	int type = 2;
	fwrite ( &type, sizeof(int), 1, fp );
	printf ( " - type %d\n", type );

	fwrite ( &nstart, sizeof(int), 3, fp );
	printf (" - map start: %d %d %d\n", nstart[0], nstart[1], nstart[2]);

	fwrite ( &n, sizeof(int), 3, fp );
	printf (" - # intervals: %d %d %d\n", n[0], n[1], n[2]);

	vSize = Vector3d ( ((double) n[0]) * step[0], ((double) n[1]) * step[1], ((double) n[2]) * step[2] );
	float cell_dim[3];
	cell_dim[0] = (float) vSize[0];
	cell_dim[1] = (float) vSize[1];
	cell_dim[2] = (float) vSize[2];
	fwrite ( &cell_dim, sizeof(float), 3, fp );
	printf (" - cell dimensions (angstroms): %f %f %f\n", cell_dim[0], cell_dim[1], cell_dim[2]);

	float cell_angles[3];
	cell_angles[0] = 90.0;
	cell_angles[1] = 90.0;
	cell_angles[2] = 90.0;
	fwrite ( &cell_angles, sizeof(float), 3, fp );
	printf (" - cell angles: %f %f %f\n", cell_angles[0], cell_angles[1], cell_angles[2]);

	int axes[3];
	axes[0] = 1;
	axes[1] = 2;
	axes[2] = 3;
	fwrite ( &axes, sizeof(int), 3, fp );
	printf (" - axes: %d %d %d\n", axes[0], axes[1], axes[2]);

	float dvals[3];
	dvals[0] = (float) dMinVal;
	dvals[1] = (float) dMaxVal;
	dvals[2] = (float) dAvgDensity;
	fwrite ( &dvals, sizeof(float), 3, fp );
	printf (" - min density %f, max density %f, mean density %f\n", dvals[0], dvals[1], dvals[2]);

	int Space_group_number = 0;
	fwrite ( &Space_group_number, sizeof(int), 1, fp );
	printf (" - space group number: %d\n", Space_group_number);

	int nb_sym_ops = 0;
	fwrite ( &nb_sym_ops, sizeof(int), 1, fp );
	printf (" - Number of bytes used for storing symmetry operators: %d\n", nb_sym_ops);

	int skew_flag = 0;
	fwrite ( &skew_flag, sizeof(int), 1, fp );
	printf (" - Flag for skew transformation, =0 none, =1 if foll: %d\n", skew_flag);

	printf (" - Skew matrix S (S11, S12, S13, S21 etc):");
	for (int i=0; i < 9; i++) {
		float me = 0.0;
		fwrite ( &me, sizeof(float), 1, fp );
		printf (" %.2f%s", me, i==8?"\n":"");
	}

	float skew_translation[3] = {0.0, 0.0, 0.0};
	fwrite ( &skew_translation, sizeof(float), 3, fp );
	printf (" - skew translation: %f %f %f\n", skew_translation[0], skew_translation[1], skew_translation[2]);

	for (int i=38; i < 53; i++) {
		// for future/ad-hoc use
		float me = 0.0; fwrite ( &me, sizeof(float), 1, fp );
	}

	char map_c[] = "MAP \0";
	fwrite ( map_c, sizeof(char), 4, fp );
	printf (" - string: [%s]\n", map_c);

	int machine_stamp = 123456789;
	fwrite ( &machine_stamp, sizeof(int), 1, fp );
	printf (" - machine stamp: %d\n", machine_stamp);

	float rms_from_mean = 0.0;
	fwrite ( &rms_from_mean, sizeof(float), 1, fp );
	printf (" - Rms deviation of map from mean density: %f\n", rms_from_mean);

	int num_labels = 1;
	fwrite ( &num_labels, sizeof(int), 1, fp );
	printf (" - Number of labels being used: %d\n", num_labels);

	char label[81] = "Created by MVis, Greg Pintilie, pintilie@mit.edu";
	fwrite ( label, sizeof(char), 80, fp );
	printf (" - label %d: %s\n", 0, label);
	memset ( label, 0, 81 * sizeof(char) );
	for (int i=1; i < 10; i++) {
		fwrite ( label, sizeof(char), 80, fp );
		printf (" - label %d: %s\n", i, label);
	}

	printf ("\n");

	//for (int i=0; i < n[0]; i++) {
	//	for (int j=0; j < n[1]; j++) {
	//		for (int k=0; k < n[2]; k++) {
	//			float val = (float) ( M[i][j][n[2]-k-1] );
	//			fwrite ( &val, sizeof(float), 1, fp );
	//		}
	//	}
	//}

	for (int k=0; k < n[2]; k++) { // z section
		for (int j=0; j < n[1]; j++) { // y column
			for (int i=0; i < n[0]; i++) { // x row
				//float val = (float) ( M[n[2]-k-1][i][j] );
				float val = (float) ( M[k][j][i] );
				fwrite ( &val, sizeof(float), 1, fp );
			}
		}
	}


	fclose (fp);
	printf ("Finished writing CCP4 map file\n");
}



void DensityMap::WriteMRC (const char *fname)
{
	name = fname;
	string map_file = name + string(".mrc");

	printf ( "Writing map to: %s\n", map_file.c_str() );
	FILE *fp = fopen ( map_file.c_str(), "wb" );

	if ( fp == NULL ) {
		printf (" - could not open\n");
		exit (0);
	}

	// format from http://www2.mrc-lmb.cam.ac.uk/image2000.html

	printf (" - map size: %d %d %d\n", n[0], n[1], n[2]);
	fwrite ( &n, sizeof(int), 3, fp );

	int type = 2;
	printf ( " - type %d\n", type );
	fwrite ( &type, sizeof(int), 1, fp );

	nstart[0] = 0; nstart[1] = 0; nstart[2] = 0;
	printf (" - map start: %d %d %d\n", nstart[0], nstart[1], nstart[2]);
	fwrite ( &nstart, sizeof(int), 3, fp );

	printf (" - # intervals: %d %d %d\n", n[0], n[1], n[2]);
	fwrite ( &n, sizeof(int), 3, fp );

	// 11-13 dimensions in angstroms
	float vSize[3];
	vSize[0] = ((double) n[0]) * step[0];
	vSize[1] = ((double) n[1]) * step[1];
	vSize[2] = ((double) n[2]) * step[2];
	printf (" - cell dimensions (angstroms): %f %f %f\n", vSize[0], vSize[1], vSize[2]);
	fwrite ( vSize, sizeof(float), 3, fp );

	float cell_angles[3] = {90.0, 90.0, 90.0};
	printf (" - cell angles: %f %f %f\n", cell_angles[0], cell_angles[1], cell_angles[2]);
	fwrite ( &cell_angles, sizeof(float), 3, fp );

	int axes[3] = {1,2,3};
	fwrite ( &axes, sizeof(int), 3, fp );
	printf (" - axes: %d %d %d\n", axes[0], axes[1], axes[2]);

	float dvals[3];
	dvals[0] = (float) dMinVal;
	dvals[1] = (float) dMaxVal;
	dvals[2] = (float) dAvgDensity;
	printf (" - min density %f, max density %f, mean density %f\n", dvals[0], dvals[1], dvals[2]);
	fwrite ( &dvals, sizeof(float), 3, fp );

	int Space_group_number = 0;
	printf (" - space group number: %d\n", Space_group_number);
	fwrite ( &Space_group_number, sizeof(int), 1, fp );

	int nb_sym_ops = 0;
	printf (" - Number of bytes used for storing symmetry operators: %d\n", nb_sym_ops);
	fwrite ( &nb_sym_ops, sizeof(int), 1, fp );

	// 25 - 49 extra space
	for (int i=25; i <= 49; i++) {
		int me = 0; fwrite ( &me, sizeof(int), 1, fp );
	}

	float o[3];
	o[0] = (float) O[0];
	o[1] = (float) O[1];
	o[2] = (float) O[2];
	printf (" - origin: %f %f %f\n", o[0], o[1], o[2]);
	fwrite ( o, sizeof(float), 3, fp );

	char map_c[] = "MAP ";
	// printf (" - string: [%s]\n", map_c);
	printf (" - string: [%s] %d %d %d %d\n", map_c, map_c[0], map_c[1], map_c[2], map_c[3]);
	fwrite ( map_c, sizeof(char), 4, fp );

	int machine_stamp = 1145307136;
	printf (" - machine stamp: %d\n", machine_stamp);
	fwrite ( &machine_stamp, sizeof(int), 1, fp );

	float rms_from_mean = 0.0;
	printf (" - Rms deviation of map from mean density: %f\n", rms_from_mean);
	fwrite ( &rms_from_mean, sizeof(float), 1, fp );

	int num_labels = 1;
	printf (" - Number of labels being used: %d\n", num_labels);
	fwrite ( &num_labels, sizeof(int), 1, fp );

	char label[81] = "Created by MVis, Greg Pintilie, pintilie@mit.edu";
	//char label[81] = "Chimera 1.2540 Fri Aug 29 11:31:20 2008";
	printf (" - label %d: %s\n", 0, label);
	fwrite ( label, sizeof(char), 80, fp );
	memset ( label, 0, 81 * sizeof(char) );
	for (int i=1; i < 10; i++) {
		printf (" - label %d: %s\n", i, label);
		fwrite ( label, sizeof(char), 80, fp );
	}

	printf ("\n");

	for (int k=0; k < n[2]; k++) { // z section
		for (int j=0; j < n[1]; j++) { // y column
			for (int i=0; i < n[0]; i++) { // x row
				//float val = (float) ( M[n[2]-k-1][i][j] );
				float val = (float) ( M[k][j][i] );
				fwrite ( &val, sizeof(float), 1, fp );
			}
		}
	}


	fclose (fp);
	printf ("Finished writing MRC map file\n");
}


void DensityMap::WriteDF3 ( const char *fname )
{
	name = fname;
	string map_file = name + string(".df3");

	printf ( "Writing map to: %s\n", map_file.c_str() );
	FILE *fptr = fopen ( map_file.c_str(), "wb" );

	if ( fptr == NULL ) {
		printf (" - could not open\n");
		exit (0);
	}

	// http://local.wasp.uwa.edu.au/~pbourke/modelling_rendering/df3/

   fputc(n[0] >> 8,fptr);
   fputc(n[0] & 0xff,fptr);
   fputc(n[1] >> 8,fptr);
   fputc(n[1] & 0xff,fptr);
   fputc(n[2] >> 8,fptr);
   fputc(n[2] & 0xff,fptr);
   
   for (int k=0; k<n[2]; k++) {
      for (int j=0; j<n[1]; j++) {
         for (int i=0; i<n[0]; i++) {
            float v = 255.0f * (float) ( (M[i][j][k]-dMinVal)/(dMaxVal-dMinVal) );
            fputc((int)v,fptr);
         }
      }
   }
   fclose (fptr);
}



void DensityMap::CopyHeaderFrom ( DensityMap *dm )
{
	vSize = dm->vSize;
	dMinVal = dm->dMinVal;
	dMaxVal = dm->dMaxVal;
	dAvgDensity = dm->dAvgDensity;
	dCellVolume = dm->dCellVolume;
	step = dm->step;
	O = dm->O;
	// nstart[0] = dm->nstart[0]; nstart[1] = dm->nstart[1]; nstart[2] = dm->nstart[2];
}


void DensityMap::CopyFrom ( DensityMap *dm )
{
	for (int k=0; k < dm->n[2]; k++) { // z section
		for (int j=0; j < dm->n[1]; j++) { // y column
			for (int i=0; i < dm->n[0]; i++) { // x row
				M[k][j][i] = dm->M[k][j][i];
			}
		}
	}
}


void DensityMap::Add ( DensityMap *dm )
{
	printf ( "adding map %s, %d %d %d\n", dm->name.c_str(), dm->n[0], dm->n[1], dm->n[2] );
	int kb = dm->n[2];
	for (int k=0; k < dm->n[2]; k++) { // z section
		// printf ("%d\n", k+kb);
		for (int j=0; j < dm->n[1]; j++) { // y column
			for (int i=0; i < dm->n[0]; i++) { // x row
				M[k+kb-1][i][j] = dm->M[k][i][j];
			}
		}
	}
}

void DensityMap::TAdd ( DensityMap *dm )
{
	int kb = dm->n[2];
	printf ( "adding T map %s, %d %d %d, dim %d\n", dm->name.c_str(), dm->n[0], dm->n[1], dm->n[2], kb );
	for (int k=kb-2, k2=1; k >=0; k--, k2++) { // z section
		for (int j=0; j < dm->n[1]; j++) { // y column
			for (int i=0; i < dm->n[0]; i++) { // x row
				M[k][i][j] = dm->M[k2][i][j];
			}
		}
	}
}


DensityMap* DensityMap::SubSample ( int sn )
{
	DensityMap *s = new DensityMap ();

	char nname[100];
	sprintf ( nname, "%s_s%d", name.c_str(), sn );
	s->name = nname;

	s->n[0] = n[0] / sn;
	s->n[1] = n[1] / sn;
	s->n[2] = n[2] / sn;

	printf ("subsampling %s (%d %d %d) -> %s (%d %d %d)\n",
		name.c_str(), n[0], n[1], n[2],
		s->name.c_str(), s->n[0], s->n[1], s->n[2] );

	s->Allocate ( s->n[0], s->n[1], s->n[2] );

	s->O = O;
	s->step = step * (double) sn;
	s->vSize = vSize;

	s->vSize = Vector3d ( ((double) s->n[0]) * s->step[0], ((double) s->n[1]) * s->step[1], ((double) s->n[2]) * s->step[2] );
	printf (" - %s size: %f %f %f\n", name.c_str(), vSize[0], vSize[1], vSize[2]);
	printf (" - %s size: %f %f %f\n", s->name.c_str(), s->vSize[0], s->vSize[1], s->vSize[2]);

	s->dCellVolume = s->step[0] * s->step[1] * s->step[2];

	s->dMaxVal = -1e99; s->dMinVal = 1e99;
	s->dAvgDensity = 0.0;

	for (int k=0; k < s->n[2]; k++) {
		for (int j=0; j < s->n[1]; j++) {
			for (int i=0; i < s->n[0]; i++) {

				double va = M[k*sn][i*sn][j*sn];
				s->M[k][i][j] = va;

				if ( va > s->dMaxVal ) s->dMaxVal = va;
				if ( va < s->dMinVal ) s->dMinVal = va;
				s->dAvgDensity += va;
			}
		}
	}

	s->dAvgDensity /= (double) ( n[0]*n[1]*n[2] );
	return s;
}






double round (double);

void DensityMap::Allocate ( int n1, int n2, int n3 )
{
	n[0] = n1;
	n[1] = n2;
	n[2] = n3;

	printf (" - allocating map: %s - %d %d %d\n", name.c_str(), n1, n2, n3 );

	M.resize ( n[2] );
	for (int k=0; k < n[2]; k++) {
		//vector < vector<double> > &vi = M[i];
		M[k].resize ( n[1] );
		for (int j=0; j < n[1]; j++) {
			//vector < double > &vj = vi[j];
			M[k][j].resize ( n[0] );
			for (int i=0; i < n[0]; i++) {
				//vj[k] = 0.0;
				M[k][j][i] = 0.0;
			}
		}
	}
}


void DensityMap::SliceX (int x, char *suffix)
{
	char fname[100];
	sprintf ( fname, "%s_%d%s.txt", name.c_str(), x, suffix );
	
	FILE *fp = fopen ( fname, "w" );
	if ( fp == NULL ) {
		printf ("Could not open %s\n", fname);
		return;
	}

	printf ("Writing slice to %s\n\n", fname);

	for (int k=0; k < n[2]; k++) { // z section
		for (int j=0; j < n[1]; j++) { // y column
			fprintf (fp, "%f%s", M[k][j][x], j==n[1]-1?"\n":" ");
		}
	}

	fclose (fp);
}


void DensityMap::GradSliceX (int x, char *suffix)
{
	char fname[100];
	sprintf ( fname, "%s_%d%s.txt", name.c_str(), x, suffix );
	
	FILE *fp = fopen ( fname, "w" );
	if ( fp == NULL ) {
		printf ("Could not open %s\n", fname);
		return;
	}

	printf ("Writing grad slice to %s\n\n", fname);

	for (int k=0; k < n[2]; k++) { // z section
		for (int j=0; j < n[1]; j++) { // y column
			Vector3d g = G[k][j][x];
			fprintf (fp, "%f%s", g.Length(), j==n[1]-1?"\n":" ");
		}
	}

	fclose (fp);
}





double DensityMap::AverageBlockDensity ( int il, int ih, int jl, int jh, int kl, int kh )
{

	double totD = 0.0, N = 0.0;

	for (int k=kl; k <= kh; k++) { // z section
		for (int j=jl; j <= jh; j++) { // y column
			for (int i=il; i <= ih; i++) { // x row
				totD += M[k][i][j];
				N += 1.0;
			}
		}
	}

	return totD / N;
}



DensityMap* DensityMap::Smooth ()
{
	int d = 5;

	DensityMap *s = new DensityMap ();
	s->name = name + string ("_sm");

	printf ("smoothing %s -> %s\n", name.c_str(), s->name.c_str() );

	s->Allocate ( n[0], n[1], n[2] );
	s->O = O; s->step = step; s->vSize = vSize; s->dCellVolume = dCellVolume;

	s->dMaxVal = -1e99; s->dMinVal = 1e99;
	s->dAvgDensity = 0.0;

	for (int k=0; k < n[2]; k++) {
		for (int j=0; j < n[1]; j++) {
			for (int i=0; i < n[0]; i++) {

				double N = 0.0;
				double V = 0.0;

				for (int ki=k-d; ki <= k+d; ki++) 
				{
					if ( ki < 0 || ki >= n[2] ) continue;

					for (int ji=j-d; ji <= j+d; ji++) 
					{
						if ( ji < 0 || ji >= n[1] ) continue;

						for (int ii=i-d; ii <= i+d; ii++) 
						{
							if ( ii < 0 || ii >= n[0] ) continue;

							N += 1.0;
							V += M[ki][ii][ji];
						}
					}
				}

				double va = V / N;
				s->M[k][i][j] = va;
				if ( va > s->dMaxVal ) s->dMaxVal = va;
				if ( va < s->dMinVal ) s->dMinVal = va;
				s->dAvgDensity += va;
			}
		}
	}

	s->dAvgDensity /= (double) ( n[0]*n[1]*n[2] );
	return s;
}



DensityMap* DensityMap::GaussianConvolve ( double dWidth )
{
	char gname[100];
	sprintf ( gname, "%s_g%.1f", name.c_str(), dWidth );

	DensityMap *s = new DensityMap ();
	s->name = gname;

	printf ("convolving %s with Gaussian, width %.4f -> %s\n",
		name.c_str(), dWidth, s->name.c_str() );

	s->Allocate ( n[0], n[1], n[2] );
	s->O = O;
	s->step = step;
	s->vSize = vSize;
	s->dCellVolume = dCellVolume;

	s->dMaxVal = -1e99; s->dMinVal = 1e99;
	s->dAvgDensity = 0.0;

	double dStd = dWidth / 2.0;
	double dm2SqStd = - 1.0 / (2.0 * dStd * dStd);
	int d = (int) ceil ( 4.0 * dStd / step[0] );
	//int d1 = (int) ceil ( 5.0 * dStd / s->step[1] );
	//int d2 = (int) ceil ( 5.0 * dStd / s->step[2] );

	printf ( " - kernel size %d\n", d );

	double dSmallestGV = 1e99;
	double dLargestGV = -1e99;

	printf ("%d\\\n", n[2]);
	for (int k=0; k < n[2]; k++)
	{
		printf ("%d,", k+1); fflush(stdout);

		for (int j=0; j < n[1]; j++) {
			for (int i=0; i < n[0]; i++) {

				double V = 0.0;
				double N = 0.0;

				for (int dk=-d; dk <= d; dk++) 
				{
					int kk = dk+k;
					if ( kk < 0 || kk >= n[2] ) continue;

					for (int dj=-d; dj <= d; dj++) 
					{
						int jj = dj+j;
						if ( jj < 0 || jj >= n[1] ) continue;

						for (int di=-d; di <= d; di++) 
						{
							int ii = di+i;
							if ( ii < 0 || ii >= n[0] ) continue;

							Vector3d p = Vector3d ( (double)di*step[0], (double)dj*step[1], (double)dk*step[2] );

							double gv = exp ( dm2SqStd * p.SquaredLength() );

							V += gv * M[kk][jj][ii];
							N += gv;

							if ( gv < dSmallestGV ) dSmallestGV = gv;
							if ( gv > dLargestGV ) dLargestGV = gv;
						}
					}
				}

				double va = V / N;
				s->M[k][j][i] = va;
				if ( va > s->dMaxVal ) s->dMaxVal = va;
				if ( va < s->dMinVal ) s->dMinVal = va;
				s->dAvgDensity += va;
			}
		}
	}
	printf ("\n");

	s->dAvgDensity /= (double) ( n[0]*n[1]*n[2] );

	printf (" - smallest gaussian val: %e\n", dSmallestGV );
	printf (" - largest gaussian val: %e\n", dLargestGV );

	return s;
}


void DensityMap::GaussianConvolveTo ( double dWidth, DensityMap *s )
{
	printf ("convolving %s with Gaussian, width %.4f -> %s\n",
		name.c_str(), dWidth, s->name.c_str() );

	double dStd = dWidth / 2.0;
	double dm2SqStd = - 1.0 / (2.0 * dStd * dStd);
	int d = (int) ceil ( 4.0 * dStd / step[0] );
	//int d1 = (int) ceil ( 5.0 * dStd / s->step[1] );
	//int d2 = (int) ceil ( 5.0 * dStd / s->step[2] );

	printf ( " - kernel size %d\n", d );

	double dSmallestGV = 1e99;
	double dLargestGV = -1e99;

	printf (" - %d: ", n[2]);
	for (int k=0; k < n[2]; k++)
	{
		printf ("%d,", k+1); fflush(stdout);
		for (int j=0; j < n[1]; j++)
		{
			for (int i=0; i < n[0]; i++) {
				double V = 0.0;
				double N = 0.0;
				for (int dk=-d; dk <= d; dk++) 
				{
					int kk = dk+k;
					if ( kk < 0 || kk >= n[2] ) continue;

					for (int dj=-d; dj <= d; dj++) 
					{
						int jj = dj+j;
						if ( jj < 0 || jj >= n[1] ) continue;

						for (int di=-d; di <= d; di++) 
						{
							int ii = di+i;
							if ( ii < 0 || ii >= n[0] ) continue;

							Vector3d p = Vector3d ( (double)di*step[0], (double)dj*step[1], (double)dk*step[2] );

							double gv = exp ( dm2SqStd * p.SquaredLength() );

							V += gv * M[kk][jj][ii];
							N += gv;

							if ( gv < dSmallestGV ) dSmallestGV = gv;
							if ( gv > dLargestGV ) dLargestGV = gv;
						}
					}
				}

				double va = V / N;
				s->M[k][j][i] = va;
				if ( va > s->dMaxVal ) s->dMaxVal = va;
				if ( va < s->dMinVal ) s->dMinVal = va;
				s->dAvgDensity += va;
			}
		}
	}
	printf ("\n");

	s->dAvgDensity /= (double) ( n[0]*n[1]*n[2] );

	printf (" - smallest gaussian val: %e\n", dSmallestGV );
	printf (" - largest gaussian val: %e\n", dLargestGV );
}






void DensityMap::ZeroBelow (double val)
{
	printf ("zeroing below %.1f\n", val);

	for (int k=0; k < n[2]; k++) {
		for (int j=0; j < n[1]; j++) {
			for (int i=0; i < n[0]; i++) {

				if ( M[k][i][j] < val )
				{
					M[k][i][j] = 0.0;
				}
			}
		}
	}
}



void DensityMap::DifToVal ( double dval )
{
	for (int k=0; k < n[2]; k++) { // z section
		for (int j=0; j < n[1]; j++) { // y column
			for (int i=0; i < n[0]; i++) { // x row
				M[k][i][j] = dval - M[k][i][j];
			}
		}
	}
}


void DensityMap::Normalize ()
{
	dMaxVal = -1e99; dMinVal = 1e99;
	dAvgDensity = 0.0;

	for (int k=0; k < n[2]; k++)
	{
		for (int j=0; j < n[1]; j++)
		{
			for (int i=0; i < n[0]; i++)
			{
				M[k][j][i] *= -1;
				double va = M[k][j][i];
				dAvgDensity += va;
				if ( va > dMaxVal ) dMaxVal = va;
				if ( va < dMinVal ) dMinVal = va;
			}
		}
	}
	dAvgDensity /= (double) (n[0]*n[1]*n[2]);
	printf ("Normalizing... before: min %f, max %f, avg %f\n", dMinVal, dMaxVal, dAvgDensity);

	for (int k=0; k < n[2]; k++)
	{
		for (int j=0; j < n[1]; j++)
		{
			for (int i=0; i < n[0]; i++)
			{
				double va = (M[k][j][i] - dMinVal) / dMaxVal;
				M[k][j][i] = va;
			}
		}
	}

	dMaxVal = -1e99; dMinVal = 1e99;
	dAvgDensity = 0.0;

	for (int k=0; k < n[2]; k++) { // z section
		for (int j=0; j < n[1]; j++) { // y column
			for (int i=0; i < n[0]; i++) { // x row
				double va = M[k][j][i];
				dAvgDensity += va;
				if ( va > dMaxVal ) dMaxVal = va;
				if ( va < dMinVal ) dMinVal = va;
			}
		}
	}

	dAvgDensity /= (double) (n[0]*n[1]*n[2]);

	printf ("After: min %f, max %f, avg %f\n", dMinVal, dMaxVal, dAvgDensity);
}


void DensityMap::Equalize ()
{
	int nsteps = 10;
	for (int s = 1; s <= nsteps; s++ ) {
		printf ( "%d/%d...\n", s, nsteps );
		EqualizeStep ( false );
		//EqualizeStep ( s == nsteps );
	}

	double globalAvgd = AverageBlockDensity ( 0, n[0]-1, 0, n[1]-1, 0, n[2]-1 );
	printf ("Global avg. density after %f\n", globalAvgd);
}


double DensityMap::EqualizeStep (bool bPrint)
{
	// printf ("\n_______ equalizing _______\n");
	double globalAvgd = AverageBlockDensity ( 0, n[0]-1, 0, n[1]-1, 0, n[2]-1 );

	int bsize = 20;

	double dLargestDif = 0.0;

	for (int kb = 0; kb < 10; kb++ )
	{
		int kl = kb*bsize, kh = (kb+1)*bsize - 1;

		for (int jb = 0; jb < 10; jb++ )
		{
			int jl = jb*bsize, jh = (jb+1)*bsize - 1;

			for (int ib = 0; ib < 10; ib++ )
			{
				int il = ib*bsize, ih = (ib+1)*bsize - 1;

				double blockAvgD = AverageBlockDensity ( il, ih, jl, jh, kl, kh );

				if ( blockAvgD < 1e-6 ) {
					if ( bPrint ) printf ("##### ");
				}
				else {
					if ( bPrint ) printf ("%.3f ", blockAvgD);

					double dif = globalAvgd - blockAvgD;
					if ( dif*dif > dLargestDif ) {
						dLargestDif = sqrt ( dif*dif );
					}

					for (int k=kl; k <= kh; k++) { // z section
						for (int j=jl; j <= jh; j++) { // y column
							for (int i=il; i <= ih; i++) { // x row
								M[k][i][j] += dif * .1;
							}
						}
					}
				}
			}

			if ( bPrint ) printf ("\n");
		}

		if ( bPrint ) printf ("\n");
	}

	printf ("- global avg. density before %f\n", globalAvgd);
	printf ("- largest diff: %f\n", dLargestDif);
	printf ("\n");

	return dLargestDif;
}


double DensityMap::DensityValue ( double *p )
{
		double i = (p[0] - O[0]) / step[0];
		double j = (p[1] - O[1]) / step[1];
		double k = (p[2] - O[2]) / step[2];

		//printf (" - pos %f %f %f - %f %f %f\n", p[0], p[1], p[2], i, j, k );

		 double xd = i - floor(i);
		int il = (int) floor(i), ih = (int) ceil(i);

        double yd = j - floor(j);
		int jl = (int) floor(j), jh = (int) ceil(j);

        double zd = k - floor(k);
		int kl = (int) floor(k), kh = (int) ceil(k);

		if (il<0 || jl<0 || kl<0 || ih>=n[0] || jh>=n[1] || kh>=n[2] ) {
			printf ("-- index out of bounds %.3f %.3f %.3f\n", i, j, k);
		}

        // trilinear interpolation
        // http://en.wikipedia.org/wiki/Trilinear_interpolation

        double i1 = M[kl][jl][il] * (1-zd) + M[kh][jl][il] * zd;
        double i2 = M[kl][jh][il] * (1-zd) + M[kh][jh][il] * zd;
        double j1 = M[kl][jl][ih] * (1-zd) + M[kh][jl][ih] * zd;
        double j2 = M[kl][jh][ih] * (1-zd) + M[kh][jh][ih] * zd;

		double w1 = i1 * (1-yd) + i2 * yd;
        double w2 = j1 * (1-yd) + j2 * yd;
        double dval = w1 * (1-xd) + w2 * xd;

		return dval;
}




DensityMap* DensityMap::ComputeGradients ( double exp_f )
{
	printf ("Allocating gradient map %s - %d %d %d\n", name.c_str(), n[0], n[1], n[2] );
	G.resize ( n[2] );
	for (int i=0; i < n[2]; i++) {
		vector < vector< Vector3d > > &vi = G[i];
		vi.resize ( n[1] );
		for (int j=0; j < n[1]; j++) {
			vector < Vector3d > &vj = vi[j];
			vj.resize ( n[0] );
			for (int k=0; k < n[0]; k++) {
				vj[k] = Vector3d (0,0,0);
			}
		}
	}

	DensityMap *gm = new DensityMap ();
	gm->Allocate ( n[0], n[1], n[2] );
	gm->CopyHeaderFrom ( this );
	gm->name = name + string("_gmap");

	// centered difference formula :
	// http://books.google.ca/books?id=_nrjAAhOeYcC&pg=PT126&lpg=PT126&dq=centered+difference+formula&source=web&ots=5Idp7-LdTf&sig=HQA5V7cbiQhTy0W89L_c6kFpqo4&hl=en&sa=X&oi=book_result&resnum=5&ct=result

	for (int k=0; k < n[2]; k++) {
		for (int j=0; j < n[1]; j++) {
			for (int i=0; i < n[0]; i++) {

				Vector3d &v = G[k][j][i];
				v = Vector3d (0,0,0);

				if ( i == 0 ) {
				}
				else if ( i == n[0]-1 ) {
				}
				else {
					//v[0] = ( M[k][j][i+1] - M[k][j][i-1] ) / ( 2.0 * step[0] );
					v[0] = ( M[k][j][i+1] - M[k][j][i-1] ) / ( 2.0 );
				}

				if ( j == 0 ) {
				}
				else if ( j == n[1]-1 ) {
				}
				else {
					//v[1] = ( M[k][j+1][i] - M[k][j-1][i] ) / ( 2.0 * step[1] );
					v[1] = ( M[k][j+1][i] - M[k][j-1][i] ) / ( 2.0 );
				}

				if ( k == 0 ) {
				}
				else if ( k == n[2]-1 ) {
				}
				else {
					//v[2] = ( M[k+1][j][i] - M[k-1][j][i] ) / ( 2.0 * step[2] );
					v[2] = ( M[k+1][j][i] - M[k-1][j][i] ) / ( 2.0 );
				}

				if ( 0 )
				{
					if ( exp_f < 0.0 ) {
						gm->M[k][j][i] = 1.0 / ( 1.0 + v.Length() );
					}
					else {
						gm->M[k][j][i] = exp ( exp_f / ( 1.0 + v.Length() ) );
					}
				}
				else {
					gm->M[k][j][i] = v.Length();
				}
			}
		}
	}

	if ( 0 )
	{
		Vector3d &g = G[40][40][40];
		printf (" %f [%f, %f, %f]\n", M[40][40][40], g[0], g[1], g[2]);

		g = G[40][40][39];
		printf (" %f [%f, %f, %f]\n", M[40][40][39], g[0], g[1], g[2]);

		g = G[40][40][41];
		printf (" %f [%f, %f, %f]\n", M[40][40][41], g[0], g[1], g[2]);

		g = G[40][41][40];
		printf (" %f [%f, %f, %f]\n", M[40][41][40], g[0], g[1], g[2]);

		g = G[41][40][40];
		printf (" %f [%f, %f, %f]\n", M[41][40][40], g[0], g[1], g[2]);
	}

	return gm;
}



void DensityMap::GradientVal ( Vector3d &p, Vector3d &g )
{
		double i = (p[0] - O[0]) / step[0];
		double j = (p[1] - O[1]) / step[1];
		double k = (p[2] - O[2]) / step[2];

		printf ( "%.3f, %.3f, %.3f :: ", i, j, k );

		//printf (" - pos %f %f %f - %f %f %f\n", p[0], p[1], p[2], i, j, k );

		double xd = i - floor(i);
		int il = (int) floor(i), ih = (int) ceil(i);

        double yd = j - floor(j);
		int jl = (int) floor(j), jh = (int) ceil(j);

        double zd = k - floor(k);
		int kl = (int) floor(k), kh = (int) ceil(k);

		if (il<0 || jl<0 || kl<0 || ih>=n[0] || jh>=n[1] || kh>=n[2] ) {
			printf ("-- index out of bounds %.3f %.3f %.3f\n", i, j, k);
		}

        // trilinear interpolation
        // http://en.wikipedia.org/wiki/Trilinear_interpolation

		if ( 0 )
		{
			Vector3d i1 = G[kl][jl][il] * (1-zd) + G[kh][jl][il] * zd;
			Vector3d i2 = G[kl][jh][ih] * (1-zd) + G[kh][jl][ih] * zd;
			Vector3d j1 = G[kl][jh][il] * (1-zd) + G[kh][jh][il] * zd;
			Vector3d j2 = G[kl][jh][ih] * (1-zd) + G[kh][jh][ih] * zd;

			Vector3d w1 = i1 * (1-yd) + i2 * yd;
			Vector3d w2 = j1 * (1-yd) + j2 * yd;
			g = w1 * (1-xd) + w2 * xd;
		}

        Vector3d i1 = G[kl][jl][il] * (1-zd) + G[kh][jl][il] * zd;
        Vector3d i2 = G[kl][jh][il] * (1-zd) + G[kh][jh][il] * zd;
        Vector3d j1 = G[kl][jl][ih] * (1-zd) + G[kh][jl][ih] * zd;
        Vector3d j2 = G[kl][jh][ih] * (1-zd) + G[kh][jh][ih] * zd;

		Vector3d w1 = i1 * (1-yd) + i2 * yd;
        Vector3d w2 = j1 * (1-yd) + j2 * yd;
        g = w1 * (1-xd) + w2 * xd;
}



void DensityMap::ComputeLaplacian ()
{
	if ( L.size() != M.size() )
	{
		printf ("- allocating Laplacian map\n" );
		L.resize ( M.size() );
		for (int k=0; k < M.size(); k++) {
			L[k].resize ( M[k].size() );
			for (int j=0; j < M[k].size(); j++) {
				L[k][j].resize ( M[k][j].size() );
				for (int i=0; i < M[k][j].size(); i++) {
					L[k][j][i] = 0.0;
				}
			}
		}
	}

	printf ( "- computing Laplacian map\n" );

	for (int k=0; k < M.size(); k++)
	{
		for (int j=0; j < M[k].size(); j++)
		{	
			for (int i=0; i < M[k][j].size(); i++)
			{
				if      ( i == 0 || i == M[k][j].size()-1 ) L[k][j][i] = 0.0;
				else if ( j == 0 || j == M[k].size()-1 )    L[k][j][i] = 0.0;
				else if ( k == 0 || k == M.size()-1 )       L[k][j][i] = 0.0;
				else
				{
					L[k][j][i] = -6.0 * M[k][j][i] + M[k][j][i-1] + M[k][j][i+1] + M[k][j-1][i] + M[k][j+1][i] + M[k+1][j][i] + M[k-1][j][i];
				}
			}
		}		
	}
}


void DensityMap::LaplacianSmooth ( double step_size )
{
	for (int k=0; k < n[2]; k++)
	{
		for (int j=0; j < n[1]; j++)
		{	
			for (int i=0; i < n[0]; i++)
			{
				if      ( i == 0 || i == n[0]-1 ) M[k][j][i] = 0.0;
				else if ( j == 0 || j == n[1]-1 ) M[k][j][i] = 0.0;
				else if ( k == 0 || k == n[2]-1 ) M[k][j][i] = 0.0;
				else
				{
					M[k][j][i] = M[k][j][i] + step_size * ( -6.0 * M[k][j][i] + M[k][j][i-1] + M[k][j][i+1] + M[k][j-1][i] + M[k][j+1][i] + M[k+1][j][i] + M[k-1][j][i] );
				}
			}
		}		
	}
}


Regions* DensityMap::Watershed ( double dThr, double dMaxVol )
{

	printf ("Sorting voxels by density value\n");

	multimap< double, Vector3i > dList;

	for (int k=0; k < n[2]; k++)
	{
		for (int j=0; j < n[1]; j++)
		{
			for (int i=0; i < n[0]; i++)
			{
				double dval = M[k][j][i];
				M[k][j][i] = 0.0;
				//printf ( "%d %d %d - %f\n", i, j, k, dval );
				if ( dval < dThr ) continue;
				Vector3i ind = Vector3i (i,j,k);
				dList.insert ( pair< double, Vector3i > ( -dval, ind ) );
				
			}
		}
	}

	printf ("Watershedding %d voxels above %f\n", dList.size(), dThr );

	Regions *regions = new Regions ();
	int region_i = 0;
	int step = 0;
	double dTotVolume = 0.0, dLastD;
	int printEvery = dList.size() / 10;
	
	// traverse the list in order of decreasing density value

	for ( multimap< double, Vector3i >::iterator it = dList.begin(); it != dList.end(); it++ )
	{
		double dval = -it->first;
		Vector3i &ind = it->second;

		// the indexes and position we're looking at...
		int *pind = ind; int ii = pind[0], jj = pind[1], kk = pind[2];
		// Vector3d p = dmap->O + Vector3d ( ((double)ii)*dmap->step[0], ((double)jj)*dmap->step[1], ((double)kk)*dmap->step[2] );

		// map of regions the voxel is adjacent to
		// by adjacent we mean here 26-connected
		
		map < int, int > nr_ind;

		for ( int i = ii-1; i <= ii+1; i++ ) {
			for ( int j = jj-1; j <= jj+1; j++ ) {
				for ( int k = kk-1; k <= kk+1; k++ )
				{
					Vector3i nind = Vector3i ( i, j, k );
					if ( i < 0 || i >= n[0] ) continue;
					if ( j < 0 || j >= n[1] ) continue;
					if ( k < 0 || k >= n[2] ) continue;
					int reg = (int) M[k][j][i];
					if ( reg > 0 ) {
						nr_ind[ reg ] += 1;
					}
				}
			}
		}

		if ( nr_ind.size() == 0 ) {
			// no adjacent regions, the voxel starts a new region
			region_i++;
			M[kk][jj][ii] = (double) region_i;
			regions->mRegs[region_i].vi.push_back ( Vector3f(ii,jj,kk) );
			
			//Vector3d &vadded = regions->mRegs[region_i].vi [ regions->mRegs[region_i].vi.size()-1 ];
			//printf ( "%d %d %d -> added %f %f %f\n", ii, jj, kk, vadded[0], vadded[1], vadded[2] );
		}
		else if ( nr_ind.size() == 1 ) {
			// one adjacent regions, the voxel is added to the region
			int region = nr_ind.begin()->first;
			M[kk][jj][ii] = (double) region;
			regions->mRegs[region].vi.push_back ( Vector3f(ii,jj,kk) );
		}
		else
		{
			// two or more adjacent regions
			// assign it to region in which it has the most adjacent voxels
			int max_num_adj = 0;
			int max_num_adj_ri = -1;
			vector < int > rrs;
			for ( map < int, int >::iterator mmi = nr_ind.begin(); mmi != nr_ind.end(); mmi++ )
			{
				if ( mmi->second > max_num_adj ) {
					max_num_adj = mmi->second;
					max_num_adj_ri = mmi->first;
				}
				rrs.push_back ( mmi->first );
			}
			M[kk][jj][ii] = (double) max_num_adj_ri;
			regions->mRegs[max_num_adj_ri].vi.push_back ( Vector3f(ii,jj,kk) );

			if ( nr_ind.size() == 2 )
			{
				// store 'connection' between the two regions
				
				int r1 = rrs[0], r2 = rrs[1];

				//map < int, int >::iterator regis = nr_ind.begin();
				//int r1 = regis->first; regis++;
				//int r2 = regis->first;
				
				RegsConnection *rc = regions->mCons[r1][r2];
				if ( rc == NULL ) {
					rc = new RegsConnection ( /*r1, r2*/ );
					regions->mCons[r1][r2] = rc;
					regions->mCons[r2][r1] = rc;
				}
				rc->totD += dval;
				rc->N += 1.0;
			}
		}

		dTotVolume += dCellVolume;

		if ( step % printEvery == 0 ) {
			printf ( "%d/%d dval %f, vol %.2f, #regions %d\n",
				step+1, dList.size(), dval, dTotVolume, region_i );
		}

		if ( dval <= dThr )
			break;

		if ( dTotVolume >= dMaxVol )
			break;

		step++;
		dLastD = dval;
	}


	printf ( "\nStopped after %d steps, density %f, vol %.3f, #regions %d\n\n",
		step+1, dLastD, dTotVolume, region_i );
		
	return regions;
}


double DensityMap::RegionsThreshold ( Regions *regs )
{
	double minD = 1e99;
	
	for ( map< int, Region >::iterator ri = regs->mRegs.begin(); ri != regs->mRegs.end(); ri++ )
	{
		Region &r = ri->second;

		for ( int vi=0; vi < r.vi.size(); vi++ )
		{
			float *ijk = r.vi[vi];
			int i = (int) ijk[0];
			int j = (int) ijk[1];
			int k = (int) ijk[2];
			
			double dval = M[k][j][i];
			if ( dval < minD ) {
				minD = dval;
			}
		}
	}
	
	return minD;
}





