
# Copyright (c) 2009 Greg Pintilie - pintilie@mit.edu

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import chimera
import chimera.match
import os
import os.path
import Tkinter
from CGLtk import Hybrid
import subprocess
import VolumeData
import VolumeData.mrc
import _multiscale
import MultiScale.surface
import _surface
import time
import numpy
import _contour
import Scientific.LA
import Matrix
import VolumeData
import FitMap
import VolumeViewer

import random
random.seed(1)          # Get reproducible random colors.
rand = random.random


dev_menus = False


OML = chimera.openModels.list


MIN_REG_SIZE = 0
SAF_DVOL = 0.75
SAF_DBRAD = 0.3
SAF_LS_DEPTH = 4
SAF_LS_NGROUPS = 1000
OPT_THR = 1
REG_OPACITY = 0.45
MAX_NUM_GROUPS = 1000


mdlg = None
#execfile("/Users/greg/dev/chimera/Movie/gui.py")
#movie()


def umsg ( txt ) :
    print txt
    status ( txt )

def status ( txt ) :
    txt = txt.rstrip('\n')
    msg.configure(text = txt)
    msg.update_idletasks()

class aClass :
    def __init__ (self) :
        pass


class Volume_Segmentation_Dialog ( chimera.baseDialog.ModelessDialog ):

    title = "Segger 1.4"
    name = "Segger"
    buttons = ('Segment', 'Group', 'Ungroup', 'Fit', 'Options', "Close")
    help = 'http://people.csail.mit.edu/gdp/segger/'

    def fillInUI(self, parent):

        tw = parent.winfo_toplevel()
        self.toplevel_widget = tw
        tw.withdraw()

        parent.columnconfigure(2, weight = 1)

        row = 1

        menubar = Tkinter.Menu(parent, type = 'menubar', tearoff = False)
        tw.config(menu = menubar)

        regions_menu_entries = (
            'separator',
            ("Show all", self.RegSurfsShowAll),
            ("Show only selected", self.RegSurfsShow),
            ("Show adjacent", self.RegSurfsShowAdjacent),
            ("Hide", self.RegSurfsHide),
            ("Make transparent", self.RegSurfsTransparent),
            ("Make opaque", self.RegSurfsOpaque),
            ("Invert selection", self.RegSelInvert),
            'separator',
            ("Volume", self.RegionsVolume),
            'separator',
            ("Delete selected", self.DelSelRegs),
            ("Group selected", self.JoinSelRegs),
            ("Ungroup selected", self.UngroupSelRegs),
            ("Save selected to .mrc file", self.WriteSelRegionsMRCFile),
            ("Save ALL to .mrc file", self.WriteAllRegionsMRCFile),
            ("Save each to .mrc file", self.WriteEachRegionMRCFile),
            ("Mask map with selected", self.MaskMapWRegions),
            "separator",
            ("Save _regions file", self.WriteRegionsFile),
            ("Close and delete file", self.DelRegionsFiles),
            ("Close all except displayed", self.CloseRegsExcl),
            ("Restart", self.RestartSeg),
            "separator",
            ("Show axes for selected", self.ShowRegionAxesSelected),
            ("Hide all axes", self.HideRegionAxes),
            )

        self.regsVisMode = Tkinter.StringVar()
        self.regsVisMode.set ( 'Voxel_Surfaces' )

        rmenu = Hybrid.cascade_menu(menubar, 'Regions', regions_menu_entries)

        if dev_menus:
            rmenu.add_separator()
            for lbl, var, val, cmd in (
                ("Surfaces around voxels", self.regsVisMode,
                 'Voxel_Surfaces', self.RegsDispUpdate),
                ("Map iso-surfaces", self.regsVisMode,
                 'Iso_Surfaces', self.RegsDispUpdate),
                #(" - delete files", self.deleteCloseFiles, 1, None),
                ):
                rmenu.add_radiobutton(label = lbl, variable = var, value = val,
                                      command = cmd)
            for lbl, cmd in (("Adjacency graph",self.RegionsAdjGraph),
                             ("Select non-placed", self.SelectNonPlacedRegions),
                             #("Apply threshold", self.RegsDispThr),
                             ("Group using all fits", self.GroupUsingFits),
                             #("Ungroup ALL", self.UngroupAllRegs),
                             #("Reduce map", self.ReduceMap),
                             #("Close", self.CloseRegions),
                             ):
                rmenu.add_command(label = lbl, command = cmd)
            #self.deleteCloseFiles = Tkinter.IntVar()
            #self.deleteCloseFiles.set ( 1 )


        self.UseAllMods = Tkinter.IntVar()
        self.UseAllMods.set ( 0 )

        self.strucCopyAfterFit = Tkinter.IntVar()
        self.strucCopyAfterFit.set ( 0 )

        fit_menu_entries = (
            #("Selected region", self.FitToSelRegion),
            #("Each selected region", self.FitToSelectedRegions),
            #("Combined selected regions", self.FitMapToSelRGroup),
            #("Regions around selected", self.FitMapToRegionsAroundSel),
            #("Groups of regions", self.FitMapToRGroups),
            #'separator',
            ("Show molecule axes", self.StrucShowAxes),
            ("Hide molecule axes", self.StrucHideAxes),
            #("Place best fits", self.PlaceBestFits),
            ("Show overlapping regions", self.ShowOverlappingRegions),
            ("Save fitted structure", self.SaveStrucFit),
            )

        fmenu = Hybrid.cascade_menu(menubar, 'Fitting', fit_menu_entries)

        if dev_menus:
            fmenu.add_separator()
            for lbl, var, val in (
                #("By principal axes", self.rotaSearch, 0),
                #("By rotation", self.rotaSearch, 1),
                ):
                fmenu.add_radiobutton(label = lbl, variable = var, value = val)
            for lbl, var in (("Chains = models", self.UseAllMods),
                             #("Save all fits", self.strucCopyAfterFit),
                             ):
                fmenu.add_checkbutton(label = lbl, variable = var)
            for lbl, cmd in (
                #("Replicate with Bio-Matrices", self.StrucBioMT),
                ("Sim chain maps", self.SimChainMaps),
                ("Make chain maps", self.StrucChainMaps),
                (" - show chain maps", self.StrucShowChainMaps),
                (" - hide chain maps", self.StrucHideChainMaps),
                (" - close chain maps", self.StrucCloseChainMaps),
                (" - overlap regions", self.StrucChMapsOvRegs),
                (" - close 1 chain map", self.StrucCloseChainMap),
                #(" - delete chain maps", self.StrucDelChainMaps),
                #(" - show chain/regions", self.ShowChRegs),
                ("Segmentation accuracy", self.SegAccuracy),
                ("Fit RMSD", self.FitRMSD),
                #("Sel Chains", self.GetSelectedChains),
                ("Align to selected chain", self.AlignToSel),
                ("Extract proteins", self.ExtractProteins),
                #("Shape match with selected regions", self.SelRegsShapeScore),
                #("Adjust threshold for best match", self.SelRegsOptimizeShapeScore),
                #("Best shape score", self.StrucBestShapeScore),
                ("Group regions", self.StrucGroupRegions),
                ("Next group", self.NextRGroup),
                #("Find group w/selected", self.FindGroupFromSelRegs),
                #("Get fits", self.GetFits),
                #("Next fit", self.NextFit),
                #("Best fit around selected", self.FitMapToRSelGroups),
                #("Fit ALL around selected", self.FitMapsToRegionsAroundSel),
                #("Fit ALL structures", self.FitMapsToRGroups),
                #("Displayed volume", self.StrucMapVolume),
                ("Fit to map (local)", self.FitSMapToDMap),
                ("Zero density map", self.ZeroDMap_with_FMap),
                ("Extract density", self.TakeDMap_with_FMap),
                ("Masked map", self.MaskedMap),
                ("Next bio Matrix", self.NextBioMt),
                ("All bio matrices", self.GoBioMt),
                             ):
                fmenu.add_command(label = lbl, command = cmd)

        from chimera.tkgui import aquaMenuBar
        aquaMenuBar(menubar, parent, row = 0, columnspan=3)


        f = Tkinter.Frame(parent)
        f.grid(column=0, row=row, sticky='ew')
        row += 1

        l = Tkinter.Label(f, text='Segment map')
        l.grid(column=0, row=0, sticky='w')

        self.dmap = Tkinter.StringVar(parent)

        self.mb  = Tkinter.Menubutton ( f, textvariable=self.dmap, relief=Tkinter.RAISED )
        self.mb.grid (column=1, row=0, sticky='we', padx=5)
        self.mb.menu  =  Tkinter.Menu ( self.mb, tearoff=0, postcommand=self.MapMenu )
        self.mb["menu"]  =  self.mb.menu


        f = Tkinter.Frame(parent)
        f.grid(column=0, row=row, sticky='ew')
        row += 1

        l = Tkinter.Label(f, text='Current segmentation')
        l.grid(column=0, row=0, sticky='w')

        self.regions_file = Tkinter.StringVar(parent)

        rmb  = Tkinter.Menubutton ( f, textvariable=self.regions_file,
                                    relief=Tkinter.RAISED )
        rmb.grid (column=1, row=0, sticky='we', padx=5)
        rmb.menu  =  Tkinter.Menu ( rmb, tearoff=0,
                                    postcommand=self.RegionsFileMenu )
        self.mbRegionsFileMenu = rmb.menu
        rmb["menu"]  = rmb.menu

        rc = Tkinter.Label(f, text='')
        rc.grid(column=2, row=0, sticky='w')
        self.regionCount = rc


        f = Tkinter.Frame(parent)
        f.grid(column=0, row=row, sticky='ew')
        row += 1

        l = Tkinter.Label(f, text='Structure to fit')
        l.grid(column=0, row=0, sticky='w')
 
        self.struc = Tkinter.StringVar(parent)
        self.strucMB  = Tkinter.Menubutton ( f, textvariable=self.struc, relief=Tkinter.RAISED )
        self.strucMB.grid (column=1, row=0, sticky='we', padx=5)
        self.strucMB.menu  =  Tkinter.Menu ( self.strucMB, tearoff=0, postcommand=self.StrucMenu )
        self.strucMB["menu"]  =  self.strucMB.menu

        row = row + 1

        self.struc_map = Tkinter.StringVar(parent)

        op = Hybrid.Popup_Panel(parent)
        opf = op.frame
        opf.grid(row = row, column = 0, sticky = 'news')
        opf.grid_remove()
        opf.columnconfigure(0, weight=1)
        self.optionsPanel = op.panel_shown_variable
        row += 1
        orow = 0

        cb = op.make_close_button(opf)
        cb.grid(row = orow, column = 0, sticky = 'e')

        l = Tkinter.Label(opf, text='Segmenting Options', font = 'TkCaptionFont')
        l.grid(column=0, row=orow, sticky='w', pady=5)
        orow += 1

        sopt = Tkinter.Frame(opf)
        sopt.grid(column=0, row=orow, sticky='ew', padx=10)
        orow += 1
        sorow = 0

        f = Tkinter.Frame(sopt)
        f.grid(column=0, row=sorow, sticky='w')
        sorow += 1

        l = Tkinter.Label(f, text='Keep only regions covering >=')
        l.grid(column=0, row=0, sticky='w')

        self.minRegionSize = Tkinter.StringVar(parent)
        self.minRegionSize.set ( '1' )
        e = Tkinter.Entry(f, width=5, textvariable=self.minRegionSize)
        e.grid(column=1, row=0, sticky='w', padx=5)

        l = Tkinter.Label(f, text='grid points')
        l.grid(column=2, row=0, sticky='w')


        f = Tkinter.Frame(sopt)
        f.grid(column=0, row=sorow, sticky='w')
        sorow += 1

        l = Tkinter.Label(f, text='Number of smoothing steps')
        l.grid(column=0, row=0, sticky='w')

        self.numSteps = Tkinter.StringVar(sopt)
        self.numSteps.set ( '5' )
        e = Tkinter.Entry(f, width=2, textvariable=self.numSteps)
        e.grid(column=1, row=0, sticky='w', padx=5)

        l = Tkinter.Label(f, text='step size')
        l.grid(column=2, row=0, sticky='w')

        self.stepSize = Tkinter.StringVar(sopt)
        e = Tkinter.Entry(f, width=5, textvariable=self.stepSize)
        e.grid(column=4, row=0, sticky='w', padx=5)


        #if dev_menus :
        #    b = Tkinter.Button(f, text="Go", command=self.Smooth_and_Group)
        #    b.grid (column=5, row=0, sticky='w', padx=5)


        f = Tkinter.Frame(sopt)
        f.grid(column=0, row=sorow, sticky='w')

        l = Tkinter.Label(f, text='Stop if number of regions <=')
        l.grid(column=0, row=0, sticky='w')

        self.targNRegions = Tkinter.StringVar(sopt)
        self.targNRegions.set ( '1' )
        e = Tkinter.Entry(f, width=5, textvariable=self.targNRegions)
        e.grid(column=1, row=0, sticky='w', padx=5)


        sorow += 1

        f = Tkinter.Frame(sopt)
        f.grid(column=0, row=sorow, sticky='w')

        self.saveIntermediateSegs = Tkinter.IntVar()
        self.saveIntermediateSegs.set ( 0 )


        c = Tkinter.Checkbutton(f, text="Save _regions files after each step",
                                variable=self.saveIntermediateSegs)
        c.grid (row = sorow, column=0, sticky='w')


        sorow += 1

        f = Tkinter.Frame(sopt)
        f.grid(column=0, row=sorow, sticky='w')

        l = Tkinter.Label(f, text='Maximum number of regions to display')
        l.grid(column=0, row=0, sticky='w')

        self.maxNumRegions = Tkinter.StringVar(sopt)
        self.maxNumRegions.set ( '2000' )
        e = Tkinter.Entry(f, width=5, textvariable=self.maxNumRegions)
        e.grid(column=1, row=0, sticky='w', padx=5)





        l = Tkinter.Label(opf, text='Fitting Options', font = 'TkCaptionFont')
        l.grid(column=0, row=orow, sticky='w', pady=5)
        orow += 1

        fopt = Tkinter.Frame(opf)
        fopt.grid(column=0, row=orow, sticky='ew', padx=10)
        orow += 1
        forow = 0

        f = Tkinter.Frame(fopt)
        f.grid(column=0, row=forow, sticky='e')
        forow += 1

        l = Tkinter.Label(f, text='Density map resolution:')
        l.grid(column=0, row=0, sticky='w')

        self.simRes = Tkinter.StringVar(fopt)
        e = Tkinter.Entry(f, width=4, textvariable=self.simRes)
        e.grid(column=1, row=0, sticky='w', padx=5)

        l = Tkinter.Label(f, text='grid spacing:')
        l.grid(column=2, row=0, sticky='w')

        self.simGridSp = Tkinter.StringVar(fopt)
        e = Tkinter.Entry(f, width=4, textvariable=self.simGridSp)
        e.grid(column=3, row=0, sticky='w', padx=5)

        b = Tkinter.Button(f, text="Calculate Map", command=self.GenStrucMap)
        b.grid (column=4, row=0, sticky='w', padx=5)



        forow += 1

        l = Tkinter.Label(fopt, text='Which regions to align the structure to:')
        l.grid(column=0, row=forow, sticky='w')

        forow += 1

        f = Tkinter.Frame(fopt)
        f.grid(column=0, row=forow, sticky='w')

        self.alignTo = Tkinter.StringVar()
        self.alignTo.set ( 'combined_selected_regions' )

        l = Tkinter.Label(f, text=' ', width=5)
        l.grid(column=0, row=0, sticky='w')

        c = Tkinter.Radiobutton(f, text="Combined selected regions", variable=self.alignTo, value = 'combined_selected_regions')
        c.grid (column=1, row=0, sticky='w')

        c = Tkinter.Radiobutton(f, text="Each selected region", variable=self.alignTo, value = 'each_selected_region')
        c.grid (column=1, row=1, sticky='w')

        c = Tkinter.Radiobutton(f, text="Groups of regions including selected region", variable=self.alignTo, value = 'around_selected')
        c.grid (column=1, row=2, sticky='w')

        c = Tkinter.Radiobutton(f, text="Groups of regions including all regions", variable=self.alignTo, value = 'all_groups')
        c.grid (column=1, row=3, sticky='w')



        forow += 1

        l = Tkinter.Label(fopt, text='Alignment Method:')
        l.grid(column=0, row=forow, sticky='w')

        forow += 1

        f = Tkinter.Frame(fopt)
        f.grid(column=0, row=forow, sticky='w')

        self.rotaSearch = Tkinter.IntVar()
        self.rotaSearch.set ( 0 )

        l = Tkinter.Label(f, text=' ', width=5)
        l.grid(column=0, row=0, sticky='w')

        c = Tkinter.Radiobutton(f, text="Align principal axes (faster)", variable=self.rotaSearch, value = 0)
        c.grid (column=1, row = 0, sticky='w')

        c = Tkinter.Radiobutton(f, text="Rotational search (more reliable)", variable=self.rotaSearch, value = 1)
        c.grid (column=1, row = 1, sticky='w')



        forow += 1

        c = Tkinter.Checkbutton(fopt, text="Save each fit when doing multiple alignments",
                                variable=self.strucCopyAfterFit)
        c.grid (row = forow, column=0, sticky='w')

        forow += 1


        f = Tkinter.Frame(fopt)
        f.grid(column=0, row=forow, sticky='w')
        forow += 1

        l = Tkinter.Label( f, text="Top number of fits to place" ) 
        l.grid(column=0, row=0, sticky='e')

        self.numFits = Tkinter.StringVar(opf)
        self.numFits.set ( '1' )
        e = Tkinter.Entry(f, width=3, textvariable=self.numFits)
        e.grid(column=1, row=0, sticky='w', padx=5)

        b = Tkinter.Button(f, text="Place", command=self.PlaceBestFits)
        b.grid (column=2, row=0, sticky='w', padx=5)


        dummyFrame = Tkinter.Frame(parent, relief='groove', borderwidth=1)
        Tkinter.Frame(dummyFrame).pack()
        dummyFrame.grid(row=row,column=0,columnspan=7, pady=7, sticky='we')

        row = row + 1

        global msg
        msg = Tkinter.Label(parent, width = 60, anchor = 'w', justify = 'left')
        msg.grid(column=0, row=row, sticky='ew')
        self.msg = msg

        umsg ( 'Select an open density map in the field above and press Segment!' )
        row += 1


        for m in OML() :
            if type(m) == VolumeViewer.volume.Volume :
                # if m.name.find ( '.mrc' ) < 0 : continue
                self.dmap.set ( m.name )
                break
        self.SetResolution()


        for m in OML() :

            try :
                self.struc.set ( m.mol.name )
                self.struc_map.set ( m.name )
            except : pass

            #if type(m) == chimera.Molecule :
            #    self.struc.set ( m.name )

            try : m.seg_map
            except : continue

            if m.display == True :
                self.regions_file.set ( m.name )
                try : self.dmap.set ( m.seg_map.name )
                except : pass

        chimera.openModels.addRemoveHandler(self.ModelClosed, None)


    def Options(self):

        self.optionsPanel.set(not self.optionsPanel.get())


    def ModelClosed(self, trigger, n, mlist):

        # Clear menus that are showing closed models.
        for mvar in (self.dmap, self.struc, self.struc_map, self.regions_file):
            if len( mvar.get() ) > 0 and getMod ( mvar.get() ) is None:
                mvar.set('')


    def MapMenu ( self ) :

        dmap = None
        if len( self.dmap.get() ) > 0 :
            dmap = getMod ( self.dmap.get() )

        self.mb.menu.delete ( 0, 'end' )        # Clear menu

        open_maps = {}
        
        for m in OML() :
            if type(m) != VolumeViewer.volume.Volume :
                continue

            else :
                # print "adding", m.name
                self.mb.menu.add_radiobutton (
                    label=m.name, variable=self.dmap, command=self.MapSelected )
            
        if dmap :
            path = os.path.dirname(dmap.data.path) + os.path.sep



    def MapSelected ( self ) :
        dmap_name = self.dmap.get()
        if len(dmap_name) == 0 : return
        
        dmap = getMod ( dmap_name )
        if dmap == None : return

        dmap.display = True

        self.SetResolution()

    def SetResolution ( self ):
        dmap_name = self.dmap.get()
        if len(dmap_name) == 0 : return
        
        dmap = getMod ( dmap_name )
        if dmap == None : return

        if len ( self.stepSize.get() ) == 0:
            res = min(dmap.data.step) * 3
            self.stepSize.set ( '%.3g' % res )

        if len ( self.simRes.get() ) == 0:
            res = min(dmap.data.step) * 3
            self.simRes.set ( '%.3g' % res )
            self.simGridSp.set ( '%.3g' % (res/3.0) )

    def RegionsFileMenu ( self ) :

        self.mbRegionsFileMenu.delete ( 0, 'end' )      # Clear menu

        sm_map = self.dmap.get()
        # print "Regions file for", sm_map

        smm = getMod ( sm_map )
        if smm == None :
            # print sm_map, "not found in open models list"
            self.regions_file.set ( "" )
            self.dmap.set ( "" )
            return

        path = os.path.dirname ( smm.data.path ) + os.path.sep
        bname = os.path.splitext ( smm.name ) [0]
        # print "-", smm.name, '-', bname, '-', path

        cur_sel_found = False
        open_names = {}
        for m in OML() :

            # if type(m) != _surface.SurfaceModel : continue
            # if m.name.find ( bname ) < 0

            try : m.seg_map
            except : continue

            # if m.name.find('_regions') < 0 : continue

            if m.name == self.regions_file.get() : cur_sel_found = True
            open_names[m.name] = 1

        if cur_sel_found == False : self.regions_file.set ( "" )

        if len(open_names.keys() ) > 0 :
            self.mbRegionsFileMenu.add_radiobutton ( label="Open regions files:" )
            self.mbRegionsFileMenu.add_separator()

            gs_files = {}
            for mname in open_names.keys() :
                glevel = NumberFromName ( mname, 'gs' )
                if glevel == None : glevel = 0.0

                try :
                    if gs_files[glevel].count ( mname ) == 0 :
                        gs_files[glevel].append ( mname )
                except :
                    gs_files[glevel] = [mname]

            glevels = gs_files.keys()
            glevels.sort()
            
            for glevel in glevels :
                names = gs_files[glevel]

                sgs_names = []
                for mname in names :
                    sglevel = NumberFromName ( mname, 'sgs' )
                    if sglevel == None : sglevel = 1e99
                    sgs_names.append ( [sglevel, mname] )

                sgs_names.sort()
                sgs_names.reverse()
                for sgs, mname in sgs_names :
                    self.mbRegionsFileMenu.add_radiobutton (
                        label=mname, variable=self.regions_file, command=self.RFileSelected )


        files = os.listdir ( path );
        names_in_path = []
        for f in files :
            if f.find ( bname ) < 0 or f.find('_regions') < 0 : continue
            if f.find ( ".txt" ) >= 0 : continue
            if f.find ( ".mrc" ) >= 0 : continue
            if open_names.has_key(f) == False :
                names_in_path.append ( f )

        if len ( names_in_path ) == 0 : return

        if len(open_names.keys() ) > 0 :
            self.mbRegionsFileMenu.add_separator()

        self.mbRegionsFileMenu.add_radiobutton ( label="In %s:" % path )
        self.mbRegionsFileMenu.add_separator()

        for f in names_in_path :
            self.mbRegionsFileMenu.add_radiobutton (
                label=f, variable=self.regions_file, command=self.RFileSelected )

        return

        rlist = {}
        sgi = 1000
        try : files = os.listdir ( path );
        except : return

        for f in files :
            if f.find('_regions') >= 0 :
                # print "-", f
                #sg = NumberFromName ( f, 'sg' )
                #if sg == None :
                sg = sgi
                sgi = sgi + 1
    
                # print sg, f
                rlist[sg] = f

        sgs = rlist.keys ()
        sgs.sort ()
        sgs.reverse ()

        for sg in sgs :
            f = rlist[sg]
            self.mbRegionsFileMenu.add_radiobutton (
                label=f, variable=self.regions_file,
                command=self.RFileSelected )

        if len ( sgs ) == 0 : self.regions_file.set ('')



    def RFileSelected ( self ) :

        rfile = self.regions_file.get()

        rmod = None
        for m in OML () :
            try : m.seg_map
            except : continue
            if m.name == rfile :
                m.display = True
                if m.adj_graph_mol : m.adj_graph_mol.display = True
                rmod = m
                self.ReportRegionCount(len(m.regions))

            else :
                m.display = False
                if m.adj_graph_mol : m.adj_graph_mol.display = False

        if rmod == None :
            mm = getMod ( self.dmap.get() )
            if mm == None : print self.dmap.get(), "not open"; return
            path = os.path.dirname(mm.data.path) + os.path.sep
            rmod = self.ReadRegions ( mm, path + rfile )

        # if rmod : self.nRegions.set ( "%d" % len(rmod.surfacePieces) )

        rmod.seg_map = getMod ( self.dmap.get() )
        if rmod.seg_map : print "Set seg map to", rmod.seg_map.name

        umsg ( "Showing %s - %d regions, %d surfaces" % (rfile, len(rmod.regions.keys()), len(rmod.surfacePieces)) )


    def StrucMenu ( self ) :

        self.strucMB.menu.delete ( 0, 'end' )   # Clear menu

        self.strucMB.menu.add_radiobutton ( label="Open structures:" )
        self.strucMB.menu.add_separator()

        slist = []
        open_mols = {}
        for m in OML() :
            if type(m) == chimera.Molecule and m.name.find("_graph") == -1:
                fn = NumberFromName ( m.name, 'f' )
                slist.append ( [fn, m.name] )
                open_mols[m.name] = True

        slist.sort ()
        cur_sel_found = False
        for fn, mname in slist :
            self.strucMB.menu.add_radiobutton ( label=mname, variable=self.struc,
                                                command = self.StrucSelected )

            if mname == self.struc.get() :
                cur_sel_found = True

        if cur_sel_found == False :
            self.struc.set ( "" )

        dmap_name = self.dmap.get()
        if len(dmap_name) == 0 : return
        
        dmap = getMod ( dmap_name )
        if dmap == None : return

        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        files = os.listdir ( path );
        mols_in_path = []
        for f in files :
            if f.find ( ".pdb" ) >= 0 and open_mols.has_key(f) == False :
                mols_in_path.append ( f )

        if len ( mols_in_path ) == 0 : return

        self.strucMB.menu.add_separator()
        self.strucMB.menu.add_radiobutton ( label="In %s:" % path )
        self.strucMB.menu.add_separator()

        for fm in mols_in_path :
            self.strucMB.menu.add_radiobutton ( label=fm, variable=self.struc,
                                                command = self.StrucSelected )


    def StrucSelected ( self ) :
        print "selected structure", self.struc.get()
        fname = self.struc.get()

        fmol = getMod ( fname )
        if fmol :
            print " - open"

            for om in chimera.openModels.list() :
                try : om_mol = om.mol
                except : continue
                if om_mol == fmol :
                    self.struc_map.set ( om.name )
                    break

            return

        dmap_name = self.dmap.get()
        if len(dmap_name) == 0 : print " - no map selected"; return        
        dmap = getMod ( dmap_name )
        if dmap == None : print " -", dmap_name, "no longer open"; return

        path = os.path.dirname(dmap.data.path) + os.path.sep
        
        fmol = chimera.openModels.open ( path + fname )[0]
        print " - opened"
        fmol.ch_colors = RandColorChains ( fmol )



    def StrucMapSelected ( self ) :

        print "struc map selected", self.struc_map.get()
        res = NumberFromName ( self.struc_map.get(), 'r' )
        if res != None :
            self.simRes.set ( "%.1f" % res )
            sp = NumberFromName ( self.struc_map.get(), 'sp' )
            if sp != None :
                self.simRes.set ( "%.0f" % (res,) )
                self.simGridSp.set ( "%.1f" % (sp,) )

        mm = getMod ( self.struc_map.get() )
        mm.display = True

        try : mm.mol
        except : return

        mm.mol.display = True

        self.struc.set ( mm.mol.name )

        for m in OML () :

            if type(m) == chimera.Molecule : continue

            if type(m) == VolumeViewer.volume.Volume and m != mm :
                m.display = False

            try : m.mol
            except : continue
            if m != mm :
                m.display = False
                m.mol.display = False




    def GetMapSelected ( self ) :

        print "getting selected map"

        try :
            s = chimera.selection.currentContents()[0][0]
        except :
            print "please select the surface of a map"
            self.dmap.set ( '' )
            return
        
        mname = s.model.name
        self.dmap.set ( mname )


    def CloseRegsExcl ( self ) :

        import _surface

        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not found"; return

        for m in chimera.openModels.list () :
            try : m.path
            except : continue
            if type(m) == _surface.SurfaceModel and m.name.find ("_regions") >= 0 :
                if m.name != smod.name :

                    map_name = m.name [ 0 : m.name.rfind ( '_D' ) ] + ".mrc"
                    if map_name == self.dmap.get() :
                        print "skipping %s and map %s" % (m.name, map_name)
                        continue

                    #print "closing and deleting:", m.path + m.name
                    try : os.remove ( m.path + m.name )
                    except : print "delete failed - ", m.path + m.name
                    cmname = m.name
                    chimera.openModels.close ( m )

                    umsg ( "Closed %s" % cmname )


                    try : os.remove ( m.path + map_name ); print " - deleted", map_name
                    except : print "- did not remove map:", m.path + map_name



    def Old_WriteRegionsFile ( self, fname = None ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not found"; return

        umsg ( "Writing %d regions" % len(smod.regions) )
        if fname : print " - to", fname

        consa = []
        for r1, cr1 in smod.cons.iteritems () :
            for r2, o in cr1.iteritems () :
                if r1.rid < r2.rid :
                    consa.append ( [r1, r2, o] )

        size = 0
        for reg in smod.regions.values() :
            size = size + 1 + 3 * len(reg.rpoints)

        size = 1 + size + 1 + 4 * len(consa) + len(smod.regions)

        # print " - size:", size

        e = numpy.zeros ( [size], numpy.float32 )
        e[0] = float ( len(smod.regions) )
        at = 1

        spi = 1
        for rid, region in smod.regions.iteritems() :

            #try : sp_cons = smod.cons[sp]
            #except : continue
            #if sp.display == False : continue

            e[at] = len(region.rpoints)
            at = at + 1

            for rp in region.rpoints :
                e[at], e[at+1], e[at+2] = rp[0], rp[1], rp[2]
                at = at + 3

            region.new_rid = spi
            spi = spi + 1

        e[at] = float ( len(consa) )
        at = at + 1
        print " - writing %d connections" % ( len(consa) )

        for r1, r2, o in consa :
            e[at+0] = r1.new_rid
            e[at+1] = r2.new_rid
            e[at+2] = o.N
            e[at+3] = o.D
            at = at + 4



        if fname != None :
            e.tofile ( fname )
            return

        #print " - at:", at

        bname = smod.name [ 0 : smod.name.rfind ('_regions') ]

        for i in range (1, 1000) :
            nname = bname + "_regions_save_%d" % i
            print " - to", smod.path + nname

            if os.path.exists ( smod.path + nname ) : continue

            smod.name = nname
            self.regions_file.set ( nname )

            e.tofile ( smod.path + nname )
            umsg ( "Wrote %s" % nname )
            break



    def RegionSize ( self, region ) :

        if len(region.cregs) == 0 :            

            rp = region
            nparents = 0
            while rp.preg != None :
                rp = rp.preg
                nparents = nparents + 1

            self.tot_write_regions = self.tot_write_regions + 1

            return 1 + len(region.rpoints)*3 + 1 + nparents

        else :
            size = 0
            for creg in region.cregs :
                size = size + self.RegionSize ( creg )
            return size


    def AddRegion ( self, smod, region, e, e_at ) :

        if len(region.cregs) == 0 :

            e[e_at] = float ( len(region.rpoints) )
            e_at = e_at + 1

            #npoints = len(region.rpoints)
            #e = numpy.concatenate ( [ e, numpy.array( [npoints], numpy.float32 ) ] )
            #lr = region.rpoints.reshape ( 1, len(region.rpoints)*3 )
            #e = numpy.concatenate ( [ e, lr[0] ] )

            for rp in region.rpoints :
                e[e_at] = rp[0]; e_at = e_at + 1
                e[e_at] = rp[1]; e_at = e_at + 1
                e[e_at] = rp[2]; e_at = e_at + 1

            rp = region
            parents = []
            while rp.preg != None :
                rp = rp.preg
                parents.append ( rp.rid )

            #parents = [ len(parents) ] + parents
            #e = numpy.concatenate ( [ e, numpy.array( parents, numpy.float32 ) ] )

            e[e_at] = float ( len(parents) )
            e_at = e_at + 1

            #if ( len(parents) > 0 ) : print "Region %d has %d parents: " % (region.rid, len(parents)),

            for pi in parents :
                e[e_at] = float ( pi )
                e_at = e_at + 1
                #print pi,

            #if ( len(parents) > 0 ) : print ""

        else :
            for creg in region.cregs :
                e_at = self.AddRegion ( smod, creg, e, e_at )

        return e_at


    def WriteRegionsFile ( self, fname = None ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not found"; return


        if fname is None:
            # Show save-file dialog.
            def save ( okay, dialog ):
                if okay:
                    paths = dialog.getPaths ( )
                    if paths:
                        self.WriteRegionsFile ( paths[0] )
                        
            bname = smod.name [ 0 : smod.name.rfind ('_regions') ]
            prefix = smod.path + bname + "_regions_save_%d"
            uupath = unusedFile ( prefix )
            from OpenSave import SaveModeless
            d = SaveModeless ( title = 'Save Regions',
                               initialdir = smod.path,
                               initialfile = os.path.basename(uupath),
                               command = save )
            return


        self.tot_write_regions = 0
        tot_e_size = 1
        for rid, region in enumerate ( smod.regions.values() ) :
            tot_e_size = tot_e_size + self.RegionSize ( region )


        num_cons = 0
        for r1, cr1 in smod.rcons.iteritems () :
            for r2, o in cr1.iteritems () :
                if r1.rid < r2.rid :
                    num_cons = num_cons + 1

        tot_e_size = tot_e_size + 1 + 4 * (num_cons)


        umsg ( "Writing %d regions, %d grouped" % ( self.tot_write_regions, len(smod.regions.values()) ) )
        if fname : print " - to", fname

        e = numpy.zeros ( [tot_e_size], numpy.float32 )

        e[0] = float ( self.tot_write_regions )
        e_at = 1

        rlist = renumberRegions(smod)
        for region in rlist :
            e_at = self.AddRegion ( smod, region, e, e_at )

        print " - writing %d connections" % ( num_cons )

        e[e_at] = float ( num_cons )
        e_at = e_at + 1

        #consa = []
        for r1, cr1 in smod.rcons.iteritems () :
            for r2, o in cr1.iteritems () :
                if r1.rid < r2.rid :
                    #consa = consa + [r1.rid, r2.rid, o.N, o.D]
                    e[e_at+0] = float ( r1.rid )
                    e[e_at+1] = float ( r2.rid )
                    e[e_at+2] = o.N
                    e[e_at+3] = o.D
                    e_at = e_at + 4

        #consa = [len(consa)/4] + consa
        #e = numpy.concatenate ( [ e, numpy.array ( consa, numpy.float32 ) ] )

        e.tofile ( fname )
        umsg ( "Wrote %s" % os.path.basename(fname) )
        
        #smod.name = os.path.basename ( fname )
        #self.regions_file.set ( smod.name )



    def SaveRegsToMRC ( self, regs, dmap, path = None ) :

        if path is None:
            # Show file chooser dialog.
            fprefix = os.path.splitext(dmap.name)[0]
            if len(regs) == 1 :
                fname = fprefix + "_region_%d.mrc" % regs[0].rid
            else :
                fname = fprefix + "_%d_regions.mrc" % len(regs)
            dir = os.path.dirname ( dmap.data.path )
            import OpenSave
            d = OpenSave.SaveModal ( title = "Save Masked Map",
                                     initialdir = dir, initialfile = fname,
                                     filters = [('MRC map', '*.mrc', '.mrc')] )
            paths_and_types = d.run ( self.toplevel_widget )
            if paths_and_types:
                path = paths_and_types[0][0]
            else:
                return

        li,lj,lk, hi,hj,hk = 1e99,1e99,1e99, -1e99,-1e99,-1e99
        for reg in regs :
            for i,j,k in reg.rpoints :
                if li > i : li = i
                if lj > j : lj = j
                if lk > k : lk = k
                if hi < i : hi = i
                if hj < j : hj = j
                if hk < k : hk = k

        bound = 2
        li = li - bound; lj = lj - bound; lk = lk - bound
        hi = hi + bound; hj = hj + bound; hk = hk + bound

        n1 = hi - li + 1
        n2 = hj - lj + 1
        n3 = hk - lk + 1

        print "Bounds - %d %d %d --> %d %d %d --> %d %d %d" % ( li, lj, lk, hi, hj, hk, n1,n2,n3 )

        umsg ( "Saving %d regions to mrc file..." % len(regs) )

        nmat = numpy.zeros ( (n3,n2,n1), numpy.float32 )
        dmat = dmap.full_matrix()

        #regs_name = ""
        for reg in regs :
            #regs_name = regs_name + "_%d" % reg.rid
            for i,j,k in reg.rpoints :
                nmat[k-lk,j-lj,i-li] = dmat[k,j,i]
                #for ii in [i-1,i,i+1] :
                #    for jj in [j-1,j,j+1] :
                #        for kk in [k-1,k,k+1] :
                #            nmat[kk-lk,jj-lj,ii-li] = dmat[kk,jj,ii]
                            

        O = dmap.data.origin
        print "origin:", O
        nO = ( O[0] + float(li) * dmap.data.step[0],
               O[1] + float(lj) * dmap.data.step[1],
               O[2] + float(lk) * dmap.data.step[2] )
        
        print "new origin:", nO

        ndata = VolumeData.Array_Grid_Data ( nmat, nO, dmap.data.step, dmap.data.cell_angles )
        try : nv = VolumeViewer.volume.add_data_set ( ndata, None )
        except : nv = VolumeViewer.volume.volume_from_grid_data ( ndata )

        nv.name = os.path.basename ( path )

        nv.openState.xform = dmap.openState.xform

        nv.write_file ( path, "mrc" )

        umsg ( "Wrote %s" % nv.name )



    def WriteSelRegionsMRCFile ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a map from which density will be taken" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        if len ( self.regions_file.get() ) == 0 : umsg ("Please select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        surfs = SelectedRegions ( smod )
        if len(surfs)==0 :
            umsg ( "Please select one ore more regions to save to .mrc file" )
            return

        regs = [ sp.region for sp in surfs ]

        self.SaveRegsToMRC ( regs, dmap )


    def WriteAllRegionsMRCFile ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a map from which density will be taken" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        if len ( self.regions_file.get() ) == 0 : umsg ("Please select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        regs = [ sp.region for sp in smod.surfacePieces ]

        self.SaveRegsToMRC ( regs, dmap )

        


    def WriteEachRegionMRCFile ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a map from which density will be taken" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        if len ( self.regions_file.get() ) == 0 : umsg ("Please select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        surfs = SelectedRegions ( smod )
        if len(surfs)==0 :
            surfs = smod.surfacePieces
        regs = [sp.region for sp in surfs]

        # Choose file path.
        dir = os.path.dirname ( dmap.data.path )
        fprefix = os.path.splitext ( dmap.name ) [0]
        fname = fprefix + "_region_%d.mrc"
        import OpenSave
        d = OpenSave.SaveModal ( title = "Save Masked Map",
                                 initialdir = dir, initialfile = fname,
                                 filters = [('MRC map', '*.mrc', '.mrc')] )
        paths_and_types = d.run ( self.toplevel_widget )
        if paths_and_types:
            path = paths_and_types[0][0]
        else:
            return
        if not '%d' in path:
            umsg ( "Must include '%d' in map file name for region number" )
            return

        print "Saving each of %d regions to .mrc files" % len(regs)

        for reg in regs :
            self.SaveRegsToMRC ( [reg], dmap, path % (reg.rid,) )



    def MaskMapWRegions ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a map from which density will be taken" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        if len ( self.regions_file.get() ) == 0 : umsg ("Please select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        surfs = SelectedRegions ( smod )
        if len(surfs)==0 :
            umsg ( "Please select one ore more regions to register the structure with" )
            return

        regs = []
        for sp in surfs : regs.append ( sp.region )

        dmat = dmap.full_matrix()
        nmat = numpy.zeros_like ( dmat )

        for reg in regs :
            for i,j,k in reg.rpoints :
                nmat[k,j,i] = dmat[k,j,i]
                            

        ndata = VolumeData.Array_Grid_Data ( nmat, dmap.data.origin, dmap.data.step, dmap.data.cell_angles )
        try : nv = VolumeViewer.volume.add_data_set ( ndata, None )
        except : nv = VolumeViewer.volume.volume_from_grid_data ( ndata )
        nv.name = os.path.splitext ( dmap.name )[0] + "_masked.mrc"
        nv.openState.xform = dmap.openState.xform
        
        nv.surface_levels = [ dmap.surface_levels[0] ]
        nv.update_surface ( False, VolumeViewer.volume.Rendering_Options() )
        for sp in nv.surfacePieces :
            v, t = sp.geometry
            if len(v) == 8 and len(t) == 12 : sp.display = False

        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        nv.write_file ( path + nv.name, "mrc" )
        umsg ( "Wrote %s" % nv.name )

        self.dmap.set ( nv.name )

        return nv

        # VolumeViewer.volume.volume_manager


    def MakeRegion ( self, smod, rpoints, rid ) :

        reg = aClass ()
        reg.color = ( 0.5*(1+rand()), 0.5*(1+rand()), 0.5*(1+rand()), 1.0 )
        reg.cregs = []
        reg.preg = None
        reg.placed = False
        reg.rid = rid

        if rpoints != None :

            reg.rpoints = rpoints
            reg.volume = float(len(rpoints)) * smod.seg_map.data.step[0] * smod.seg_map.data.step[1] * smod.seg_map.data.step[2]

            tpoints = numpy.copy ( reg.rpoints )
            _contour.affine_transform_vertices ( tpoints, smod.seg_map.data.ijk_to_xyz_transform )
            com = numpy.sum(tpoints, axis=0) / len(tpoints)
            reg.COM = chimera.Vector ( com[0], com[1], com[2] )
            comv = numpy.ones_like ( tpoints ) * com
            points = tpoints - comv
            reg.bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points), 1 ) ) )

            i,j,k = reg.rpoints[0][0], reg.rpoints[0][1], reg.rpoints[0][2]
            reg.map_ind = [ i,j,k ]
            reg.map_inds = [ [i,j,k] ]

        return reg



    def ReadRegions ( self, mm, regions_file_path, createMod=True ) :

        regions_file = os.path.basename ( regions_file_path )

        print "Looking for regions surfaces", regions_file
        smod = getMod ( regions_file )

        try : e = numpy.fromfile ( regions_file_path, numpy.float32 )
        except : return None

        if smod == None :
            smod = _surface.SurfaceModel()
            smod.name = regions_file
            smod.oneTransparentLayer = True
            chimera.openModels.add( [smod] )
            smod.openState.xform = mm.openState.xform
        else :
            print " - found", smod.name
            for p in smod.surfacePieces : smod.removePiece ( p )
            try : smod.rid_to_sp_map.clear()
            except : pass
            try : smod.regions.clear()
            except : pass

        mm.regions_file = regions_file
        smod.path = os.path.dirname ( regions_file_path ) + os.path.sep

        smod.adj_graph_mol = None
        smod.seg_map = mm
        print " - " + smod.path + smod.name

        nregions = int ( e[0] )
        print " - reading %d regions..." % nregions
        at = 1

        smod.regions = {}
        all_regions = {}

        for i in range ( nregions ) :

            try : nvoxels = int ( e[at] )
            except :
                print " - reached end of file before reading all regions"
                break

            at = at + 1
            rvs = e [ at : at + (nvoxels*3) ]
            at = at + nvoxels*3

            rpoints = numpy.reshape ( rvs, (nvoxels, 3) )

            nparents = int ( e[at] )
            at = at + 1
            parents = []
            for ni in range ( nparents ) :
                parents.append ( int( e[at] ) )
                at = at + 1

            if nvoxels < int ( self.minRegionSize.get() ) :
                continue

            reg = self.MakeRegion ( smod, rpoints, i+1 )

            all_regions[reg.rid] = reg
            smod.regions [ reg.rid ] = reg

            last_reg = reg
            reg.preg = None

            # if ( len(parents) > 0 ) : print "Region %d has %d parents:" % (reg.rid, len(parents)),

            for pi in parents :
                #print pi,
                if pi in all_regions:
                    preg = all_regions[pi]
                else :
                    preg = self.MakeRegion ( smod, None, pi )
                    all_regions[pi] = preg

                last_reg.preg = preg

                if preg.cregs.count ( last_reg ) == 0 :
                    preg.cregs.append ( last_reg )

                last_reg = preg




        try : ncon = int ( e[at] )
        except :
            print " - reached end of file before reading connections"
            ncon = 0

        at = at + 1
        smod.rcons = {}

        if ncon > 0 :

            print " - reading %d connections..." % ( ncon )
        
            am = e [ at : at + (ncon*4) ]
            cm = numpy.reshape ( am, (ncon, 4) )

            for i in range ( ncon ) :
                c = cm[i]
                o = aClass()
                try :
                    rid1, rid2, o.N, o.D = int(c[0]), int(c[1]), c[2], c[3]
                    # print "&", c
                except :
                    print "*", c

                #try : r1 = smod.regions[rid1]
                #except : print "Bad region id", rid1; continue
                #try : r2 = smod.regions[rid2]
                #except : print "Bad region id", rid2; continue

                try : r1 = smod.regions[rid1]
                except : continue
                try : r2 = smod.regions[rid2]
                except : continue

                if r1 == r2 : print " - connection to self"

                if smod.rcons.has_key (r1) == False : smod.rcons[r1] = {}
                if smod.rcons.has_key (r2) == False : smod.rcons[r2] = {}

                smod.rcons[r1][r2] = o
                smod.rcons[r2][r1] = o

        print ""


        smod.regions = {}

        for rid, reg in all_regions.iteritems() :
            if reg.preg is None:
                smod.regions [ rid ] = reg

        print " - making %d regions..." % len( smod.regions.values() )
        for ri, reg in enumerate ( smod.regions.values() ) :
            #print "Making region %d/%d" % ( ri, len(smod.regions.values()) )
            self.MakeAllRegions ( smod, reg )


        ComputeConnections ( smod )

        # mm.display = False
        self.regions_file.set ( regions_file )
        if createMod : self.RegsDispUpdate ()


        return smod


    def MakeAllRegions ( self, smod, reg, lev=0 ) :

        #for i in range(lev) : print " ",

        #print reg.rid, smod.name

        for creg in reg.cregs :
            self.MakeAllRegions ( smod, creg, lev+1 )

        if len(reg.cregs) > 0 :
            MakeRegionFromCRegs ( smod, reg ) 

        


    def RegsDispUpdate ( self ) :

        if len(self.regions_file.get()) == 0 : umsg ( "No regions file selected" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s not open", self.regions_file.get() ); return

        try : smod.seg_map.name
        except : smod.seg_map = getMod ( self.dmap.get() )

        print "%s - updating %s, %d regions" % (
            smod.name, self.regsVisMode.get(), len(smod.regions) )

        for p in smod.surfacePieces : smod.removePiece ( p )

        smod.rid_sp_map = {}

        dmap = None
        init_mat = None

        maxnr = self.MaximumRegionsToDisplay()
        for reg in smod.regions.values() :

            if maxnr > 0 and len(smod.surfacePieces) >= maxnr  :
                umsg('Only showing %d of %d regions.' %
                     (len(smod.surfacePieces), len(smod.regions)))
                break
            if len(reg.rpoints) == 0 : print "region %d has 0 voxels!" % reg.rid; return
            if len(reg.rpoints) < MIN_REG_SIZE : break

            vt = None

            if self.regsVisMode.get() == 'Voxel_Surfaces' :
                vt = MultiScale.surface.surface_points ( reg.rpoints, 1.0, 0.1, .25, 5 )
                _contour.affine_transform_vertices ( vt[0], smod.seg_map.data.ijk_to_xyz_transform )

            elif self.regsVisMode.get() == 'Density_Maxima' :
                rpts = numpy.array ( reg.map_inds, numpy.float32 )
                #rpts = numpy.array ( [reg.map_ind], numpy.float32 )
                vt = MultiScale.surface.surface_points ( rpts, 1.0, 0.1, .25, 5 )
                _contour.affine_transform_vertices ( vt[0], smod.seg_map.data.ijk_to_xyz_transform )

            elif self.regsVisMode.get() == 'Iso_Surfaces' :

                if dmap == None:
                    dmap = smod.seg_map.writable_copy ( require_copy=True )
                    init_mat = dmap.data.full_matrix().copy()

                tpoints = reg.rpoints.copy()
                _contour.affine_transform_vertices ( tpoints, dmap.data.ijk_to_xyz_transform )

                vt = None
                print " - reg %d, %d voxels" % (reg.rid, len(tpoints))

                sg = VolumeData.zone_masked_grid_data( dmap.data, tpoints, dmap.data.step[0] )
                m = sg.full_matrix()

                cmatrix = dmap.data.full_matrix()
                cmatrix[:,:,:] = m[:,:,:]
                dmap.surface_levels = [ dmap.surface_levels[0] ]
                dmap.region = ( dmap.region[0], dmap.region[1], [1,1,1] )
                ro = VolumeViewer.volume.Rendering_Options()
                #ro.surface_smoothing = True
                #ro.smoothing_factor = .2
                #ro.smoothing_iterations = 5
                dmap.update_surface ( False, ro )

                surf_sp = None
                surf_sp0 = None
                for sp in dmap.surfacePieces :
                    v, t = sp.geometry
                    #print "- %d vertices, %d tris" % (len(v), len(t))
                    if len(v) == 8 and len(t) == 12 :
                        continue
                    if len(v) == 0 and len(t) == 0 :
                        surf_sp0 = sp
                    else :
                        surf_sp = sp
 
                try : vt = surf_sp.geometry
                except : vt = surf_sp0.geometry

                cmatrix[:,:,:] = init_mat[:,:,:]
                dmap.data.values_changed()

            nsp = makeRegionSurface ( smod, vt[0], vt[1], reg )


        if dmap : dmap.close()

        umsg ( "%s: %d regions, %d surfaces" % (
            smod.name, len(smod.regions), len(smod.surfacePieces) ) )

        self.ReportRegionCount(len(smod.regions))

    def MaximumRegionsToDisplay ( self ) :

        try:
            maxnr = int(self.maxNumRegions.get())
        except:
            maxnr = 0
        return maxnr

    def ReportRegionCount ( self, c ):

        if c is None:
            s = ''
        else:
            s = "%d regions" %  c
        self.regionCount["text"] = s


    def RegsDispThr0 ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None :
            print self.regions_file.get(), "not open"
            return

        print "%s - thresholding %d regions" % ( smod.name, len(smod.regions) )

        for p in smod.surfacePieces : smod.removePiece ( p )

        smod.rid_sp_map = {}

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print "Map %s not open" % self.dmap.get(); return
        print "Using map:", dmap.name


        maxnr = self.MaximumRegionsToDisplay()
        for rid, reg in smod.regions.iteritems() :

            if maxnr > 0 and len(smod.surfacePieces) >= maxnr  :
                umsg('Only showing %d of %d regions.' %
                     (len(smod.surfacePieces), len(smod.regions)))
                break
            if len(reg.rpoints) == 0 : print "region %d has 0 voxels!" % rid; return
            if len(reg.rpoints) < MIN_REG_SIZE : break

            tpoints = reg.rpoints.copy()
            _contour.affine_transform_vertices ( tpoints, dmap.data.ijk_to_xyz_transform )

            vt = None
            print " - reg %d, %d voxels" % (rid, len(tpoints))

            sg = VolumeData.zone_masked_grid_data( dmap.data, tpoints, 0.1 )
            m = sg.full_matrix()

            cmatrix = dmap.data.full_matrix()
            cmatrix[:,:,:] = m[:,:,:]
            dmap.surface_levels = [ dmap.surface_levels[0] ]
            dmap.region = ( dmap.region[0], dmap.region[1], [1,1,1] )
            dmap.update_surface ( False, VolumeViewer.volume.Rendering_Options () )

            surf_sp = None
            surf_sp0 = None
            for sp in dmap.surfacePieces :
                v, t = sp.geometry
                #print "- %d vertices, %d tris" % (len(v), len(t))
                if len(v) == 8 and len(t) == 12 :
                    continue
                if len(v) == 0 and len(t) == 0 :
                    surf_sp0 = sp
                else :
                    surf_sp = sp

            try : vt = surf_sp.geometry
            except : vt = surf_sp0.geometry

            nsp = makeRegionSurface ( smod, vt[0], vt[1], reg )

            cmatrix[:,:,:] = init_mat[:,:,:]
            dmap.data.values_changed()



    def RegsDispThr ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None :
            print self.regions_file.get(), "not open"
            return

        print "%s - thresholding %d regions" % ( smod.name, len(smod.regions) )

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print "Map %s not open" % self.dmap.get(); return
        print " - using map:", dmap.name
        dthr = dmap.surface_levels[0]

        for p in smod.surfacePieces : smod.removePiece ( p )

        smod.rid_sp_map = {}

        maxnr = self.MaximumRegionsToDisplay()
        for rid, reg in smod.regions.iteritems() :

            #print "%d -" % rid,

            if maxnr > 0 and len(smod.surfacePieces) >= maxnr  :
                umsg('Only showing %d of %d regions.' %
                     (len(smod.surfacePieces), len(smod.regions)))
                break
            if len(reg.rpoints) == 0 : print "region %d has 0 voxels!" % rid; return
            if len(reg.rpoints) < MIN_REG_SIZE : break

            tpoints = reg.rpoints.copy()
            _contour.affine_transform_vertices ( tpoints, dmap.data.ijk_to_xyz_transform )

            dvals = dmap.interpolated_values ( tpoints, dmap.openState.xform )
            min_d = numpy.min ( dvals )
            #print "%d points, min d: %.4f," % (len(tpoints), min_d),

            ge = numpy.greater_equal ( dvals, dthr )
            tpoints = numpy.compress ( ge, tpoints, 0 )
            dvals = numpy.compress ( ge, dvals )
            min_d = numpy.min ( dvals )

            #print "/ %d points, min d: %.4f" % (len(tpoints), min_d)        

            _contour.affine_transform_vertices ( tpoints, dmap.data.xyz_to_ijk_transform )
            vt = MultiScale.surface.surface_points ( tpoints, 1.0, 0.1, .25, 5 )
            _contour.affine_transform_vertices ( vt[0], dmap.data.ijk_to_xyz_transform )

            nsp = makeRegionSurface ( smod, vt[0], vt[1], reg )


    def RegsPrint ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None :
            print self.regions_file.get(), "not open"
            return

        for rid, reg in smod.regions.iteritems() :

            print "%d - %d %d %d" % ( rid, reg.map_ind[0], reg.map_ind[1], reg.map_ind[2] )


    def ReduceMap ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a map first" ); return
        mm = getMod ( self.dmap.get() )
        if mm == None : umsg ( "%s is not open" % self.dmap.get() ); return

        path = os.path.dirname ( mm.data.path ) + os.path.sep
        mname = os.path.splitext ( mm.name )[0]

        d = mm.data
        m2 = d.matrix ( ijk_step=(2,2,2) )
        step2 = ( d.step[0]*2.0, d.step[1]*2.0, d.step[2]*2.0 )
        ld = VolumeData.Array_Grid_Data(m2, d.origin, step2, d.cell_angles, d.rotation,
                           name = mname + '_s2.mrc')

        gv = VolumeViewer.volume.add_data_set ( ld, None )
        gv.name = mname + '_s2.mrc'

        print "writing", path + gv.name
        mod.write_file ( path + gv.name, "mrc" )




    def RunSeg ( self, dmap, file_out, thrD ) :
        
        # thrD = dmap.surface_levels[0]

        import Segger
        prog_path = os.path.join ( Segger.__path__[0], 'seg' )
        args = [ prog_path, dmap.data.path, "%f" % thrD ]

        print "- running: "
        for a in args : print a,
        print ""

        t0 = time.clock()

        try : subprocess.call ( args )
        except : tseg ( dmap, file_out )

        took_sec = time.clock() - t0
        print "Time: %f sec" % took_sec



    def Segment ( self, createMod = True ) :

        if len(self.dmap.get()) == 0 : umsg ("Select an open density map to segment first" ); return
        mm = getMod ( self.dmap.get() )
        if mm == None : umsg ( "%s is not open" % self.dmap.get() ); return

        thrD = mm.surface_levels[0]
        path = os.path.dirname ( mm.data.path ) + os.path.sep
        base = os.path.splitext ( mm.name )[0]


        rfilename = "%s_%s%.3f_regions" % (base, "D", thrD)
        regions_file = path + rfilename

        self.regions_file.set ( rfilename )

        print "\n___________________________"
        umsg ( "Segmenting %s, density threshold %f" % (mm.name, thrD) )

        self.t0 = time.clock()

        self.RunSeg ( mm, regions_file, thrD )
        sregs = self.ReadRegions ( mm, regions_file, createMod )

        if sregs :
            #self.nRegions.set ( "%d" % len(sregs.surfacePieces) )
            for m in chimera.openModels.list() :
                if m != sregs and m.name.find ( '_regions' ) >= 0 :
                    m.display = False

            mm.display = False

            self.SmoothAndGroup()

        else :
            umsg ( "Segmentation not completed - make sure the selected map is saved as .mrc to disk." )
            # return


    def Fit ( self ) :


        #("Selected region", self.FitToSelRegion), 'selected_region'
        #("Each selected region", self.FitToSelectedRegions), 'each_selected_region'
        #("Combined selected regions", self.FitMapToSelRGroup), 'combined_selected_regions'
        #("Regions around selected", self.FitMapToRegionsAroundSel), 'around_selected'
        #("Groups of regions", self.FitMapToRGroups), 'all_groups'

        alignTo = self.alignTo.get()

        if alignTo == 'combined_selected_regions' :
            print "Fitting to combined selected regions..."
            self.FitMapToSelRGroup ()

        elif alignTo == 'each_selected_region' :
            print "Fitting to each selected region..."
            self.FitToSelectedRegions ()

        elif alignTo == 'around_selected' :
            print "Fitting groups around selected..."
            self.FitMapToRegionsAroundSel ()

        elif alignTo == 'all_groups' :
            print "Fitting to all groups..."
            self.FitMapToRGroups ()




    def SegmentSeq ( self, createMod = True ) :

        if len(self.dmap.get()) == 0 : umsg ("Select an open density map to segment first" ); return
        mm = getMod ( self.dmap.get() )
        if mm == None : umsg ( "%s is not open" % self.dmap.get() ); return

        thrD_min = mm.surface_levels[0]
        path = os.path.dirname ( mm.data.path ) + os.path.sep
        base = os.path.splitext ( mm.name )[0]

        last_sregs = None

        nsteps = 300
        for i in range ( 1, nsteps+1 ) :

            print 

            f = float(i)/float(nsteps)
            dist = mm.data.mrc_data.max_intensity - thrD_min

            thrD = mm.data.mrc_data.max_intensity - f * dist

            rfilename = "%s_%s%f_regions" % (base, "D", thrD)
            regions_file = path + rfilename

            self.regions_file.set ( rfilename )

            print "\n___________________________"
            umsg ( "Segmenting %s, density threshold %f" % (mm.name, thrD) )

            self.RunSeg ( mm, regions_file, thrD )
            sregs = self.ReadRegions ( mm, regions_file, createMod )

            if sregs :
                #self.nRegions.set ( "%d" % len(sregs.surfacePieces) )
                for m in chimera.openModels.list() :
                    if m != sregs and m.name.find ( '_regions' ) >= 0 :
                        m.display = False

                if last_sregs :
                    #self.RecolorRegs ( sregs, last_sregs )
                    last_sregs.display = False

                    for rid, creg in last_sregs.regions.iteritems() :
                        reg = sregs.regions [ rid ]
                        sregs.rid_sp_map[reg.rid].color = last_sregs.rid_sp_map[rid].color
                        

                mdlg.Step ( 1 )

                if last_sregs :
                    chimera.openModels.close ( [last_sregs] )
                last_sregs = sregs

            else :
                umsg ( "\n*** Segmentation not completed or regions file could not be opened\n\n" )
                # return


            mm.display = False




        

    def RecolorRegs ( self, regs, regs_c ) :

        print "Making index map for %d regions in %s" % ( len(regs.regions), regs.name )

        for si, sreg in enumerate ( regs.regions.values() ) :

            # status ( "Making index map %d of %d" % (si, len(srmod.regions)) )

            sreg.imap = {}
            for i, j, k in sreg.rpoints :
                try : mi = sreg.imap[i]
                except : mi = {}; sreg.imap[i] = mi
                try : mij = mi[j]
                except : mij = {}; mi[j] = mij
                mij[k] = 1


        print "Recoloring regions in %s by overlap..." % regs.name

        gregs = {}
        
        for reg_i, sp in enumerate ( regs_c.surfacePieces ) :

            region = sp.region
            # print "%d/%d" % ( reg_i, len(regs_c.surfacePieces) )

            max_overlap_num = 0
            max_overlap_ssp = None

            ipoints = numpy.cast[numpy.int] ( region.rpoints )

            for sreg_i, sp2 in enumerate ( regs.surfacePieces ) :

                sreg = sp2.region
                v = (sreg.COM - region.COM).length
                if v > 2.0 + sreg.bRad + region.bRad :
                    continue

                noverlap = 0
                for i,j,k in ipoints :
                    try : noverlap = noverlap + sreg.imap[i][j][k]
                    except : continue

                if noverlap > max_overlap_num :
                    max_overlap_num = noverlap
                    max_overlap_ssp = sp2

            if max_overlap_num > 0 :
                max_overlap_ssp.color = sp.color






    def Smooth ( self ) :
        self.SmoothInSteps ( False )


    def SmoothInSteps ( self, segWhileSmoothing ) :

        if len(self.dmap.get()) == 0 : umsg ("Select a density map in the Segger dialog first" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "Selected density map %s not open" % self.dmap.get() ); return

        try : targNRegs = int(self.targNRegions.get())
        except : umsg ( "Enter an integer for target # of regions" ); return

        if targNRegs <= 0 :
            umsg ( "Enter an integer > 0 for target # of regions" )
            return

        try :
            numit = int ( self.numSteps.get() )
            step_size = sdev = float ( self.stepSize.get() )
        except :
            umsg ( "Enter <int> for # smoothing steps, ,<float> for step size" )
            return


        umsg ( "%d smoothing steps of %.4f, aiming for %d regions" % (
            numit, step_size, targNRegs ) )


        try : dmap = dmap.orig_map
        except : pass

        mname = os.path.splitext ( dmap.name ) [0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        try :
            dmap.smoothed_map.name
            prev_map = True
        except :
            prev_map = False


        gs_s_level = None
        if prev_map :
            print "Prev smoothed map:", dmap.smoothed_map.name
            gs_s_level = NumberFromName ( dmap.smoothed_map.name, 'gs' )
            if gs_s_level == None :
                print "- did not get smoothing level"
            else :
                print "%s --- %.1f" % (dmap.smoothed_map.name, gs_s_level)
        else :
            dmap.smoothed_map = dmap.writable_copy (require_copy=True)
            dmap.smoothed_map.name = mname + "_smoothed"
            dmap.smoothed_map.orig_map = dmap
            gs_s_level = 0

        sm_m = dmap.smoothed_map
        print "- smoothed map:", sm_m.name
        sm_mat = sm_m.data.full_matrix()


        dmap.display = False
        sm_m.display = True # not segWhileSmoothing
        smod = getMod ( self.regions_file.get() )


        for iti in range ( numit ) :

            print "\nStep %d, step size %.1f" % (iti+1, sdev)

            gs_s_level = gs_s_level + sdev
            sm_name = mname + "_gs%.0f.mrc" % gs_s_level

            try :
                import GaussianFilter
                use_this = True
            except:
                use_this = False

            if use_this :
                v = sm_m
                tf = v.matrix_indices_to_xyz_transform(False, False)
                origin = [tf[a][3] for a in range(3)]
                step = [Matrix.length([tf[a][b] for a in range(3)]) for b in range(3)]
                ijk_linewidths = [sdev*2.0 / s for s in step]
                sm = GaussianFilter.gaussian_convolution(sm_mat, ijk_linewidths)

            else :
                import VolumeFilter
                region = sm_m.subregion(1, None)
                origin, step = sm_m.region_origin_and_step ( region )
                ijk_sdev = [sdev / s for s in step]
                sm = VolumeFilter.gaussian.gaussian_convolution (sm_mat, ijk_sdev)

            sm_mat[:,:,:] = sm[:,:,:]
            sm_m.name = sm_name
            sm_m.write_file ( path + sm_m.name )
            print " - wrote", path + sm_m.name

            if smod :
                min_d = 1e99
                for rid, reg in smod.regions.iteritems() :
                    if len(reg.rpoints) < 20 : continue
                    tpoints = reg.rpoints.copy()
                    _contour.affine_transform_vertices ( tpoints, sm_m.data.ijk_to_xyz_transform )
                    dvals = sm_m.interpolated_values ( tpoints, sm_m.openState.xform )
                    rmin_d = numpy.min ( dvals )
                    if rmin_d < min_d : min_d = rmin_d

                if min_d < 0.001 : min_d = 0.001
                print " - regions threshold: %f" % min_d
                sm_m.surface_levels[0] = min_d
                smod.display = False

            sm_m.display = True
            sm_m.matrix_changed ()
            sm_m.update_surface ( False, VolumeViewer.volume.Rendering_Options() )
            self.dmap.set ( sm_name )
            for sp in sm_m.surfacePieces :
                if len(sp.geometry[0]) == 8 and len(sp.geometry[1]) == 12 : sp.display = False

            umsg ( "Smoothed %s -> %s" % (dmap.name, sm_name) )

            if segWhileSmoothing :

                self.Segment ( True )

                sm_m.display = False
                self.dmap.set ( dmap.name )

                sgroups_f = self.regions_file.get()
                print "New regions file:", sgroups_f,

                smod = getMod ( sgroups_f )
                if smod == None :
                    umsg ( "Could not find segmentation for smoothed map %s" % sm_m.name )
                    break

                num_regs = len(smod.regions)
                print " - %d regions" % num_regs

                #self.GroupByOverlap ( regs, smod )
                #self.regions_file.set ( regs.name )
                #self.dmap.set ( dmap.name )
                #self.RegsDispUpdate ()
                #chimera.openModels.close ( [sregs] )
                #rfiles.insert ( 0, sgroups_f )

                if num_regs <= int(self.targNRegions.get()) :
                    print "\nReached target number of regions (%d)!" % int(self.targNRegions.get())
                    break


        # sm_m.close ()
        # self.dmap.set ( dmap.name )


    def DoSharpen ( self, path, dmap, s_smod, smod, sharp_rfile ) :

        import Segger
        prog_path = os.path.join ( Segger.__path__[0], 'sharpen' )
        args = [ prog_path, path+smod.name, path+s_smod.name, path+sharp_rfile ]
        #print "- running: "
        #for a in args : print a,
        #print ""
        t0 = time.clock()

        doSh = False
        try : subprocess.call ( args )
        except : doSh = True
        took_sec = time.clock() - t0
        #print "Time: %f sec" % took_sec

        if doSh :
            self.GroupByOverlap ( s_smod, smod )
            # self.WriteRegions ( path+sharp_rfile )
            smod.display = False
            s_smod.display = True
            return s_smod

        else :
            sh_smod = self.ReadRegions ( dmap, path+sharp_rfile )
            if sh_smod == None :
                umsg ( "Did not find sharpened regions file" )
                return

            s_smod.display = False
            smod.display = False
            sh_smod.display = True
            return sh_smod



    def Sharpen ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Select a density map in the Segger dialog first" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "Selected density map %s not open" % self.dmap.get() ); return

        mname = os.path.splitext ( dmap.name ) [0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        if len(self.regions_file.get()) == 0 : umsg ( "No regions file selected" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s not open", self.regions_file.get() ); return


        print "Sharpening %s (%d regions) in %s" % (
            smod.name, len(smod.surfacePieces), dmap.name)

        gs_s_level = NumberFromName ( smod.name, 'gs' )
        if gs_s_level == None :
            print "Regions file should be from a smoothed map"
            return

        base = smod.name [ 0 : smod.name.find ( '_gs' ) ]
        root_mm = None
        gsl_max = NumberFromName ( smod.name, 'gs' )
        print "Base:", base, gsl_max

        gsl_smods = []

        for m in chimera.openModels.list() :

            if m.name.find ( '_regions' ) < 0 or m.name.find ( base ) < 0 :
                continue

            gs_level = NumberFromName ( m.name, 'gs' )
            dThr = NumberFromName ( m.name, 'D' )
            sgs_level = NumberFromName ( m.name, 'sgs' )

            if sgs_level != None : continue

            if gs_level == None :
                gs_level = 0.0

            if gs_level <= gsl_max :
                gsl_smods.append ( [gs_level, dThr, m] )
                print " - added:", m.name, gs_level, dThr


        print ""

        base = smod.name [ 0 : smod.name.find ( '_D' ) ]
        print "Base:", base
        shf_name = ""

        hfile = smod.name [ 0 : smod.name.find ( '_D' ) ] + "_hfile.txt"
        print "HFile:", hfile

        print ""
        umsg ( "Sharpening __ %s __" % smod.name )

        import Segger
        prog_path = os.path.join ( Segger.__path__[0], 'sharpen' )
        args = [ prog_path ]

        gsl_smods.sort()
        gsl_smods.reverse()

        for gsl, dThr, s_smod in gsl_smods :
            shf_name = base + "_sgs%.0f_D%.2f_regions" % (gsl, dThr)
            print s_smod.name
            print " ->", shf_name
            args.append ( path+s_smod.name )

        print ""
        print "- running: "
        for a in args : print a,
        print ""
        t0 = time.clock()

        try : subprocess.call ( args )
        except : "Failed to run sharpen"
        took_sec = time.clock() - t0
        print "Time: %f sec" % took_sec

        print ""

        sh_smod = self.ReadRegions ( dmap, path+shf_name )
        if sh_smod == None :
            umsg ( "Did not find sharpened regions file" )
            return

        smod.display = False
        sh_smod.display = True




            


    def GroupByOverlap ( self, rmod, srmod ) :

        print "Making index map for %d regions in %s" % (
            len(srmod.regions), srmod.name )

        for si, sreg in enumerate ( srmod.regions.values() ) :

            status ( "Making index map %d of %d" % (si, len(srmod.regions)) )

            sreg.imap = {}
            for i, j, k in sreg.rpoints :
                try : mi = sreg.imap[i]
                except : mi = {}; sreg.imap[i] = mi
                try : mij = mi[j]
                except : mij = {}; mi[j] = mij
                mij[k] = 1


        print "Grouping regions in %s by overlap..." % rmod.name

        gregs = {}
        
        for reg_i, region in enumerate ( rmod.regions.values() ) :

            status ( "Testing region %d of %d" % (reg_i, len(rmod.regions)) )

            max_overlap_num = 0
            max_overlap_sreg = -1

            ipoints = numpy.cast[numpy.int] ( region.rpoints )

            for sreg_i, sreg in enumerate ( srmod.regions.values() ) :

                v = (sreg.COM - region.COM).length
                if v > 2.0 + sreg.bRad + region.bRad :
                    continue

                noverlap = 0
                for i,j,k in ipoints :
                    try : noverlap = noverlap + sreg.imap[i][j][k]
                    except : continue

                if noverlap > max_overlap_num :
                    max_overlap_num = noverlap
                    max_overlap_sreg = sreg_i

            if max_overlap_num > 0 :
                try : gregs [max_overlap_sreg].append ( region )
                except : gregs [max_overlap_sreg] = [ region ]


        print "Grouping regions..."
        ngroups = len ( gregs )
        ngi = 1
        for gi, jregs in gregs.iteritems() :

            print " - group %d - %d regions" % ( gi+1, len(jregs) )
            status ( "Grouping %d of %d" % (ngi, ngroups) )
            ngi = ngi + 1

            nreg = jregs[0]
            for jreg in jregs[1:] :
                nreg = JoinRegions ( rmod, [nreg, jreg] )

            nreg.sub_regs_ids = []
            nreg.sub_regs = []

        self.regions_file.set ( rmod.name )
        #self.RegsDispUpdate ()



    def Old_Smooth_and_Group ( self ) :

        dmap = getMod ( self.dmap.get() )
        print "Grouping regions by smoothing", dmap.name

        mname = os.path.splitext ( dmap.name ) [0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        smod = getMod ( self.regions_file.get() )
        if smod == None :
            print self.regions_file.get(), "not open"
            return

        print smod.name, " - %d regions" % len(smod.surfacePieces)


        targNRegs = int ( self.targNRegions.get() )

        numit = int ( self.numSteps.get() )
        step_size = sdev = float ( self.stepSize.get() )

        print "%d scale space steps of %.4f :" % (numit, step_size)
        print " - aiming for %d regions" % targNRegs

        try :
            dmap.smoothed_map.name
        except :
            dmap.smoothed_map = dmap.writable_copy (require_copy=True)
            dmap.smoothed_map.name = mname + "_smoothed"

        sm_m = dmap.smoothed_map
        print "- smoothed map:", sm_m.name
        m = sm_mat = sm_m.data.full_matrix()


        for iti in range ( numit ) :

            print "%d" % (iti+1),

            sm = laplacian ( sm_mat )
            numpy.multiply ( sm, step_size, sm )
            sm = sm + sm_mat
            sm_mat[:,:,:] = sm[:,:,:]

            print ">",

            rm = {}

            for rid, reg in smod.regions.iteritems() :

                while 1 :
                    i, j, k = reg.map_ind[0], reg.map_ind[1], reg.map_ind[2]
                    max_v = m[k,j,i]
                    max_inds = None
                    for ii in [i-1, i, i+1] :
                        for jj in [j-1, j, j+1] :
                            for kk in [k-1, k, k+1] :
                                ival = m[kk,jj,ii]
                                if ival > max_v :
                                    max_v = ival
                                    max_inds = [ii,jj,kk]

                    if max_inds == None :
                        # at local maximum
                        break
                    else :
                        i, j, k = max_inds[0], max_inds[1], max_inds[2]
                        reg.map_ind[0], reg.map_ind[1], reg.map_ind[2] = i,j,k
                        reg.map_inds.append ( [i,j,k] )

                ii, jj, kk = reg.map_ind[0], reg.map_ind[1], reg.map_ind[2]
                cur_val = m[kk,jj,ii]
                ind_stack = [ [ii,jj,kk] ]

                while len(ind_stack) > 0 :

                    ind_at = ind_stack.pop ( 0 )
                    i, j, k = ind_at[0], ind_at[1], ind_at[2]

                    try : rm[i][j][k].index(reg); continue
                    except : pass

                    try : rmi = rm[i]
                    except : rmi = {}; rm[i] = rmi
                    try : rmij = rmi[j]
                    except : rmij = {}; rmi[j] = rmij
                    try : rmijk = rmij[k]
                    except : rmijk = []; rmij[k] = rmijk
                    rmijk.append ( reg )

                    if m[ k+1, j, i ] >= cur_val : ind_stack.append ( [i, j, k+1] )                    
                    if m[ k-1, j, i ] >= cur_val : ind_stack.append ( [i, j, k-1] )                
                    if m[ k, j+1, i ] >= cur_val : ind_stack.append ( [i, j+1, k] )
                    if m[ k, j-1, i ] >= cur_val : ind_stack.append ( [i, j-1, k] )
                    if m[ k, j, i+1 ] >= cur_val : ind_stack.append ( [i+1, j, k] )
                    if m[ k, j, i-1 ] >= cur_val : ind_stack.append ( [i-1, j, k] )


            for i, rmi in rm.iteritems() :
                for j, rmij in rmi.iteritems() :
                    for k, regs in rmij.iteritems() :
                        if len(regs) >= 2 :
                            JoinRegions ( smod, regs )

            print "[%d]" % len(smod.regions),

            if len (smod.regions) <= targNRegs :
                print " - target!"
                break


            if iti > 0 and iti % 100 == 0 :

                sm_m.matrix_changed ()
                #ro = VolumeViewer.volume.Rendering_Options()
                # sm_m.surface_colors[0] = (.6, .6, .6, .4)
                #sm_m.update_surface ( False, ro )
                #for sp in sm_m.surfacePieces :
                #    if len(sp.geometry[0]) == 8 and len(sp.geometry[1]) == 12 : sp.display = False

                #self.RegsDispUpdate ()
                #mdlg.Step ( 1 )


        print ""

        self.RegsDispUpdate ()
        #mdlg.Step ( 10 )

        dmap.smoothed_map.close();
        dmap.smoothed_map = None        





    def Group_by_Tracking_Maxima ( self, smod, m, task = None ) :

        print "Grouping regions in %s..." % smod.name
        
        rm = {}
        for c, reg in enumerate ( smod.regions.values() ) :

            if c % 1000 == 0 :
                s = "Grouping region %d/%d" % ( c, len( smod.regions ) )
                status ( s )
                if task:
                    task.updateStatus ( s )

            #print "region %d/%d" % ( c+1, len( smod.regions.values() ) )

            while 1 :
                i, j, k = reg.map_ind[0], reg.map_ind[1], reg.map_ind[2]
                # print " - ", i,j,k
                max_v = m[k,j,i]
                max_inds = None
                for ii in [i-1, i, i+1] :
                    if ii >= 0 and ii < m.shape[2] :
                        for jj in [j-1, j, j+1] :
                            if jj >= 0 and jj < m.shape[1] :
                                for kk in [k-1, k, k+1] :
                                    if kk >= 0 and kk < m.shape[0] :
                                        ival = m[kk,jj,ii]
                                        if ival > max_v :
                                            max_v = ival
                                            max_inds = [ii,jj,kk]

                if max_inds == None :
                    # at local maximum
                    break
                else :
                    i, j, k = max_inds[0], max_inds[1], max_inds[2]
                    reg.map_ind[0], reg.map_ind[1], reg.map_ind[2] = i,j,k
                    reg.map_inds.append ( [i,j,k] )

            ii, jj, kk = reg.map_ind[0], reg.map_ind[1], reg.map_ind[2]
            cur_val = m[kk,jj,ii]
            ind_stack = [ (ii,jj,kk) ]

            while len(ind_stack) > 0 :

                i, j, k = ind_at = ind_stack.pop ( 0 )

                if ind_at in rm:
                    rm[ind_at].append ( reg )
                else:
                    rm[ind_at] = [ reg ]

                break

                if k+1 < m.shape[2] and m[ k+1, j, i ] > cur_val : ind_stack.append ( (i, j, k+1) )                    
                if k   > 0          and m[ k-1, j, i ] > cur_val : ind_stack.append ( (i, j, k-1) )                
                if j+1 < m.shape[1] and m[ k, j+1, i ] > cur_val : ind_stack.append ( (i, j+1, k) )
                if j   > 0          and m[ k, j-1, i ] > cur_val : ind_stack.append ( (i, j-1, k) )
                if i+1 < m.shape[0] and m[ k, j, i+1 ] > cur_val : ind_stack.append ( (i+1, j, k) )
                if i > 0            and m[ k, j, i-1 ] > cur_val : ind_stack.append ( (i-1, j, k) )


        status ( "Merging regions..." )

        for regs in rm.values():
            if len(regs) >= 2 :
                JoinRegions ( smod, regs )

        ComputeConnections ( smod )
        umsg ( "Grouping produced %d regions." % len(smod.regions) )



    def RestartSeg ( self ) :

        umsg ( "Restarting - deleting smoothed map and regions files." )
        
        dmap = getMod ( self.dmap.get() )
        if dmap != None :
            dmap.display = True

            files = []
            path = os.path.dirname ( dmap.data.path )
            try : files = os.listdir ( path )
            except : pass

            for f in files :
                if f.find ( "_regions_(" ) >= 0 :
                    try :
                        os.remove ( path + os.path.sep + f )
                        print " - removed " + path + os.path.sep + f
                    except : pass
        

        for m in chimera.openModels.list() :
            print m.name, " - ", 

            try :
                m.orig_map
                chimera.openModels.close ( m );
                print "closed smoothed map"
                continue
            except : pass

            if m.name.find ( "_regions" ) >= 0 :
                chimera.openModels.close ( m );
                print "closed regions model"
                continue
                

            print ""

        self.regions_file.set ( "" )
        self.ReportRegionCount(None)


    def SmoothAndGroup ( self ):

        try:
            from chimera import tasks, CancelOperation
            task = tasks.Task('Segger segmentation %s' % self.dmap.get(),
                              modal = True)
        except:
            task = None  # No task support in this Chimera version.

        if task:
            try:
                self.Smooth_and_Group(task)
            except CancelOperation:
                umsg('Cancelled segmentation')
                return
            task.finished()
        else:
            self.Smooth_and_Group()


    def Smooth_and_Group ( self, task = None ) :

        if len(self.dmap.get()) == 0 : umsg ("Select a density map in the Segger dialog first" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "Selected density map %s not open" % self.dmap.get() ); return

        smod = getMod ( self.regions_file.get() )
        if smod == None :
            # umsg ( "Please segment the map first." ); return
            self.Segment ()
            smod = getMod ( self.regions_file.get() )
            if smod == None :
                umsg ( "The map was not segmented successfully" ); return

        try : targNRegs = int(self.targNRegions.get())
        except : umsg ( "Enter an integer for target # of regions" ); return

        if targNRegs <= 0 :
            umsg ( "Enter an integer > 0 for target # of regions" )
            return

        try :
            numit = int ( self.numSteps.get() )
            step_size = sdev = float ( self.stepSize.get() )
        except :
            umsg ( "Enter <int> for # smoothing steps, <float> for step size" )
            return


        umsg ( "%d smoothing steps of %.4f, aiming for %d regions" % (
            numit, step_size, targNRegs ) )


        gs_s_level = 0

        from numpy import single as floatc
        sm_mat = dmap.data.full_matrix().astype(floatc)

        dmap.display = False

        last_num_regs = len ( smod.regions )

        num_steps = 0

        for iti in range ( numit ) :

            num_steps = num_steps + 1

            umsg ( "Smoothing... (step %d, step size %.1f) ..." % (iti+1, sdev) )
            if task:
                task.updateStatus('Smoothing step %d of %d' % (iti + 1, numit))

            gs_s_level += sdev

            try :
                import GaussianFilter
                use_gf = True
            except:
                use_gf = False

            if use_gf :
                tf = dmap.matrix_indices_to_xyz_transform(False, False)
                step = [Matrix.length([tf[a][b] for a in range(3)])
                        for b in range(3)]
                ijk_linewidths = [sdev*2.0 / s for s in step]
                sm = GaussianFilter.gaussian_convolution(sm_mat, ijk_linewidths)

            else :
                from VolumeFilter import gaussian
                region = dmap.subregion(1, None)
                origin, step = dmap.region_origin_and_step ( region )
                ijk_sdev = [sdev / s for s in step]
                try:
                    sm = gaussian.gaussian_convolution (sm_mat, ijk_sdev,
                                                        task = task)
                except:
                    # No task support in this Chimera version.
                    sm = gaussian.gaussian_convolution (sm_mat, ijk_sdev)

            sm_mat[:,:,:] = sm

            self.Group_by_Tracking_Maxima ( smod, sm_mat, task )

            num_regs = len ( smod.regions.values() )
            self.ReportRegionCount(num_regs)

            if self.saveIntermediateSegs.get() :
                
                sname = smod.name.split ( "_(" ) [0] + "_(%3.0f_%d)" % (gs_s_level, len(smod.regions))

                #smod.name = smod.name.split ( "_(" ) [0] + "_(%3.0f_%d)" % (gs_s_level, len(smod.regions))
                #self.regions_file.set ( smod.name )

                if num_regs != last_num_regs :
                    try : self.WriteRegionsFile ( smod.path + sname )
                    except : pass

                last_num_regs = num_regs

            if num_regs <= int(self.targNRegions.get()) : break


        try : self.WriteRegionsFile ( smod.path + smod.name )
        except : pass

        umsg ( "Got %d regions after %d steps." % ( len(smod.regions), num_steps) )

        took_sec = time.clock() - self.t0
        print "Time: %f sec" % took_sec



    def GroupUsingFits ( self ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print "Map %s not open" % self.dmap.get(); return

        smod = getMod ( self.regions_file.get() )
        if smod == None : print "Regions :", self.regions_file.get(), "not open"; return

        if len(smod.regions) == 0 : print "No regions in", smod.name; return
        try : dmap.fitted_mols
        except : dmap.fitted_mols = []
        if len(dmap.fitted_mols) == 0 : print "No fits found for", dmap.name; return

        print "Grouping %d regions by overlap to %d fitted structures" % (
            len(smod.regions), len(dmap.fitted_mols) )

        dmap.chain_maps = []

        for mol in dmap.fitted_mols :
            try : mol.fmap.imap
            except : mol.fmap.imap = self.MapIndexesInMap ( dmap, mol.fmap )
            mol.fmap.surf_color = ( rand(), rand(), rand(), 1 )
            dmap.chain_maps.append ( mol.fmap )

        self.SegAccuracy ( "_fits_acc", True )



    def RegSurfsShowAll ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return
        
        for sp in smod.surfacePieces :
            sp.display = True


    def RegSurfsShow ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return

        for p in smod.surfacePieces : p.display = False
        for r in SelectedRegions ( smod ) : r.display = True


    def RegSurfsHide ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return

        sregs = SelectedRegions ( smod )        
        if len(sregs) == 0 : sregs = smod.surfacePieces

        for p in sregs : p.display = False


    def RegSurfsShowAdjacent ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return

        sregs = SelectedRegions ( smod )        
        if len(sregs) == 0 : sregs = smod.surfacePieces

        for s in sregs :
            try :
                cregs = smod.cons[s.region]
            except :
                umsg ( "Region has no adjacency information" )
                continue

            umsg ( "Region has %d adjacent regions" % len(cregs.values()) )
            for cr, o in cregs.iteritems() :
                smod.rid_sp_map[cr.rid].display = True



    def RegSurfsTransparent ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None :
            umsg ( self.regions_file.get() + "not open" )
            return

        sregs = SelectedRegions ( smod )        
        if len(sregs) == 0 : sregs = smod.surfacePieces

#        chimera.viewer.keyLight.specular = chimera.MaterialColor(0,0,0,1)
        for sp in sregs :
            clr = sp.region.color
            sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )

        # smod.piecesAreSelectable = False



    def RegSurfsOpaque ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None :
            umsg ( self.regions_file.get() + "not open" )
            return

#        chimera.viewer.keyLight.specular = chimera.MaterialColor(1,1,1,1)

        sregs = SelectedRegions ( smod )        
        if len(sregs) == 0 : sregs = smod.surfacePieces

        for sp in sregs :
            clr = sp.region.color
            sp.color = ( clr[0], clr[1], clr[2], 1.0 )


        # smod.piecesAreSelectable = True


    def RegSelInvert ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None :
            umsg ( self.regions_file.get() + "not open" )
            return

        sel_surfs = SelectedRegions ( smod )

        nsel = []
        for sp in smod.surfacePieces :
            if sel_surfs.count ( sp ) == 0 :
                nsel.append ( sp )
                
        chimera.selection.clearCurrent ()
        chimera.selection.addCurrent ( nsel )        
    

    def Group ( self ):

        self.JoinSelRegs()

    def JoinSelRegs ( self ) :

        if len ( self.regions_file.get() ) == 0 : umsg ("Please select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        surfs = SelectedRegions ( smod )
        if len(surfs)==0 :
            umsg ( "No regions selected" )
            return

        regs = []
        for sp in surfs : regs.append ( sp.region )

        jreg = JoinRegions ( smod, regs )
        ComputeConnections ( smod )

        chimera.selection.setCurrent([smod.rid_sp_map[jreg.rid]])

        self.ReportRegionCount(len(smod.regions))
        umsg ( "Grouped %d regions" % len(regs) )
        # self.RegsDispUpdate ()




    def DelSelRegs ( self ) :

        if len ( self.regions_file.get() ) == 0 : umsg ("Please select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        surfs = SelectedRegions ( smod )
        if len(surfs)==0 :
            umsg ( "Please select one ore more regions to register the structure with" )
            return

        for sp in surfs :
            DeleteRegion ( smod, sp.region )

        ComputeConnections ( smod )
        self.ReportRegionCount(len(smod.regions))
        umsg ( "Deleted %d regions" % len(surfs) )




    def Ungroup ( self ):

        self.UngroupSelRegs()


    def UngroupSelRegs ( self ) :

        if len(self.regions_file.get()) == 0 : umsg ( "Select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "Regions file %s not open" % self.regions_file.get() ); return

        sps = SelectedRegions ( smod )

        if len(sps) == 0 :
            umsg ( "No regions selected" )
            return

        print "%d selected regions" % len(sps)
        chimera.selection.clearCurrent ()

        plist = []
        for spi, sp in enumerate ( sps ) :

            if len(sp.region.cregs) == 0 :
                plist.append(sp)
                umsg ( "Region %d/%d has no sub-groups..." % (spi+1, len(sps)) )

            else :

                umsg ( "Ungrouping region %d/%d into %d subregions..." % (
                    spi+1, len(sps), len(sp.region.cregs)) )

                for creg in sp.region.cregs :

                    #creg.rid = max(smod.regions.keys()) + 1
                    smod.regions[creg.rid] = creg

                    vt = MultiScale.surface.surface_points ( creg.rpoints, 1.0, 0.1, .25, 5 )
                    _contour.affine_transform_vertices ( vt[0], smod.seg_map.data.ijk_to_xyz_transform )
                    nsp = makeRegionSurface ( smod, vt[0], vt[1], creg )
                    plist.append(nsp)

                    creg.preg = None

                print "Deleting old region %d" % sp.region.rid
                del smod.regions [ sp.region.rid ]
                del smod.rid_sp_map [ sp.region.rid ]
                smod.removePiece ( sp )

        ComputeConnections ( smod )

        chimera.selection.setCurrent (plist)

        self.ReportRegionCount(len(smod.regions))



    def UngroupAllRegs ( self ) :

        if len(self.regions_file.get()) == 0 : umsg ( "Select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "Regions file %s not open" % self.regions_file.get() ); return

        sps = smod.surfacePieces

        if len(sps) == 0 :
            umsg ( "No regions to ungroup" )
            return

        
        for spi, sp in enumerate ( sps ) :

            status ( "Ungrouping region %d/%d..." % (spi, len(sps)) )

            for creg in sp.region.cregs :

                creg.rid = max(smod.regions.keys()) + 1
                smod.regions[creg.rid] = creg

                vt = MultiScale.surface.surface_points ( creg.rpoints, 1.0, 0.1, .25, 5 )
                _contour.affine_transform_vertices ( vt[0], smod.seg_map.data.ijk_to_xyz_transform )
                nsp = makeRegionSurface ( smod, vt[0], vt[1], creg )

                creg.preg = None

            del smod.regions [ sp.region.rid ]
            del smod.rid_sp_map [ sp.region.rid ]
            smod.removePiece ( sp )

        ComputeConnections ( smod )
        umsg ( "Ungrouped %d regions -> %d regions" % (len(sps), len(smod.surfacePieces)) )
        self.ReportRegionCount(len(smod.regions))






    def CloseRegions ( self ) :

        rfile = self.regions_file.get()

        smod = getMod ( rfile )
        if smod == None :
            print rfile, "not open"
            return

        chimera.openModels.remove ( smod )
        rfile = self.regions_file.set('')
        self.ReportRegionCount(None)

        try : chimera.openModels.remove ( smod.adj_graph_mol )
        except : pass


    def DelRegionsFiles ( self ) :

        rfile = self.regions_file.get()

        smod = getMod ( rfile )
        if smod == None :
            print rfile, "not open"
            return

        print "deleting", smod.path + rfile
        try : os.remove ( smod.path + rfile )
        except : print " - failed"

        chimera.openModels.remove ( smod )
        rfile = self.regions_file.set('')

        try : chimera.openModels.remove ( smod.adj_graph_mol )
        except : pass
            





    def RegionsVolume ( self ) :

        if len(self.regions_file.get()) == 0 : umsg ( "Select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "Regions file %s not open" % self.regions_file.get() ); return

        sregs = []
        sel_sps = SelectedRegions ( smod )
        print "%d selected surfaces" % len(sel_sps)

        if len(sel_sps) > 0 :
            for sp in sel_sps :
                sregs.append ( sp.region )
        else :
            sregs = smod.regions.values()

        if len(sregs) == 0 :
            umsg ( "No regions found in %s" % smod.name )
            return

        tvol = 0.0
        for reg in sregs :
            # print " - %d, volume %.3f" % ( r.rid, r.volume )
            tvol = tvol + reg.volume

        rw = "region"
        if len(sregs) > 1 : rw = "regions"
        umsg ( "Volume of %d %s: %.3e Angstroms^3" % ( len(sregs), rw, tvol ) )




    def RegionsAdjGraph ( self ) :

        if len(self.regions_file.get()) == 0 : print "Please select a regions file"; return
        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return

        print "Adjacencies for", smod.name

        gname = smod.name + "_graph"

        gmod = getMod ( gname )
        if gmod : chimera.openModels.close ( gmod )

        nmol = chimera.Molecule()
        chimera.openModels.add ( [nmol] )
        nmol.name = gname
        smod.adj_graph_mol = nmol
        aMap = dict()
        nres = nmol.newResidue( 'SEG', chimera.MolResId('SegR', 1))

        for sp in smod.surfacePieces :

            nv = len ( sp.rpoints )
            aname = "%d, vol %.3f" % ( sp.rid, sp.volume )

            com = numpy.sum(sp.tpoints, axis=0) / float(nv)
            COM = chimera.Vector ( com[0], com[1], com[2] )

            nat = nmol.newAtom ( aname, chimera.Element(1))
            aMap[sp] = nat
            nres.addAtom ( nat )
            nat.setCoord ( chimera.Point(COM[0], COM[1], COM[2]) )
            nat.drawMode = nat.Sphere
            nat.radius = 2.0
            nat.color = chimera.MaterialColor ( sp.region.color[0], sp.region.color[1], sp.region.color[2], 1.0 )
            nat.display = True
            nat.name = aname


        if smod.cons == None :
            print " - no connections"
            return

        for r1 in smod.cons.keys() :
            for r2 in smod.cons[r1].keys() :
                if r2 > r1 :
                    o = smod.cons[r1][r2]
                    nb = nmol.newBond ( aMap[r1], aMap[r2] )
                    nb.display = True
                    nb.drawMode = nb.Stick
                    nb.halfbond = False
                    nb.radius = 0.5
                    #f = float(w) / float(max_w)
                    #nb.color = chimera.MaterialColor ( 1, 1-f, 1-f, 1 )
                    nb.color = chimera.MaterialColor ( .5, .5, .5, 1 )



    def ShowRegionsAxes ( self, regs ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        for sp in regs :

            try :
                sp.axes.display = True
                chimera.openModels.close ( sp.axes )
            except :
                pass

            tpoints = sp.region.rpoints.copy()
            _contour.affine_transform_vertices ( tpoints, dmap.data.ijk_to_xyz_transform )

            sp.COM, sp.U, sp.S, sp.V = prAxes ( tpoints )

            com = numpy.sum(tpoints, axis=0) / len(tpoints)
            comv = numpy.ones_like ( tpoints ) * com
            points = tpoints - comv

            ppoints = points * sp.U
            sp.Extents = numpy.asarray ( numpy.max ( numpy.abs ( ppoints ), 0 ) )[0]

            if 0 and rand() > 0.5 :
                sp.U[0,0] = sp.U[0,0] * -1.0
                sp.U[1,0] = sp.U[1,0] * -1.0
                sp.U[2,0] = sp.U[2,0] * -1.0
            if rand() > 0.5 : r_y = -1.0
            if rand() > 0.5 : r_z = -1.0

            sp.Extents[0] = sp.Extents[0] + 5.0
            sp.Extents[1] = sp.Extents[1] + 5.0
            sp.Extents[2] = sp.Extents[2] + 5.0

            sp.axes = AxesMod ( sp.COM, sp.U, sp.Extents, 1.0, 1.1,
                                alignTo = sp.model )
            sp.axes.name = "region_%d_axes" % sp.region.rid



    def ShowRegionAxesSelected ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None :
            print self.regions_file.get(), "not open"
            return

        sregs = SelectedRegions ( smod )        
        if len(sregs)==0 : print "no selected regions found"; return

        self.ShowRegionsAxes ( sregs )


    def HideRegionAxes ( self ) :

        print "hiding axes"
        for m in OML() :
            t = m.name.split ("_")
            if t[0] == "region" and t[2] == "axes" :
                print "- removing", m.name
                chimera.openModels.close( m )



    def StrucCenter ( self ) :

        if len(self.struc.get()) == 0 : umsg ("Please select a structure first"); return
        fmol = getMod ( self.struc.get() )
        if fmol == None : umsg ( "%s not open" % self.struc.get() ); return

        centerMol ( fmol )

        mm = getMod ( self.dmap.get() )
        if mm : fmol.openState.xform = mm.openState.xform



    def StrucShowAxes ( self ) :

        if len ( self.struc.get() ) == 0 :
            print "Please select a structure"
            return
        
        print "Showing axes for", self.struc.get()

        fmol = getMod ( self.struc.get() )
        if fmol == None :
            print self.struc.get(), "not open";
            return

        try :
            chimera.openModels.close ( fmol.axes )
            fmol.axes = None
        except :
            pass

        if not hasattr(fmol, 'centered'):
            centerMol(fmol)
        print fmol.COM
        print fmol.Extents

        #fmol.axes = AxesMod ( fmol.COM, fmol.U, fmol.Extents, 1.0 )
        fmol.axes =  AxesMod ( Extents = fmol.Extents, rad = 1.0,
                               alignTo = fmol )
        fmol.axes.name = os.path.splitext (fmol.name)[0] + "_axes"



    def StrucHideAxes ( self ) :

        fmol = getMod ( self.struc.get() )
        if fmol == None : print self.struc.get(), "not open"; return

        try :
            chimera.openModels.close ( fmol.axes )
            fmol.axes = None
        except :
            pass




    def SelStrucMap_G ( self, sel_str, mol, dmap, color ) :

        atoms = chimera.selection.OSLSelection( sel_str ).atoms()

        #points = _multiscale.get_atom_coordinates ( atoms, transformed = False )

        # transform to index reference frame of ref_map
        f1 = Matrix.xform_matrix ( mol.openState.xform )
        f2 = Matrix.xform_matrix ( dmap.openState.xform.inverse() )
        f3 = dmap.data.xyz_to_ijk_transform

        tf = Matrix.multiply_matrices( f2, f1 )
        tf = Matrix.multiply_matrices( f3, tf )

        #_contour.affine_transform_vertices ( points, tf )

        M = numpy.zeros_like ( dmap.data.matrix() )
        n1, n2, n3 = dmap.data.size[0], dmap.data.size[1], dmap.data.size[2]

        #for fi, fj, fk in points :
            #for i in [ int(numpy.floor(fi)), int(numpy.ceil(fi)) ] :
            #    for j in [ int(numpy.floor(fj)), int(numpy.ceil(fj)) ] :
            #        for k in [ int(numpy.floor(fk)), int(numpy.ceil(fk)) ] :
            #            if i < 0 or i >= n3 : continue
            #            if j < 0 or j >= n2 : continue
            #            if k < 0 or k >= n1 : continue
            #            mm[k,j,i] = 1.0


        for at in atoms :

            c = at.coord()
            p = numpy.array ( [ [ c[0], c[1], c[2] ] ], numpy.float32 );
            _contour.affine_transform_vertices ( p, tf )
            fi, fj, fk = p[0][0], p[0][1], p[0][2]

            # trilinearly interpolate onto 8 nearest grid points
            il, jl, kl = numpy.floor(fi), numpy.floor(fj), numpy.floor(fk)
            ih, jh, kh = numpy.ceil(fi), numpy.ceil(fj), numpy.ceil(fk)

            val = float ( at.element.number );
            # val = 1.0
            vil = (fi-il) * val
            vih = (ih-fi) * val

            viljl = (fj-jl)*vil;
            viljh = (jh-fj)*vil;
            vihjl = (fj-jl)*vih;
            vihjh = (jh-fj)*vih;

            viljlkl = (fk-kl)*viljl;
            viljlkh = (kh-fk)*viljl;
            viljhkl = (fk-kl)*viljh;
            viljhkh = (kh-fk)*viljh;
            vihjlkl = (fk-kl)*vihjl;
            vihjlkh = (kh-fk)*vihjl;
            vihjhkl = (fk-kl)*vihjh;
            vihjhkh = (kh-fk)*vihjh;

            M [kl][jl][il] = M[kl][jl][il] + viljlkl;
            M [kh][jl][il] = M[kh][jl][il] + viljlkh;
            M [kl][jh][il] = M[kl][jh][il] + viljhkl;
            M [kh][jh][il] = M[kh][jh][il] + viljhkh;
            M [kl][jl][ih] = M[kl][jl][ih] + vihjlkl;
            M [kh][jl][ih] = M[kh][jl][ih] + vihjlkh;
            M [kl][jh][ih] = M[kl][jh][ih] + vihjhkl;
            M [kh][jh][ih] = M[kh][jh][ih] + vihjhkh;


        tf = dmap.matrix_indices_to_xyz_transform ( False, False )
        origin = [tf[a][3] for a in range(3)]
        step = [Matrix.length([tf[a][b] for a in range(3)]) for b in range(3)]
        ijk_linewidths = [ 2.244 / s for s in step]
        M = GaussianFilter.gaussian_convolution ( M, ijk_linewidths )

        sgd = VolumeData.Array_Grid_Data (M, dmap.data.origin, dmap.data.step, dmap.data.cell_angles, dmap.data.rotation)
        gv = VolumeViewer.volume.add_data_set ( sgd, None )
        gv.openState.xform = dmap.openState.xform

        points = _multiscale.get_atom_coordinates ( atoms, transformed = False )
        dvals = gv.interpolated_values ( points, mol.openState.xform )
        min_d = numpy.min ( dvals )
        print " min d: %.4f," % min_d,

        gv.display = True
        gv.surface_levels = [ min_d ]
        gv.surface_colors = [ color ]
        gv.surf_color = color
        ro = VolumeViewer.volume.Rendering_Options()
        ro.surface_smoothing = True
        ro.smoothing_factor = .2
        ro.smoothing_iterations = 5
        gv.region = ( gv.region[0], gv.region[1], [1,1,1] )
        gv.update_surface ( False, ro )

        for sp in gv.surfacePieces :
            v, t = sp.geometry
            if len(v) == 8 and len(t) == 12 : sp.display = False

        cm = gv.data.full_matrix()

        #nze = numpy.nonzero ( cm )
        #print "%d NZ" % len ( nze[0] ),

        cmt = numpy.where ( cm > min_d, cm, numpy.zeros_like(cm) )
        nze = numpy.nonzero ( cmt )
        print " %d NZ" % len ( nze[0] ),

        gv.imap = {}
        for ei, i in enumerate ( nze[0] ) :
            j = nze[1][ei]
            k = nze[2][ei]

            try : mi = gv.imap[i]
            except : mi = {}; gv.imap[i] = mi

            try : mij = mi[j]
            except : mij = {}; mi[j] = mij

            mij[k] = 1


        com = numpy.sum(points, axis=0) / len(points)
        gv.COM = chimera.Vector ( com[0], com[1], com[2] )
        comv = numpy.ones_like ( points ) * com
        points_v = points - comv
        gv.bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points_v), 1 ) ) )

        return gv



    def SelStrucMap_Thr ( self, sel_str, mol, dmap, color ) :

        atoms = chimera.selection.OSLSelection( sel_str ).atoms()

        if len(atoms) == 0 :
            print "- empty"
            return None
        
        points = _multiscale.get_atom_coordinates ( atoms, transformed = True )
        _contour.affine_transform_vertices ( points, Matrix.xform_matrix(dmap.openState.xform.inverse()) )

        step = chimera.Vector ( dmap.data.step[0], dmap.data.step[1], dmap.data.step[2] )
        sg = VolumeData.zone_masked_grid_data ( dmap.data, points, step.length )

        sgm = sg.matrix()
        #sgmt = numpy.where ( sgm > 0, sgm, numpy.ones_like(sgm)*1e99 )
        #min_d = numpy.min ( sgmt )

        #sgmt = numpy.where ( sgm > 0, numpy.ones_like(sgm), numpy.zeros_like(sgm) )
        #sgd = VolumeData.Array_Grid_Data (sgmt, dmap.data.origin, dmap.data.step, dmap.data.cell_angles, dmap.data.rotation)

        try : gv = VolumeViewer.volume.add_data_set ( sg, None )
        except : gv = VolumeViewer.volume.volume_from_grid_data ( sg )
        gv.openState.xform = dmap.openState.xform

        dvals = dmap.interpolated_values ( points, mol.openState.xform )
        min_d = numpy.min ( dvals ) / 2.0
        print " - min d: %.4f," % min_d,

        gv.region = ( gv.region[0], gv.region[1], [1,1,1] )

        gv.display = True
        gv.surface_levels = [ min_d ]
        #gv.surface_levels = [ dmap.surface_levels[0] ]
        #gv.surface_levels = [20.0]
        gv.surface_colors = [ color ]
        gv.surf_color = color
        ro = VolumeViewer.volume.Rendering_Options()
        ro.surface_smoothing = True
        #ro.smoothing_factor = .25
        #ro.smoothing_iterations = 10
        gv.update_surface ( False, ro )

        for sp in gv.surfacePieces :
            v, t = sp.geometry
            if len(v) == 8 and len(t) == 12 : sp.display = False

        cm = gv.data.matrix()

        nze = numpy.nonzero ( cm )
        print "%d NZ \\" % len ( nze[0] ),

        cmt = numpy.where ( cm > min_d, cm, numpy.zeros_like(cm) )
        nze = numpy.nonzero ( cmt )
        print " %d NZ" % len ( nze[0] )
        gv.d_thr = min_d

        gv.imap = {}
        for ei, i in enumerate ( nze[0] ) :
            j = nze[1][ei]
            k = nze[2][ei]

            try : mi = gv.imap[i]
            except : mi = {}; gv.imap[i] = mi

            try : mij = mi[j]
            except : mij = {}; mi[j] = mij

            mij[k] = 1


        com = numpy.sum(points, axis=0) / len(points)
        gv.COM = chimera.Vector ( com[0], com[1], com[2] )
        comv = numpy.ones_like ( points ) * com
        points_v = points - comv
        gv.bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points_v), 1 ) ) )

        return gv



    def SelStrucMap ( self, sel_str, mol, dmap, color ) :

        atoms = chimera.selection.OSLSelection( sel_str ).atoms()

        if len(atoms) == 0 :
            print "- empty"
            return None
        
        points = _multiscale.get_atom_coordinates ( atoms, transformed = True )
        _contour.affine_transform_vertices ( points, Matrix.xform_matrix(dmap.openState.xform.inverse()) )

        step = chimera.Vector ( dmap.data.step[0], dmap.data.step[1], dmap.data.step[2] )
        sg = VolumeData.zone_masked_grid_data ( dmap.data, points, 2.0 )

        sgm = sg.matrix()
        #sgmt = numpy.where ( sgm > 0, sgm, numpy.ones_like(sgm)*1e99 )
        #min_d = numpy.min ( sgmt )

        #sgmt = numpy.where ( sgm > 0, numpy.ones_like(sgm), numpy.zeros_like(sgm) )
        #sgd = VolumeData.Array_Grid_Data (sgmt, dmap.data.origin, dmap.data.step, dmap.data.cell_angles, dmap.data.rotation)

        try : gv = VolumeViewer.volume.add_data_set ( sg, None )
        except : gv = VolumeViewer.volume.volume_from_grid_data ( sg )
        gv.openState.xform = dmap.openState.xform

        dvals = dmap.interpolated_values ( points, mol.openState.xform )
        min_d = numpy.min ( dvals ) / 2.0
        min_d = dmap.surface_levels[0]

        print " - min d: %.4f," % min_d,

        gv.region = ( gv.region[0], gv.region[1], [1,1,1] )

        gv.display = True
        gv.surface_levels = [ min_d ]
        #gv.surface_levels = [20.0]
        gv.surface_colors = [ color ]
        gv.surf_color = color
        ro = VolumeViewer.volume.Rendering_Options()
        ro.surface_smoothing = True
        #ro.smoothing_factor = .25
        #ro.smoothing_iterations = 2
        gv.update_surface ( False, ro )
        gv.atoms = atoms

        for sp in gv.surfacePieces :
            v, t = sp.geometry
            if len(v) == 8 and len(t) == 12 : sp.display = False

        cm = gv.data.matrix()

        nze = numpy.nonzero ( cm )
        print "%d NZ \\" % len ( nze[0] ),

        cmt = numpy.where ( cm > min_d, cm, numpy.zeros_like(cm) )
        nze = numpy.nonzero ( cmt )
        print " %d NZ" % len ( nze[0] )
        gv.d_thr = min_d

        gv.imap = {}
        for ei, i in enumerate ( nze[0] ) :
            j = nze[1][ei]
            k = nze[2][ei]

            try : mi = gv.imap[i]
            except : mi = {}; gv.imap[i] = mi

            try : mij = mi[j]
            except : mij = {}; mi[j] = mij

            mij[k] = 1


        com = numpy.sum(points, axis=0) / len(points)
        gv.COM = chimera.Vector ( com[0], com[1], com[2] )
        comv = numpy.ones_like ( points ) * com
        points_v = points - comv
        gv.bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points_v), 1 ) ) )

        return gv



    def MakeChainMaps ( self, mols, dmap ) :

        try : print len(dmap.chain_maps), "chain maps so far"
        except : dmap.chain_maps = []


        if len(mols) == 1 :

            mol = mols[0]
            print "Making chain maps for %s in %s:" % (mol.name, dmap.name)

            for cid, clr in mol.chain_colors.iteritems() :

                map_name = os.path.splitext ( dmap.name )[0]
                cname = map_name + "_" + cid
                sel_str = "#%d:.%s" % (mol.id, cid)
                print "%s [%s]" % (cname, sel_str),

                gv = self.SelStrucMap ( sel_str, mol, dmap, clr.rgba() )
                #gv = self.SelStrucMap_G ( sel_str, mol, dmap, clr.rgba() )
                if gv :
                    gv.name = cname
                    gv.chain_id = cid
                    dmap.chain_maps.append ( gv )
                    gv.mol = mol

                #return


        else :

            print "Making struc maps for %d strucs" % len(mols)

            for mol in mols :

                map_name = os.path.splitext ( dmap.name )[0]
                mol_name = os.path.splitext ( mol.name )[0]
                cname = map_name + "_" + mol_name
                sel_str = "#%d" % (mol.id)
                print "%s [%s]" % (cname, sel_str),
                
                clr = (rand(), rand(), rand(), 1.0)
                gv = self.SelStrucMap ( sel_str, mol, dmap, clr )
                gv.name = cname
                gv.chain_id = mol_name
                dmap.chain_maps.append ( gv )




    def StrucChainMaps ( self ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        mols = []
        if self.UseAllMods.get() :
            for mod in chimera.openModels.list() :
                if type(mod) == chimera.Molecule :
                    mols.append ( mod )

        else :
            m = getMod ( self.struc.get() )
            if m != None :
                mols.append ( m )
            m.chain_colors = RandColorChains ( m )

        if len(mols) == 0 :
            print "No structures"; return

        self.MakeChainMaps ( mols, dmap )
        print "- %d chain or unit maps" % len ( dmap.chain_maps )



    def StrucShowChainMaps ( self ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        print "%s - showing %d chain maps" % (dmap.name, len(dmap.chain_maps))
        for chm in dmap.chain_maps :
            chm.display = True


    def StrucHideChainMaps ( self ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        print "%s - hiding %d chain maps" % (dmap.name, len(dmap.chain_maps))
        for chm in dmap.chain_maps :
            chm.display = False


    def StrucCloseChainMaps ( self ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return
        
        print "%s - closing %d chain maps" % (dmap.name, len(dmap.chain_maps))

        for chm in dmap.chain_maps :
            chm.close()

        dmap.chain_maps = []



    def StrucDelChainMaps ( self ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        print "%s - deleting %d chain maps" % (dmap.name, len(dmap.chain_maps))
        for chm in dmap.chain_maps :
            try : os.remove ( path + chm.name )
            except : print " - could not delete", chm.name
            chm.close()            

        dmap.chain_maps = []


    def RMSD ( self, ress, fmol ) :

        N, D, posr = 0.0, 0.0, {}
        for r in ress : posr[r.id.position] = r
        for r in fmol.residues :
            v = r.atomsMap["CA"][0].xformCoord() - posr[r.id.position].atomsMap["CA"][0].xformCoord()
            N, D = N + 1, D + (v.length * v.length)

        return numpy.sqrt ( D / N )
        

    def StrucChMapsOvRegs ( self ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        print "%s - %d chain maps" % (dmap.name, len(dmap.chain_maps))

        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return
        if len(smod.regions) == 0 : print " - no regions!"; return


        for mn in ["2AW4.pdb", "2AVY.pdb"] :
            m = getMod (mn)
            if m :
                for res in m.residues :
                    res.ribbonDisplay = False


        for chm in dmap.chain_maps :
            try : chm.display = False
            except : continue


        for chm in dmap.chain_maps :

            try : chm.display = False
            except : continue

            if chm.chain_id == "A" : continue
            if chm.chain_id == "B" and chm.mol.name == "2AW4.pdb" : continue

            chm.display = True
            print "%s - %s" % (chm.name, chm.chain_id)
            self.struc_map.set ( chm.name )

            oregs = self.ShowOverlappingRegions ()
            chimera.viewer.viewAll ()


            chm.display = False

            if len(oregs) == 0 :
                print " - NO overlapping regions"
                continue


            ress = chimera.selection.OSLSelection( "#%d:.%s" % (chm.mol.id, chm.chain_id) ).residues()
            for res in ress : res.ribbonDisplay = True

            mname = os.path.splitext ( chm.mol.name )[0]
            fname = mname + "_%s.pdb" % chm.chain_id
            fmol = getMod ( fname )
            if fmol == None :
                fmol = chimera.openModels.open ( path + fname )[0]
                fmol.ch_colors = RandColorChains ( fmol )

            self.struc.set ( fname )
            self.StrucCenter ()
            self.simRes.set ( "14" )
            self.simGridSp.set ( "2" )
            self.GenStrucMap ()
            sim_dmap = getMod ( self.struc_map.get() )
            sim_dmap.chmap = chm

            self.rotaSearch.set ( 0 )
            self.FitMapToSelRGroup ()
            rmsd = self.RMSD ( ress, fmol )
            print "\nRMSD: %f " % rmsd,
            tag = chm.chain_id

            msk_dmap = None
            
            if rmsd > 15.0 :
                msk_dmap = self.MaskMapWRegions ()
                self.FitMapToSelRGroup ()
                rmsd = self.RMSD ( ress, fmol )
                print "\nRMSD: %f " % rmsd,
                tag = chm.chain_id + "m"

            if rmsd > 15.0 :
                self.dmap.set ( dmap.name )
                self.rotaSearch.set ( 1 )
                self.FitMapToSelRGroup ()
                rmsd = self.RMSD ( ress, fmol )
                print "\nRMSD: %f " % rmsd,
                tag = chm.chain_id + "r"

            if rmsd > 15.0 :
                self.dmap.set ( msk_dmap.name )
                self.rotaSearch.set ( 1 )
                self.FitMapToSelRGroup ()
                rmsd = self.RMSD ( ress, fmol )
                print "\nRMSD: %f " % rmsd,
                tag = chm.chain_id + "rm"


            if rmsd < 15.0 :

                print "_________ YES _______\n"

                self.SelRegsOptimizeShapeScore ()
                oregs = self.ShowOverlappingRegions ()
                self.SelRegsOptimizeShapeScore ()
                oregs = self.ShowOverlappingRegions ()

                path = os.path.dirname ( dmap.data.path ) + os.path.sep
                log_file = path + "ribo_fits_sms_corr.txt"
                print "SMS log to", log_file
                fpl = open ( log_file, 'a' )
                fpl.write ( "%s %f " % (tag, sim_dmap.sms) )
                fpl.close ()

                if len(oregs) == 1 :
                    oregs[0].placed = True
                if len(oregs) >= 2 :
                    nreg = JoinRegions ( smod, oregs )
                    #self.RegsDispUpdate ()

            else :

                print "_________ NO _______\n"


            #sim_dmap.close()
            sim_dmap.display = False

            self.dmap.set ( dmap.name )
            if msk_dmap : msk_dmap.close()

            self.StrucCloseChainMap()

            #return
            

    def StrucCloseChainMap ( self ) :

        chm = getMod ( self.struc_map.get() )
        if chm :
            print " --- closing ", chm.mol.name
            chimera.openModels.close ( [chm.mol] )
            print " --- closing ", chm.chmap.name
            chm.chmap.close()
            print " --- closing ", chm.name
            chm.close()
            self.struc_map.set ( "" )




    def StrucBestShapeScore ( self ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.fmap.get(), "not found"; return

        # first find out which region overlaps the most
        
        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return
        if len(smod.regions) == 0 : print " - no regions!"; return

        max_overlap_num = 0
        max_overlap_reg = None

        print "Region with highest overlap: ",

        for reg_i, region in enumerate ( smod.regions.values() ) :

            ipoints = numpy.cast[numpy.int] ( region.rpoints )

            noverlap = 0
            for i,j,k in ipoints :
                try : noverlap = noverlap + imap[k][j][i]
                except : continue

            if noverlap > max_overlap_num :
                max_overlap_num = noverlap
                max_overlap_reg = region

        if max_overlap_num == 0 :
            print "no region overlaps"
            return None

        else :
            print "%d (%d)" % (max_overlap_reg.rid, max_overlap_num)    
            sms = self.ShapeMatchScore ( fmap.fmol.atoms, dmap, [max_overlap_reg] )
            print " - shape match : %f" % sms

        return sms

    



    def NotSegAccuracy ( self ) :

        mol = getMod ( self.struc.get() )
        if mol == None : print self.struc.get(), "not open"; return
        print "Structure:", mol.name

        try : chain_colors = mol.chain_colors
        except :  chain_colors = RandColorChains ( mol )
        mol.chain_colors = chain_colors

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        try : mol.chain_maps
        except : self.MakeChainMaps ( mol, dmap )
        print "- %d chain maps" % len ( mol.chain_maps )


        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return
        print "- %d regions" % len ( smod.surfacePieces )

        ch_regs = {}

        for reg_i, region in enumerate ( smod.regions.values() ) :

            max_overlap_num = 0
            max_overlap_cmap = None

            ipoints = numpy.cast[numpy.int] ( region.rpoints )

            for cm in mol.chain_maps :
                
                noverlap = 0
                for i,j,k in ipoints :
                    try : noverlap = noverlap + cm.imap[k][j][i]
                    except : continue

                if noverlap > max_overlap_num :
                    max_overlap_num = noverlap
                    max_overlap_cmap = cm

            if max_overlap_num == 0 :
                print "%d of %d - %d %d points - " % (
                    reg_i+1, len(smod.regions), region.rid, len(ipoints) )

            else :
                print "%d of %d - %d %d points - max overlap %d with %s" % (
                    reg_i+1, len(smod.regions), region.rid, len(ipoints), max_overlap_num, max_overlap_cmap.name )

                c = max_overlap_cmap.surf_color
                region.color = c
                # smod.rid_sp_map[reg_i].color = ( c[0], c[1], c[2], 1.0 )

                try : ch_regs[max_overlap_cmap].append ( region )
                except : ch_regs[max_overlap_cmap] = [ region ]


        thr = dmap.surface_levels[0]
        scores = []

        for cm in mol.chain_maps :
            print cm.name,
            if ch_regs.has_key ( cm ) == False :
                print " - no regions!"
                continue

            regs = ch_regs[cm]
            print "%d regions" % len(regs),


            points = regs[0].rpoints.copy()
            for r in regs[1:] : points = numpy.concatenate ( [points, r.rpoints], axis=0 )
            _contour.affine_transform_vertices ( points, smod.seg_map.data.ijk_to_xyz_transform )


            sg = VolumeData.zone_masked_grid_data ( dmap.data, points, 0.5 )
            regsm = sg.matrix()

            nz = len(regsm.nonzero()[0])
            if nz != len(points) :
                print "mask failed - %d of %d pts nonzero" % (nz, len(points))
                continue

            print "R %d nz" % nz,

            cmm = cm.data.matrix()
            thr = dmap.surface_levels[0]
            cmm = numpy.where ( cmm > thr, numpy.ones_like(cmm), numpy.zeros_like(cmm) )

            print "C %d >%.3f" % (len(cmm.nonzero()[0]), thr),

            nz_int =  numpy.shape ( (regsm * cmm).nonzero () )[1]
            nz_uni =  numpy.shape ( (regsm + cmm).nonzero () )[1]

            sm_score = float(nz_int) / float (nz_uni)

            print "I : %d | U : %d = %.3f" % (nz_int, nz_uni, sm_score)

            scores.append ( sm_score )


        dname = os.path.splitext ( dmap.name ) [0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        log_f = path + dname + "_acc.txt"
        print "Accuracies log file:", log_f

        try : fp = open ( log_f, "a" )
        except : print " - could not open log file"; return

        fp.write ( "%f %f %f " % ( min(scores), max(scores), sum(scores)/float(len(scores))) )
        for sm_score in scores : fp.write ( "%f " % sm_score )

        fp.write ( "\n" )
        fp.close ()



    def SegAccuracy ( self, tag = "_acc", joinRegs=True ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return
        map_name = os.path.splitext ( dmap.name )[0]

        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return

        cmaps = dmap.chain_maps
        print "\nMapping %d regions -> %d chain maps..." % (
            len(smod.regions), len(cmaps) )

        if len(smod.regions) == 0 : print " - no regions!"; return
        if len(cmaps) == 0 : print " - no chain maps!"; return

        ch_regs = {}

        for reg_i, region in enumerate ( smod.regions.values() ) :

            max_overlap_num = 0
            max_overlap_cmap = None

            ipoints = numpy.cast[numpy.int] ( region.rpoints )

            for cm in cmaps :

                noverlap = 0
                for i,j,k in ipoints :
                    try : noverlap = noverlap + cm.imap[k][j][i]
                    except : continue

                if noverlap > max_overlap_num :
                    max_overlap_num = noverlap
                    max_overlap_cmap = cm

            if max_overlap_num == 0 :
                print "%d of %d - %d %d points - " % (reg_i+1, len(smod.regions), region.rid, len(ipoints) )

            else :
                # print "%d/%d - %d %d points - max overlap %d with %s" % (reg_i+1, len(smod.regions), region.rid, len(ipoints), max_overlap_num, max_overlap_cmap.name )
                c = max_overlap_cmap.surf_color
                region.color = c
                # smod.rid_sp_map[reg_i].color = ( c[0], c[1], c[2], 1.0 )

                try : ch_regs[max_overlap_cmap].append ( region )
                except : ch_regs[max_overlap_cmap] = [ region ]

        thr = dmap.surface_levels[0]
        scores = []

        for cm in cmaps :
            print cm.name,
            if ch_regs.has_key ( cm ) == False :
                print " - no regions!"
                continue

            regs = ch_regs[cm]
            print " - %d regions" % len(regs),

            sms = self.ShapeMatchScore ( cm.atoms, dmap, regs )
            print " - sms %.4f" % (sms)

            if sms > 0.3 and len(regs) < 10 :
                scores.append ( sms )

            else :
                print "sms low: %.3f, nregs: %d" % (sms, len(regs))

            if joinRegs : JoinRegions ( smod, regs )


        map_name = os.path.splitext ( dmap.name ) [0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        log_f = path + map_name + tag + ".txt"
        print "\nAccuracies log file:", log_f

        res = NumberFromName ( dmap.name, 'r' );
        if res == None : res = 0.0
        print " - resolution: %.0f, %d scores" % ( res, len(scores) )

        try : fp = open ( log_f, "a" )
        except : print " - could not open log file"; return
        fp.write ( "%f %f %f %f " % ( res, min(scores), max(scores), sum(scores)/float(len(scores))) )
        for sm_score in scores : fp.write ( "%f " % sm_score )
        fp.write ( "\n" )
        fp.close ()

        # if joinRegs : self.RegsDispUpdate ()





    def ShowChRegs ( self ) :

        mol = getMod ( self.struc.get() )
        if mol == None : print self.struc.get(), "not open"; return
        print "Structure:", mol.name

        try : chain_colors = mol.chain_colors
        except :  chain_colors = RandColorChains ( mol )
        mol.chain_colors = chain_colors

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        try : mol.chain_maps
        except : self.MakeChainMaps ( mol, dmap )
        print "- %d chain maps" % len ( mol.chain_maps )

        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return
        print "- %d regions" % len ( smod.surfacePieces )


        # for sp in smod.surfacePieces : sp.display = False

        rgroups = {}
        rgroups['all'] = []
        ov_regs = []

        for chmap in mol.chain_maps :

            # if chmap.chain_id == 'A' or chmap.chain_id == 'B' : continue
            # chmap.display = True

            regs = self.OverlappingRegions ( dmap, chmap, smod )

            ov_regs = ov_regs + regs

            if len(regs) == 0 :
                print " - no regions found";
                return

            # rgroups[chmap.chain_id] = regs

            #sms = self.ShapeMatchScore ( chmap, dmap, regs )
            #print " - chain %s, map overlaps %d regions - %.4f" % ( chmap.chain_id, len(regs), sms )

            for r in regs :
                #r.display = True
                #r.color = chmap.surf_color
                rgroups['all'].append ( r )

            # break

        if 0 :
            # for overlapping regs, which chain do they overlap the most
            for r in ov_regs :
                max_ov = 0.0
                max_ov_chm = None
                for chm in mol.chain_maps :
                    imap = self.MapIndexesInMap ( dmap, chm ) 
                    ipoints = numpy.cast[numpy.int] ( r.rpoints )
                    noverlap = 0
                    for i,j,k in ipoints :
                        try : imap[k][j][i]
                        except : continue
                        noverlap = noverlap + 1

                    ov = float(noverlap) / float ( len(r.rpoints) )
                    if ov > max_ov :
                        max_ov = ov
                        max_ov_chm = chm

                try : rgroups[max_ov_chm.chain_id]
                except : rgroups[max_ov_chm.chain_id] = []
                rgroups[max_ov_chm.chain_id].append ( r )

        return rgroups



    def SimChainMaps ( self ) :

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print "dmap not open"; return

        map_name = os.path.splitext ( dmap.name ) [0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        res, grid = None, None

        try : res = float ( self.simRes.get() )
        except : print "Invalid number entered for resolution:", self.simRes.get(); return

        try : grid = float ( self.simGridSp.get() )
        except : grid = res / 3.0

        mol = getMod ( self.struc.get() )
        if mol == None : print "Structure", self.struc.get(), "not found"; return

        try : mol.chain_colors
        except : mol.chain_colors = RandColorChains ( mol )
            
        print "Simulating %d chain maps for %s, res %.3f, grid %.3f" % (
            len(mol.chain_colors.keys()), mol.name, res, grid)


        dmap.fitted_mols = []

        for cid, clr in mol.chain_colors.iteritems() :

            cname = map_name + "_" + cid + ".mrc"

            mv = getMod ( cname )

            if mv == None :

                sel_str = "#%d:.%s" % (mol.id, cid)
                print "%s [%s]" % (cname, sel_str),

                cmd = "molmap %s %f sigmaFactor 0.187 gridSpacing %f" % ( sel_str, res, grid )
                print " -", cmd
                chimera.runCommand ( cmd )

                for mod in chimera.openModels.list() :
                    ts = mod.name.split()
                    if len(ts) > 1 and mod.name.find("map") >=0 and mod.name.find("res") >=0 :
                        print " - saving to:", path + cname
                        mod.write_file ( path + cname, "mrc" )
                        chimera.openModels.close ( mod )
                        break

                mv = VolumeViewer.open_volume_file ( path + cname )[0]
                print " - loaded:", mv.name

            fmol = aClass()
            fmol.fmap = mv
            
            dmap.fitted_mols.append ( fmol )

            #gv.imap = imap
            #gv.name = cname
            #gv.chain_id = cid
            #dmap.chain_maps.append ( gv )

        


    def GenStrucMap ( self , show = True ) :

        res, grid, ress = None, None, None
        res_grid = (self.simRes.get(), self.simGridSp.get())

        if len(res_grid) >= 1 :
            rs = res_grid[0].split (":")
            if len(rs) == 3 :
                try :
                    first_res = float(rs[0])
                    incr_res = float(rs[1])
                    last_res = float(rs[2])
                except :
                    print "Invalid number entered for resolutions:", res_grid[0]
                    return

                ress = range ( first_res, last_res+1.0, incr_res )
                print "Resolutions:", ress
            else :
                try :
                    res = float ( rs[0] )
                except :
                    print "Invalid number entered for resolution:", res_grid[0]
                    return

        if len(res_grid) >= 2 and res_grid[1]:
            try :
                grid = float ( res_grid[1] )
            except :
                print "Invalid number for grid spacing:", res_grid[1]
                return

        
        if res == None and ress == None :
            print "Please enter a resolution"
            return

        elif res != None :
            if grid == None : grid = res/3.0
            dmap = self.DoGenStrucMap ( res, grid, show )
            if dmap == None: return
            self.struc_map.set ( dmap.name )


        elif ress != None :

            dmap = getMod ( self.dmap.get() )
            if dmap == None : print "dmap not open"; return

            start_thr = dmap.surface_levels[0]

            for r in ress :

                if grid == None : grid = r/3.0
                print "____________", r, "____________", grid, "____________"

                mv = self.DoGenStrucMap ( r, grid, show )

                print "Starting threshold: %f" % start_thr
                mv.surface_levels[0] = start_thr
                mv.update_surface ( False, VolumeViewer.volume.Rendering_Options() )

                try : dmap.chain_maps
                except : print "No chain maps for %s" % dmap.name; return

                if dmap.name == mv.name :
                    print "First model"
                else :
                    mv.chain_maps = []
                    print "Transferring %d chain maps from %s..." % ( len(dmap.chain_maps), dmap.name )
                    for cmap in dmap.chain_maps :
                        mv.chain_maps.append ( TransferChainMap (cmap, mv) )

                self.dmap.set ( mv.name )
                self.Segment ();
                self.SegAccuracy ( "_acc_max" )
                self.Smooth_and_Group ();
                self.SegAccuracy ( "_acc" )

                if dmap.name == mv.name :
                    print "First model"
                else :
                    for cm in mv.chain_maps : cm.close()
                    mv.close()

                smod = getMod ( self.regions_file.get() )
                chimera.openModels.close ( smod )



    def DoGenStrucMap ( self, res, grid, show = True ) :

        mols = []
        if self.UseAllMods.get() :
            for mod in chimera.openModels.list() :
                if type(mod) == chimera.Molecule : mols.append ( mod )
        else :
            if len(self.struc.get()) == 0 :
                umsg ( "Please select a structure first" )
                return
            m = getMod ( self.struc.get() )
            if m == None :
                umsg ( "%s not open" % self.struc.get() )
                return
            mols.append ( m )

        if len(mols) == 0 :
            umsg ( "Structure not found" )
            return

        for m in mols:
            if not hasattr ( m, 'centered' ):
                centerMol ( m )

        path = os.path.dirname ( mols[0].openedAs[0] ) + os.path.sep

        #try : m.centered
        #except : self.StrucCenter ()

        base = ""
        sel_str = ""
        for m in mols :
            base = base + os.path.splitext(m.name)[0] + "_"
            sel_str = sel_str + "#%d," % m.id

        sel_str = sel_str [:-1]
        mname = base + "r%.1f_sp%.1f.mrc" % (res, grid)

        mv = getMod ( mname )
        if mv != None :
            print "Found", mname
            return mv
        
        print "Generating", mname

        cmd = "molmap %s %f sigmaFactor 0.187 gridSpacing %f" % ( sel_str, res, grid )
        print " -", cmd
        chimera.runCommand ( cmd )

        mv = None

        for mod in chimera.openModels.list() :
            ts = mod.name.split()
            if len(ts) > 1 and mod.name.find("map") >=0 and mod.name.find("res") >=0 :
                print " - found", mod.name
                #print " - saving to:", path + mname
                #mod.write_file ( path + mname, "mrc" )
                #print " - closing:", mod.name
                #chimera.openModels.close ( mod )
                mv = mod
                mv.name = mname
                break

        #mv = VolumeViewer.open_volume_file ( path + mname )[0]
        #print " - loaded:", mv.name
        mv.display = show
        clr = mols[0].color.rgba()
        mv.surfacePieces[0].color = ( clr[0], clr[1], clr[2], 1.0 )
        mv.mol = mols[0]

        return mv



    def StrucMapVolume ( self ) :

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        vol = _surface.enclosed_volume ( *(fmap.surfacePieces[0].geometry) )[0]
        print fmap.name + " volume: %f" % vol

        print " - surface levels",fmap.surface_levels
        thr = fmap.surface_levels[0]

        mm = fmap.data.matrix()
        mmab = numpy.where ( mm > thr, numpy.ones_like(mm), numpy.zeros_like(mm) )
        nz = numpy.shape ( numpy.nonzero ( mmab ) )[1]
        vvol = fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]
        tvol = vvol * float(nz)
        print "%s - %d above %f, volume %.3f" % (fmap.name, nz, thr, tvol)

        vvol = fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]

        tvol = vvol * float(nz)
        print " - %d above %f, volume %.3f" % (nz, thr, tvol)




    def FitSMapToDMap ( self ) :

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        print "Fitting %s to %s" % ( fmap.name, dmap.name )

        ov, corr = self.FitMapLocal ( fmap, dmap )

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA

        try : fmap.mol.chain_maps
        except : fmap.mol.chain_maps = []
        for chm in fmap.mol.chain_maps : chm.openState.xform = xfA


    def StrucBioMT ( self ) :

        if len(self.struc.get()) == 0 :
            umsg ( "Please select a structure first" )
            return
        mol = getMod ( self.struc.get() )
        if mol == None :
            umsg ( "%s not open" % self.struc.get() )
            return


        mats = BioMatrices ( mol )

        umsg ( "%s - %d matrices" % ( mol.name, len(mats) ) )




    def NextBioMt ( self ) :

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return False

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return False

        try : mol = fmap.mol
        except : "no mol"; return False

        try : mol.bio_mts
        except : mol.bio_mts = BioMatrices ( mol )

        print " - %d bio matrices" % len(mol.bio_mts)

        try : mol.bio_mt_at = mol.bio_mt_at + 1
        except : mol.bio_mt_at = 1

        print "_______________ at matrix %d _______________" % mol.bio_mt_at

        try : amat = mol.bio_mts[ mol.bio_mt_at ]
        except : print " - no such matrix"; return False

        fmap.M = fmap.M0 * am_2_M(amat)

        # orthogonalize M
        T = fmap.M; xf = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3], True )
        mt, mr = xf_2_M ( xf )
        fmap.M = mt * mr

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3], True )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA

        try : mol.chain_maps
        except : mol.chain_maps = []
        for chm in mol.chain_maps : chm.openState.xform = xfA

        return True



    def GoBioMt ( self ) :

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return False
        try : mol = fmap.mol
        except : "no mol"; return False

        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return

        smod.bioM_groups = {}

        mol.bio_mt_at = 0

        for r in smod.surfacePieces :
            r.max_ov = 0.0
            r.max_ov_cid = None
            r.max_ov_bioM = None

        smod.p_ctrs = []            

        while self.NextBioMt () :

            self.FitSMapToDMap ()

            smod.bio_mt_at = mol.bio_mt_at
            smod.bioM_groups[mol.bio_mt_at] = self.ShowChRegs ()

            #self.ZeroDMap_with_FMap ()


            if 1 :
                atoms = chimera.selection.OSLSelection( "#%d:.%s" % (fmap.mol.id, 'P') ).atoms()
                points = _multiscale.get_atom_coordinates ( atoms, transformed = False )
                print " - %d points" % len(points)
                com = numpy.sum(points, axis=0) / len(points)
                comv = numpy.array ( [ [ com[0], com[1], com[2] ] ], numpy.float32 )
                _contour.affine_transform_vertices( comv, Matrix.xform_matrix(fmap.openState.xform) )
                _contour.affine_transform_vertices( comv, Matrix.xform_matrix(smod.openState.xform.inverse()) )

                print com[0], com[1], com[2]
                print comv[0][0], comv[0][1], comv[0][2]

                C = chimera.Vector ( comv[0][0], comv[0][1], comv[0][2] )
                smod.p_ctrs.append ( C )

            #return



    def ExtractProteins ( self ) :

        dmap = getMod ( self.dmap.get() )
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        if len(self.struc.get()) == 0 : umsg ("Please select a structure first"); return
        fmol = getMod ( self.struc.get() )
        if fmol == None : umsg ( "%s not open" % self.struc.get() ); return

        try : fmol.ch_colors
        except : fmol.ch_colors = RandColorChains ( fmol )

        mname = os.path.splitext ( fmol.name )[0]

        print "%s - %d chains" % (fmol.name, len(fmol.ch_colors.keys()))

        for cid, clr in fmol.ch_colors.iteritems () :

            print cid,
            cmol = copyMolChain ( None, fmol, cid, cid, None, clr.rgba() )
            cmol.name = "%s_%s.pdb" % (mname, cid)

            print " - %d atoms" % len(cmol.atoms)

            if len(cmol.atoms) > 0 :
                print " - writing", path + cmol.name
                chimera.PDBio().writePDBfile ( [cmol], path + cmol.name )


        print ""
        


    def SaveStrucFit ( self ) :
        
        if len ( self.struc_map.get() ) == 0 : umsg ("Please select a registered map first" ); return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : umsg ( "%s is not open" % self.struc_map.get() ); return

        self.SaveFit ( fmap )


    def SaveFit ( self, fmap, clr=(rand(), rand(), rand(), 1) ) :

        dmap = getMod ( self.dmap.get() )
        if dmap :
            try : fit_m_at = len ( dmap.fitted_mols ) + 1
            except : fit_m_at = 1; dmap.fitted_mols = []
        else :
            umsg ( "Density map %s not open" % self.dmap.get() )
            return

        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        mol = CopyMol ( fmap.mol )
        mol.name = os.path.splitext ( mol.name )[0] + "_f%d.pdb" % fit_m_at

#        r = fmap.mol.residues[0]
#        c = r.ribbonColor
#        if c == None:
#            c = r.molecule.color
#        clr = c.rgba()
        clr = ( rand(), rand(), rand(), 1.0 )

        molApplyT ( mol, fmap.M )
        print "saving", path + mol.name

        chimera.PDBio().writePDBfile ( [mol], path + mol.name )
        m = chimera.openModels.open ( path + mol.name )[0]
        print "opened", m.name
        for res in m.residues :
            res.ribbonDisplay = True
            res.ribbonDrawMode = 2
            res.ribbonColor = chimera.MaterialColor( clr[0], clr[1], clr[2], 1.0 );
            for at in res.atoms :
                at.display = False

        if dmap : dmap.fitted_mols.append ( m )

        m.display = True

        umsg ( "Saved %s" % ( path + mol.name ) )

        if 0 :
            m.fmap = fmap.copy ()
            m.fmap.name = os.path.splitext (m.name)[0] + ".mrc"
            print "Saved fit map %s (thr %f)" % ( m.fmap.name, m.fmap.surface_levels[0] )


        return m
        


    def FitToSelRegion ( self ) :

        if len ( self.dmap.get() ) == 0 : umsg ("Please select a segmented map first" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        if len ( self.regions_file.get() ) == 0 : umsg ("Please select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        fmap = self.SimulatedMap()
        if fmap == None:
            return

        surfs = SelectedRegions ( smod )
        if len(surfs) != 1 :
            umsg ( "Please select one region to register the structure with" )
            return

        regs = [ sp.region for sp in surfs ]
   
        self.FitToEachRegion ( smod, fmap, dmap, regs, 1 )



    def FitToSelectedRegions ( self ) :

        if len ( self.dmap.get() ) == 0 : umsg ("Please select a density map first" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        if len ( self.regions_file.get() ) == 0 : umsg ("Please select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        fmap = self.SimulatedMap()
        if fmap == None:
            return

        surfs = SelectedRegions ( smod )
        if len(surfs) == 0 :
            #umsg ( "Please select one or more regions to register the structure with" )
            #return
            surfs = smod.surfacePieces
            print "Fitting to all regions"

        regs = [ sp.region for sp in surfs if hasattr(sp, 'region') ]

        self.FitToEachRegion ( smod, fmap, dmap, regs, 0 )



    def FitToEachRegion ( self, smod, fmap, dmap, regs, show=1 ) :

        if not hasattr(fmap.mol, 'centered'):
            self.StrucCenter()

        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        reportFitRegions(fmap.name, regs)
        scores = []
        corrs = []

        #if mdlg : mdlg.Step ( 60 )

#        chimera.viewer.keyLight.specular = chimera.MaterialColor(0,0,0,1)

        for reg in regs :

            for sp in smod.surfacePieces :
                try : sp.region
                except : continue
                if sp.region == reg :
                    sp.display = True
                    clr = sp.region.color
                    sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
                else :
                    pass
                    #sp.display = False

            if mdlg :
                if show : mdlg.movieOn.set ( 1 )
                else : mdlg.movieOn.set ( 0 )

            # if mdlg : mdlg.Step ( 90 )

            fmap.display = False
            fmap.mol.display = True

            self.FitMapToRegion ( reg )

            if mdlg :
                mdlg.movieOn.set ( 1 )
                #mdlg.Step ( 30 )

            if 0 :
                #fmap.sms_score = self.OptShapeScore ( fmap, dmap, [ reg ] )
                fmap.sms_score = self.ShapeMatchScore ( fmap.mol.atoms, dmap, [reg] )
                # fmap.fit_score = correlation ( fmap, dmap, True )

            scores.append ( 0 )
            corrs.append ( fmap.fit_score )

            if self.strucCopyAfterFit.get() :
                reg.placed = self.SaveFit ( fmap, reg.color )

            umsg ( "Cross-correlation of fit: %f" % fmap.fit_score )


        res = NumberFromName ( fmap.name, 'r' )
        if res == None : res = 0.0

        acc_log_file = path + smod.name + "_fits_acc.txt"
        print "Accuracies log to", acc_log_file

        try : fp = open ( acc_log_file, "a" )
        except : print " - could not open log file"; return

        fp.write ( "%f %f %f %f " % ( res, min(scores), max(scores), sum(scores)/float(len(scores))) )
        for sm_score in scores : fp.write ( "%f " % sm_score )
        fp.write ( "\n" )

        fp.write ( "C%f %f %f %f " % ( res, min(corrs), max(corrs), sum(corrs)/float(len(corrs))) )
        for sm_score in corrs : fp.write ( "%f " % sm_score )
        fp.write ( "\n" )

        fp.close ()



    def FitMapToRegion ( self, sreg ) :

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        tpoints = sreg.rpoints.copy()
        _contour.affine_transform_vertices ( tpoints, dmap.data.ijk_to_xyz_transform )


        if self.rotaSearch.get () :
            self.saFitMapToPoints_byRot ( fmap, tpoints, dmap )
        else :
            self.saFitMapToPoints ( fmap, tpoints, dmap )





    def saFitMapToPoints ( self, fmap, points, dmap ) :

        print "fitting %s in map %s, to %d points" % (fmap.name, dmap.name, len(points))
        COM, U, S, V = prAxes ( points )

        # if mdlg : mdlg.Step ( 90 )

        # mT = self.CSegMol
        highest_fit, highest_M, highest_MI, opt_af = 0.0, None, None, None

        comT = numpy.matrix ( [
            [ 1, 0, 0, COM[0] ],
            [ 0, 1, 0, COM[1] ],
            [ 0, 0, 1, COM[2] ],
            [ 0, 0, 0,      1 ]  ] )

        tXO, tXR = xf_2_M ( dmap.openState.xform )

        fmap.fpoints = VolumeData.grid_indices ( fmap.data.size, numpy.single)        # i,j,k indices
        _contour.affine_transform_vertices( fmap.fpoints, fmap.data.ijk_to_xyz_transform )
        fmap.fpoint_weights = numpy.ravel( fmap.data.matrix() ).astype(numpy.single)

        threshold = fmap.surface_levels[0]
        #threshold = .3 * max ( fmap.fpoint_weights )

        ge = numpy.greater_equal(fmap.fpoint_weights, threshold)
        fmap.fpoints = numpy.compress(ge, fmap.fpoints, 0)
        fmap.fpoint_weights = numpy.compress(ge, fmap.fpoint_weights)
        nz = numpy.nonzero( fmap.fpoint_weights )[0]

        print " - threshold %.4f, %d nonzero, center %s" % ( threshold, len(nz), str(COM) )

        if len(nz) < len (fmap.fpoint_weights) :
            fmap.fpoints = numpy.take( fmap.fpoints, nz, axis=0 )
            fmap.fpoint_weights = numpy.take(fmap.fpoint_weights, nz, axis=0)

        mass = numpy.sum(fmap.fpoint_weights, dtype=numpy.single)
        fmap.rotation_center = numpy.dot(fmap.fpoint_weights,fmap.fpoints) / mass


        # the 4 alignments to try...
        amult = [ [1.0,1.0,1.0], [-1.0,-1.0,1.0], [-1.0,1.0,-1.0], [1.0,-1.0,-1.0] ]

        for j in range(4) :

            af = amult[j]
            print " ~ %.1f*X %.1f*Y %.1f*Z" % ( af[0], af[1], af[2] )

            mR = numpy.matrix ( [
                [ af[0]*U[0,0], af[1]*U[0,1], af[2]*U[0,2], 0 ],
                [ af[0]*U[1,0], af[1]*U[1,1], af[2]*U[1,2], 0 ],
                [ af[0]*U[2,0], af[1]*U[2,1], af[2]*U[2,2], 0 ],
                [            0,            0,            0, 1 ] ] )

            fmap.M = comT * mR
            T = tXO * tXR * fmap.M
            xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
            fmap.openState.xform = xfA

            # if j == 0 : return

            #if mdlg : mdlg.Step ( 90 )

            ov, corr = self.FitMap_T ( fmap, dmap, num_tries=4 )

            #if mdlg : mdlg.Step ( 60 )

            # ov, corr = map_overlap_and_correlation ( fmap, dmap, True )
            #print "- final overlap: %f, correlation: %f\n" % (ov, corr)
            #mA.avD, n = self.AvgDVal ( dmap, mA )

            if corr > highest_fit :
                highest_fit = corr
                highest_M = fmap.M
                highest_MI = comT * mR
                opt_af = af

        print " * highest fit: %f for %.1f*X %.1f*Y %.1f*Z" % (
            highest_fit, opt_af[0], opt_af[1], opt_af[2] )

        fmap.M = highest_M
        fmap.MI = highest_MI
        fmap.fit_score = highest_fit

        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA


        #if mdlg : mdlg.Step ( 60 )


        if 0 :
            f_m = fmap.data.full_matrix(); size = list(f_m.shape); size.reverse()
            fmap.fpoints = VolumeData.grid_indices(size, numpy.single)        # i,j,k indices
            _contour.affine_transform_vertices( fmap.fpoints, fmap.data.ijk_to_xyz_transform )
            fmap.fpoint_weights = numpy.ravel(f_m).astype(numpy.single)

            threshold = fmap.surface_levels[0]
            ge = numpy.greater_equal(fmap.fpoint_weights, threshold)
            fmap.fpoints = numpy.compress(ge, fmap.fpoints, 0)
            fmap.fpoint_weights = numpy.compress(ge, fmap.fpoint_weights)
            nz = numpy.nonzero( fmap.fpoint_weights )[0]
            print " - threshold %.4f, %d nonzero" % ( threshold, len(nz) ), COM

            ov, corr = self.FitMap_T ( fmap, dmap, num_tries=2 )
            #ov, corr = map_overlap_and_correlation ( fmap, dmap, True )
            print " - correlation with full threshold: %f" % corr
            fmap.fit_score = corr

            T = tXO * tXR * fmap.M
            xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
            fmap.openState.xform = xfA
            fmap.mol.openState.xform = xfA



    def saFitMapToPoints_byRot ( self, fmap, points, dmap, N=8, M=8 ) :

        fpLog = None
        if 0 : fpLog = open ( "fit_rot_log.txt", "w" )

        print "fitting %s in map %s, to %d points, by rotation" % (fmap.name, dmap.name, len(points))
        COM, U, S, V = prAxes ( points )

        comT = numpy.matrix ( [
            [ 1, 0, 0, COM[0] ],
            [ 0, 1, 0, COM[1] ],
            [ 0, 0, 1, COM[2] ],
            [ 0, 0, 0,      1 ]  ] )

        tXO, tXR = xf_2_M ( dmap.openState.xform )

        f_m = fmap.data.full_matrix(); size = list(f_m.shape); size.reverse()
        fmap.fpoints = VolumeData.grid_indices(size, numpy.single)        # i,j,k indices
        _contour.affine_transform_vertices( fmap.fpoints, fmap.data.ijk_to_xyz_transform )
        fmap.fpoint_weights = numpy.ravel(f_m).astype(numpy.single)

        #threshold = fmap.surface_levels[0]
        threshold = .3 * max ( numpy.ravel(f_m).astype(numpy.single) )

        ge = numpy.greater_equal(fmap.fpoint_weights, threshold)
        fmap.fpoints = numpy.compress(ge, fmap.fpoints, 0)
        fmap.fpoint_weights = numpy.compress(ge, fmap.fpoint_weights)
        nz = numpy.nonzero( fmap.fpoint_weights )[0]

        print " - threshold %.4f, %d nonzero" % ( threshold, len(nz) ), COM

        if len(nz) < len (fmap.fpoint_weights) :
            fmap.fpoints = numpy.take( fmap.fpoints, nz, axis=0 )
            fmap.fpoint_weights = numpy.take(fmap.fpoint_weights, nz, axis=0)

        mass = numpy.sum(fmap.fpoint_weights, dtype=numpy.single)
        fmap.rotation_center = numpy.dot(fmap.fpoint_weights,fmap.fpoints) / mass


        # dmap = self.DoMask ( dmap, mol_set )

        highest_fit, highest_M, highest_MI, opt_af = 0.0, None, None, None

        # tt = self.fitByRot_NM.get().split(',')
        # N, M = int ( tt[0] ), int ( tt[1] )
        # N, M = 10, 10
        print "N %d, M %d" % (N, M)


        thetas, phis = [], []
        # http://www.math.niu.edu/~rusin/known-math/97/spherefaq


        for k in range ( 1, N+1 ) :
            h = -1.0 + ( 2.0*float(k-1)/float(N-1) )
            thetas.append ( numpy.arccos(h) )
            phi = 0.0
            if k == 1 or k == N : phi = 0
            else : phi = (phis[k-2] + 3.6/numpy.sqrt(float(N)*(1.0-h**2.0)))
            while phi > 2.0*numpy.pi : phi -= 2.0*numpy.pi
            phis.append ( phi )


        for ti, thet in enumerate ( thetas ) :
            phi, theta = phis[ti], thet - (numpy.pi / 2.0)
            v = chimera.Vector (0,0,0)
            v[2] = numpy.sin(theta)
            v[1] = numpy.cos(theta) * numpy.sin (phi)
            v[0] = numpy.cos(theta) * numpy.cos (phi)

            print "%d/%d : %.3f (%.3f), %.3f %.3f, %f" % (
                ti, len(thetas), theta, theta*180.0/numpy.pi, phi, phi*180.0/numpy.pi, v.length )

            for m in range ( M ) :

                ang = 360.0 * float(m)/float(M)
                print "- %d/%d - %.3f deg" % (m, M, ang)

                status ( "Rotational fit %d/%d" % (ti*M+m+1, N*M) )

                xfR = chimera.Xform.rotation ( v, ang )
                rrO, rrR = xf_2_M ( xfR )

                fmap.M = comT * rrR
                T = tXO * tXR * fmap.M
                xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
                fmap.openState.xform = xfA
                fmap.mol.openState.xform = xfA

                if ti == 0 and m == 0 :
                    if mdlg : mdlg.Step ( 90 )

                if mdlg : mdlg.Step ( 1 )

                ov, corr = self.FitMap_T ( fmap, dmap, num_tries=1 )

                # ov, corr = map_overlap_and_correlation( fmap, dmap, True )
                #print "- corr %f" % corr

                if fpLog : fpLog.write ( "%f\n" % corr )

                if corr > highest_fit :
                    highest_fit = corr
                    highest_M = fmap.M
                    highest_MI = comT * rrR


        print "- highest fit: %f" % ( highest_fit )
        print ""

        if fpLog : fpLog.close()
        

        fmap.M = highest_M
        fmap.MI = highest_MI
        fmap.fit_score = highest_fit

        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA





    def FitMap_T ( self, fmap, dmap, Log=False, num_tries=1, bTrans=True, bRot=True ) :

        fpLog = None
        if 0 :
            try : self.fitno = self.fitno + 1
            except : self.fitno = 0
            fpLog = open ( "fit_log_%d.txt" % self.fitno, "a" )

        olap, cor = 0, 0

        for gd_i in range ( num_tries ) :

            #print "%d: fitting of %s in %s" % (gd_i, fmap.name, dmap.name)
            print "   try %d of %d:" % (gd_i+1, num_tries),

            T = fmap.M; data_xf = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
            data_xyz_to_xyz = Matrix.xform_matrix( data_xf )

            xyz_to_ijk = dmap.data.xyz_to_ijk_transform
            ttf = Matrix.multiply_matrices(xyz_to_ijk, data_xyz_to_xyz)

            data_array = dmap.data.matrix()
            max_steps = 100
            ijk_step_size_max = 0.5
            ijk_step_size_min = 0.01
            optimize_translation, optimize_rotation = True, True

            ##### from FitMap.locate_maximum ####

            segment_steps = 4
            cut_step_size_threshold = .5
            step_cut_factor = .5

            ijk_step_size = ijk_step_size_max

            move_tf = Matrix.identity_matrix()

            step = 0
            while step < max_steps and ijk_step_size > ijk_step_size_min:
                xyz_to_ijk_tf = Matrix.multiply_matrices( ttf, move_tf )

                try :
                    seg_tf = FitMap.step_to_maximum(
                        fmap.fpoints, fmap.fpoint_weights,
                                             data_array, xyz_to_ijk_tf,
                                             segment_steps, ijk_step_size,
                                             bTrans, bRot,
                                             fmap.rotation_center)
                except :
                    seg_tf = FitMap.step_to_maximum(
                        fmap.fpoints, fmap.fpoint_weights,
                                             data_array, xyz_to_ijk_tf,
                                             segment_steps, ijk_step_size,
                                             bTrans, bRot,
                                             fmap.rotation_center,
                                             'sum product' )
                    
                step += segment_steps
                mm = FitMap.maximum_ijk_motion(
                    fmap.fpoints, xyz_to_ijk_tf, seg_tf)
                if mm < cut_step_size_threshold * segment_steps * ijk_step_size:
                    ijk_step_size *= step_cut_factor
                move_tf = Matrix.multiply_matrices(seg_tf, move_tf)

                #shift, angle = FitMap.shift_and_angle(seg_tf, fmap.rotation_center)
                #print '- %d steps: shift %.3g, rotation %.3g degrees, step %.3f' % (
                #    step, shift, angle, ijk_step_size)
                # print "^",
                # add move_tf to overall transform

                if 0 and mdlg :
                    xT, xR = xf_2_M ( Matrix.chimera_xform ( move_tf ) )
                    tXO, tXR = xf_2_M ( dmap.openState.xform )
                    # this is the transform after the gradient descent movement
                    T = tXO * tXR * fmap.M * xT * xR
                    xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
                    fmap.openState.xform = xfA
                    fmap.mol.openState.xform = xfA

                    print "^",
                    if mdlg : mdlg.Step ( 1 )

                if fpLog :
                    map_values, outside = VolumeData.interpolate_volume_data(
                        fmap.fpoints, xyz_to_ijk_tf, data_array )
                    olap, cor = overlap_and_correlation ( fmap.fpoint_weights, map_values )

                    fpLog.write ( "%f %f\n" % (olap, cor) )


                #if Log :
                #    avgd, n = self.AvgDVal ( dmap, mol )
                #    self.AvgDValue.set ( "%f" % avgd )

                #    log_f = self.path.get() + "_%s_%s_ConvergeMOIAlign.txt" % (dmap.name, mol.name)
                #    print "Log file: %s" % log_f
                #    f = open ( log_f, 'a' )
                #    f.write ( "%7d %7d %15f\n" % (9, step, avgd) )
                #    f.close()


            xT, xR = xf_2_M ( Matrix.chimera_xform ( move_tf ) )
            fmap.M = fmap.M * xT * xR

            map_values, outside = VolumeData.interpolate_volume_data(
                fmap.fpoints, xyz_to_ijk_tf, data_array )

            olap, cor = overlap_and_correlation ( fmap.fpoint_weights, map_values )

            try : shift, angle = FitMap.shift_and_angle( move_tf, fmap.rotation_center )
            except : shift, angle = Matrix.shift_and_angle( move_tf, fmap.rotation_center )
                
            # print ""
            print ' - %d steps: d %.3g, r %.3g, ov %.3e, cor %f' % (
                step, shift, angle, olap, cor )

            if angle < .1 and shift < .1 :
                break

        # tXO, tXR = xf_2_M ( dmap.openState.xform )
        # T = tXO * tXR * fmap.M
        # xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        # fmap.openState.xform = xfA

        if fpLog : fpLog.close()

        return olap, cor


    def FitMapLocal ( self, fmap, dmap ) :

        f_m = fmap.data.full_matrix(); size = list(f_m.shape); size.reverse()
        fmap.fpoints = VolumeData.grid_indices(size, numpy.single)        # i,j,k indices
        _contour.affine_transform_vertices( fmap.fpoints, fmap.data.ijk_to_xyz_transform )
        fmap.fpoint_weights = numpy.ravel(f_m).astype(numpy.single)

        threshold = fmap.surface_levels[0]
        #threshold = .3 * max ( numpy.ravel(f_m).astype(numpy.single) )

        ge = numpy.greater_equal(fmap.fpoint_weights, threshold)
        fmap.fpoints = numpy.compress(ge, fmap.fpoints, 0)
        fmap.fpoint_weights = numpy.compress(ge, fmap.fpoint_weights)
        nz = numpy.nonzero( fmap.fpoint_weights )[0]

        print " - threshold %.4f, %d nonzero" % ( threshold, len(nz) )

        if len(nz) < len (fmap.fpoint_weights) :
            fmap.fpoints = numpy.take( fmap.fpoints, nz, axis=0 )
            fmap.fpoint_weights = numpy.take(fmap.fpoint_weights, nz, axis=0)

        mass = numpy.sum(fmap.fpoint_weights, dtype=numpy.single)
        fmap.rotation_center = numpy.dot(fmap.fpoint_weights,fmap.fpoints) / mass

        olap, cor = self.FitMap_T ( fmap, dmap, num_tries=1 )

        fmap.fpoints = None
        fmap.fpoint_weights = None
        fmap.rotation_center = None

        return olap, cor






    def StrucGroupRegions ( self ) :

        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        if len(self.struc_map.get()) == 0 : print "Please select a structure map"; return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        if len(self.regions_file.get()) == 0 : print "Please select a regions file"; return
        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return


        sel_sps = SelectedRegions ( smod )
        regs = []
        for sp in sel_sps :
            regs.append ( sp.region )

        if len(regs)==0 :
            print "\nGrouping all regions"
            tvol = self.MapVolume ( fmap )
            smod.rgroups = self.GroupAllRegions ( smod, tvol )

        elif len(regs) == 1 :
            tvol = self.MapVolume ( fmap )
            bRad = -1.0 # self.MapBoundingRad ( fmap )

            print "\nMaking groups around region %d - target vol %.3f, b-Rad %.3f" % (regs[0].rid, tvol, bRad)
            smod.rgroups, maxDepthReached = self.GroupAroundReg ( smod, regs[0], tvol, bRad )
            print " - depth reached: %d" % maxDepthReached


        else :
            print "Please select no regions (for global groups) or one region (for local groups)"
            return

        if len ( smod.rgroups ) == 0 :
            print "No groups result!"
            return

        smod.rgroups.sort()
        smod.rgroup_at = 0
        dv, regs = smod.rgroups[smod.rgroup_at]
        print "Group %d of %d - dv %f, regions:" % (smod.rgroup_at+1, len(smod.rgroups), dv),
        for r in regs : print r.rid,
        print ""

        for sp in smod.surfacePieces :
            if regs.count ( sp.region ) > 0 : sp.display = True
            else : sp.display = False




    def GroupAroundReg ( self, smod, reg, target_volume, bRad=-1.0 ) :

        dv_rgroups = []
        maxDepthReached = 0

        stack = [ [[reg], reg.volume, 0] ]

        print "Making groups around %d" % reg.rid,
        status ( "Making groups around %d" % reg.rid ), 

        while len(stack) > 0 :

            regs, vol_at, depth_at = stack.pop()

            if depth_at > maxDepthReached : maxDepthReached = depth_at

            if depth_at >= SAF_LS_DEPTH : continue
            if vol_at >= target_volume * (1.0 + SAF_DVOL)  : continue

            dv = abs ( vol_at - target_volume ) / target_volume
            dv_rgroups.append ( [dv, regs] )

            if len(dv_rgroups) % 100 == 0 :
                status ( "Making groups around %d - %d groups" % (reg.rid, len(dv_rgroups)) ), 

            reg_at = regs[0]

            try : cmap = smod.cons[reg_at]
            except : continue

            for cr, o in cmap.iteritems() :
                if regs.count ( cr ) != 0 : continue
                if cr.placed : continue
                #sim = sim_at + ( o.D / o.N )
                vol = vol_at + cr.volume
                stack.insert ( 0, [ [cr]+regs, vol, depth_at+1 ] )

        dv_rgroups_f = self.FilterGroups ( dv_rgroups, bRad )

        print " - %d groups --> %d filtered groups" % (len(dv_rgroups), len(dv_rgroups_f))
        return [dv_rgroups_f, maxDepthReached]

        



    def FilterGroups ( self, dv_rgroups, bRad = -1.0 ) :

        dv_rgroups_f = []
        len_groups = {}
        #          [ Included, ri ]
        inc_regs_map = {}

        gi, ngroups = 0, len(dv_rgroups)

        # print "Filtering %d groups..." % ngroups

        for dv, regs in dv_rgroups :

            gi = gi + 1
            if gi % 100 == 0 :
                status ( "Filtering group %d/%d" % (gi,ngroups) )

            if dv > SAF_DVOL : continue

            included = False

            regs.sort()
            inc_regs_at = inc_regs_map
            for reg in regs :
                try : included, inc_regs_at = inc_regs_at[reg]
                except : included = False; break

            #num_regs = len(regs)
            #try : i_dv_groups = len_groups [ num_regs ]
            #except : i_dv_groups = None; len_groups [ num_regs ] = []
            #if i_dv_groups != None :
            #    for idv, iregs in i_dv_groups :
            #        bAllIn = True
            #        for r in regs :
            #            if iregs.count ( r ) == 0 : bAllIn = False; break
            #        if bAllIn == True : included = True; break

            if included : continue

            if bRad > 0.0 :

                points = regs[0].rpoints.copy()
                for r in regs[1:] : points = numpy.concatenate ( [points, r.rpoints], axis=0 )
                com = numpy.sum(points, axis=0) / len(points)
                C = chimera.Vector ( com[0], com[1], com[2] )
                comv = numpy.ones_like ( points ) * com
                points = points - comv
                regs_bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points), 1 ) ) )

                #com = regs[0].COM
                #for r in regs[1:] : com = com + r.COM
                #com = com / float( len(regs) )
                #regs_bRad = 0.0
                #for r in regs :
                #    d = (r.COM - com).length + r.bRad
                #    if d > regs_bRad : regs_bRad = d

                brad_d = abs ( regs_bRad - bRad ) / bRad

                if brad_d > SAF_DBRAD : continue

            dv_rgroups_f.append ( [dv, regs] )

            #len_groups[ num_regs ].append ( [dv,regs] )

            inc_regs_at = inc_regs_map
            last_arr = None
            for reg in regs :
                try :
                    last_arr = inc_regs_at[reg]
                    inc_regs_at = last_arr[1]
                except :
                    last_arr = [ False, {} ]
                    inc_regs_at[reg] = last_arr
                    inc_regs_at = last_arr[1]

            last_arr[0] = True


        return dv_rgroups_f




    def GroupAllRegions ( self, smod, target_volume, bRad=-1.0) :

        dv_rgroups = []
        maxDepthReached = 0

        print "Grouping %d regions in %s, target volume %.2f, bounding radius %.2f" % (
            len(smod.surfacePieces), smod.name, target_volume, bRad )

        ri, nregs = 0, len(smod.regions)

        for rid, reg in smod.regions.iteritems() :

            ri = ri + 1
            # status ( "Grouping around region %d/%d" % (ri,nregs) )

            dv_rgroupsR, maxDepthReachedR = self.GroupAroundReg ( smod, reg, target_volume, bRad )

            dv_rgroups = dv_rgroups + dv_rgroupsR

            if maxDepthReachedR > maxDepthReached : maxDepthReached = maxDepthReachedR

            
        print "\n - max depth reached: %d" % maxDepthReached

        print " - filtering %d groups..." % len( dv_rgroups )
        return self.FilterGroups ( dv_rgroups, bRad )
        #return self.FilterGroups ( dv_rgroups )





    def NextRGroup ( self ) :

        smod = getMod ( self.regions_file.get() )
        if smod == None :
            print "Please select a regions file"
            return

        smod.rgroup_at = smod.rgroup_at + 1

        if smod.rgroup_at >= len(smod.rgroups) : smod.rgroup_at = 0

        dv, regs = smod.rgroups[smod.rgroup_at]

        print "Group %d/%d - dv %f, regions:" % (smod.rgroup_at+1, len(smod.rgroups), dv),
        for r in regs : print r.rid,
        print ""

        for sp in smod.surfacePieces :
            if regs.count ( sp.region ) > 0 : sp.display = True
            else : sp.display = False



    def FindGroupFromSelRegs ( self ) :

        if len(self.regions_file.get()) == 0 : print "Please select a regions file"; return
        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return

        print "\nStructure:",
        if len(self.struc_map.get()) == 0 : print "Please select a structure map"; return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return
        tvol = self.MapVolume ( fmap )


        regs = SelectedRegions ( smod )        
        if len(regs)==0 : print "no selected regions found"; return

        print "Finding group for %d selected regions:" % ( len(regs) ),
        regs_vol = 0.0
        for r in regs :
            print r.rid,
            regs_vol = regs_vol + r.volume
        print ""

        dv = abs ( tvol - regs_vol ) / tvol;
        print " - total regions volume: %.3f - DV %.5f" % (regs_vol, dv)

        points = regs[0].tpoints.copy()
        for r in regs[1:] : points = numpy.concatenate ( [points, r.tpoints], axis=0 )

        com = numpy.sum(points, axis=0) / len(points)
        C = chimera.Vector ( com[0], com[1], com[2] )
        comv = numpy.ones_like ( points ) * com
        points = points - comv
        regs_bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points), 1 ) ) )

        print " - COM: %.3f %.3f %.3f, bounding rad %.3f" % ( C[0], C[1], C[2], regs_bRad )


        print "Searching %d groups" % len(smod.rgroups)
        gi = 0
        for dv, gregs in smod.rgroups :
            gi = gi + 1
            if len(regs) != len (gregs) : continue
            allIn = True
            for r in regs :
                if gregs.count ( r ) == 0 : allIn = False; break
            if allIn :
                print " - FOUND! %d - dv %f - regs" % (gi, dv),
                for r in gregs : print r.rid,
                print ""

        print " - done searching"



    def FitMapToSelRGroup ( self ) :

        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        fmap = self.SimulatedMap()
        if fmap == None : return
        thr = fmap.surface_levels[0]
        mm = fmap.data.matrix()
        mmab = numpy.where ( mm > thr, numpy.ones_like(mm), numpy.zeros_like(mm) )
        nz = numpy.shape ( numpy.nonzero ( mmab ) )[1]
        tvol = float(nz) * fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]
        print "%s - %d above %f, volume %.3f" % (fmap.name, nz, thr, tvol)


        if len(self.regions_file.get()) == 0 : print "Please select a regions file"; return
        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return

        sel_sps = SelectedRegions ( smod )
        if len(sel_sps)==0 : print "no selected regions found"; return

        regs = [sp.region for sp in sel_sps]
        reportFitRegions(fmap.name, regs)

        for reg, sp in zip(regs, sel_sps) :
            print reg.rid,
            clr = sp.region.color
            sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
        print ""

        points = regs[0].rpoints.copy()
        regions_vol = regs[0].volume
        for r in regs[1:] :
            points = numpy.concatenate ( [points, r.rpoints], axis=0 )
            regions_vol = regions_vol + r.volume

        dv = abs(regions_vol - tvol) / tvol
        print " - regions volume: %.2f - dv %.5f" % (regions_vol, dv)

        _contour.affine_transform_vertices ( points, dmap.data.ijk_to_xyz_transform )

        # self.saFitMapToPoints ( fmap, points, dmap )

        if self.rotaSearch.get () :
            self.saFitMapToPoints_byRot ( fmap, points, dmap )
        else :
            self.saFitMapToPoints ( fmap, points, dmap )


        #fmap.sms = self.ShapeMatchScore ( fmap.mol.atoms, dmap, regs )
        fmap.sms = 0 # self.OptShapeScore ( fmap, dmap, regs )
        #print " - shape match score: %.3f" % fmap.sms

        if self.strucCopyAfterFit.get() :
            self.SaveFit ( fmap, regs[0].color )

        umsg ( "Cross-correlation: %.4f\n" % (fmap.fit_score) )

        #dmap_name = os.path.splitext ( dmap.name )[0]
        #path = os.path.dirname ( dmap.data.path ) + os.path.sep
        #log_f = path + "%s_fits.txt" % (dmap_name)
        #print "Log file: %s" % log_f

        #fp = open ( log_f, 'a' )
        #fp.write ( "%s %f %f %f" % (fmap.mol.name, fmap.fit_score, fmap.sms, dv) )
        #T = fmap.M ; fp.write ( " %f %f %f %f %f %f %f %f %f %f %f %f" % (T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3]) )
        #T = fmap.MI; fp.write ( " %f %f %f %f %f %f %f %f %f %f %f %f" % (T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3]) )
        #for r in regs : fp.write ( " %d" % r.rid )
        #fp.write ( "\n" )
        #fp.close()





    def ShapeMaskedCorr0 ( self, fmap, dmap, regs, mask_f=0.8, bPrint=False ) :

        points = regs[0].tpoints.copy()
        for r in regs[1:] :
            points = numpy.concatenate ( [points, r.tpoints], axis=0 )

        tfd = Matrix.xform_matrix( dmap.openState.xform )
        tff = Matrix.xform_matrix( fmap.openState.xform.inverse() )
        _contour.affine_transform_vertices( points, tfd )
        _contour.affine_transform_vertices( points, tff )

        sg = VolumeData.zone_masked_grid_data( fmap.data, points, mask_f * fmap.data.step[0] )


        #r_mask_matrix, p2m_transform = self.matrix_and_transform(point_xform,
        #                                                  use_subregion,
        #                                                  use_step)
        r_mask_matrix = sg.matrix()
        p2m_transform = Matrix.xform_matrix ( chimera.Xform() )


        fmapm = fmap.data.matrix()
        f_weights = numpy.ravel(fmapm).astype(numpy.single)

        #tf = fmap.data.ijk_to_xyz_transform
        #f_t = Matrix.xform_matrix( mp.openState.xform )
        #tf = Matrix.multiply_matrices(f_t, f_i2s)
        size = list(fmapm.shape)
        size.reverse()
        f_points = VolumeData.grid_indices(size, numpy.single)        # i,j,k indices
        #_contour.affine_transform_vertices(f_points, tf)

        thr = fmap.surface_levels[0]
        ge = numpy.greater_equal(f_weights, thr)
        f_points = numpy.compress(ge, f_points, 0)
        f_weights = numpy.compress(ge, f_weights)

        m_weights, outside = VolumeData.interpolate_volume_data (
            f_points, p2m_transform, r_mask_matrix, 'linear' )

        o, shape_cor = overlap_and_correlation ( f_weights, m_weights )

        # m_weights = mp.interpolated_values ( f_points, chimera.Xform() )
        #o, shape_cor = matrix_to_map_correlation ( sg.matrix(), fmap, None )
        #sg = VolumeData.zone_masked_grid_data( dmap.data, points, .01 )
        #o, shape_cor = map_overlap_and_correlation(fmap, dm, False)

        print " * shape masked correlation %.3f" % shape_cor

        return shape_cor


    def ShapeMaskedCorr ( self, fmap, dmap, regs ) :

        points = regs[0].tpoints.copy()
        for r in regs[1:] :
            points = numpy.concatenate ( [points, r.tpoints], axis=0 )

        #tf = Matrix.xform_matrix ( dmap.openState.xform )
        #_contour.affine_transform_vertices(points, tf)

        sg = VolumeData.zone_masked_grid_data ( dmap.data, points, 0.5 )

        #gv = VolumeViewer.volume.add_data_set ( sg, None )
        #gv.name = 'regions_mask'
        #gv.openState.xform = dmap.openState.xform

        nz = numpy.shape ( numpy.nonzero ( sg.matrix() ) )[1]
        if len(points) != nz :
            print "mask failed [%d points, %d masked]" % ( len(points), nz )
            return 0.0

        r_mask_matrix = sg.matrix()

        fmapm = fmap.data.matrix()
        f_weights = numpy.ravel(fmapm).astype(numpy.single)

        size = list(fmapm.shape)
        size.reverse()
        f_points = VolumeData.grid_indices(size, numpy.single)        # i,j,k indices

        thr = fmap.surface_levels[0]
        ge = numpy.greater_equal(f_weights, thr)
        f_points = numpy.compress(ge, f_points, 0)
        f_weights = numpy.compress(ge, f_weights)

        from Matrix import multiply_matrices as MM
        from Matrix import xform_matrix as XFM

        p2m_transform = MM ( dmap.data.xyz_to_ijk_transform,
                        MM ( XFM ( dmap.openState.xform.inverse() ),
                        MM ( XFM ( fmap.openState.xform ), fmap.data.ijk_to_xyz_transform ) ) )

        #_contour.affine_transform_vertices(f_points, tf)

        m_weights, outside = VolumeData.interpolate_volume_data (
            f_points, p2m_transform, r_mask_matrix, 'linear' )

        o, shape_cor = overlap_and_correlation ( f_weights, m_weights )

        print " * shape map-masked correlation %.3f" % shape_cor

        return shape_cor




    def ShapeMatchScore0 ( self, fmap, dmap, regs, mask_f=0.8, bPrint=False ) :

        points = regs[0].tpoints.copy()
        for r in regs[1:] : points = numpy.concatenate ( [points, r.tpoints], axis=0 )

        print "1/5"
        
        tfd = Matrix.xform_matrix( dmap.openState.xform )
        tff = Matrix.xform_matrix( fmap.openState.xform.inverse() )
        tf = Matrix.multiply_matrices( tff, tfd )
        _contour.affine_transform_vertices( points, tf )

        print "2/5"

        fmapm = fmap.data.matrix()
        regsm = fmapm # numpy.ones_like ( fmapm )
        rmd = VolumeData.Array_Grid_Data ( regsm, fmap.data.origin, fmap.data.step, fmap.data.cell_angles )
        sg = VolumeData.zone_masked_grid_data( rmd, points, mask_f * fmap.data.step[0] )

        print "3/5"

        regsm = sg.matrix()
        nz = numpy.shape ( numpy.nonzero ( regsm ) )[1]
        regs_vol = float(nz) * fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]
        regs_vol_o = float(len(points)) * fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]
        regs_vol_r = 100.0 * numpy.fabs(regs_vol_o - regs_vol) / regs_vol_o

        print "4/5"

        if bPrint : print " - original regions %d points, volume %.2f" % ( len(points), regs_vol_o )
        if bPrint : print " - masked %s with regions, %d non-zero, volume %.2f (%.2f%%)" % (fmap.name, nz, regs_vol, regs_vol_r)

        #gv = VolumeViewer.volume.add_data_set ( sg, None )
        #gv.name = 'regions_mask'
        #gv.openState.xform = fmap.openState.xform

        thr = fmap.surface_levels[0]
        fmap_ab = numpy.where ( fmapm > thr, numpy.ones_like(fmapm), numpy.zeros_like(fmapm) )
        nz = numpy.shape ( numpy.nonzero ( fmap_ab ) )[1]
        tvol = float(nz) * fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]

        print "5/5"

        if bPrint : print " - values %s above %f, %d, volume %.2f" % ( fmap.name, thr, nz, tvol )

        nz_int =  numpy.shape ( (regsm * fmap_ab).nonzero () )[1]
        nz_uni =  numpy.shape ( (regsm + fmap_ab).nonzero () )[1]

        sm_score = float(nz_int) / float (nz_uni)

        if bPrint : print " - intersection %d, union %d" % (nz_int, nz_uni)

        print " * shape match score %.3f" % sm_score

        return sm_score



    def OptShapeScore ( self, fmap, dmap, regs ) :

        thr_at = fmap.surface_levels[0]
        sms = self.ShapeMatchScore ( fmap, dmap, regs )

        umsg ( "Optimizing threshold for %s - %.2f/%.4f" % (fmap.name, thr_at, sms) )

        while 1 :
            fmap.surface_levels[0] = thr_at + .01
            new_sms = self.ShapeMatchScore ( fmap, dmap, regs )

            if new_sms > sms :
                thr_at = fmap.surface_levels[0]
                sms = new_sms
                #print "- %.2f/%.4f" % (thr_at, sms),
                
                ro = VolumeViewer.volume.Rendering_Options()
                fmap.update_surface ( False, ro )
                for sp in fmap.surfacePieces :
                    v, t = sp.geometry
                    if len(v) == 8 and len(t) == 12 : sp.display = False

                umsg ( "Optimizing threshold for %s - %.2f/%.4f" % (fmap.name, thr_at, sms) )

            else :
                fmap.surface_levels[0] = thr_at
                break

        while 1 :
            fmap.surface_levels[0] = thr_at - .01
            new_sms = self.ShapeMatchScore ( fmap, dmap, regs )
            if new_sms > sms :
                thr_at = fmap.surface_levels[0]
                sms = new_sms
                print "- %.2f/%.4f" % (thr_at, sms),

                ro = VolumeViewer.volume.Rendering_Options()
                fmap.update_surface ( False, ro )
                for sp in fmap.surfacePieces :
                    v, t = sp.geometry
                    if len(v) == 8 and len(t) == 12 : sp.display = False

                umsg ( "Optimizing threshold for %s - %.2f/%.4f" % (fmap.name, thr_at, sms) )

            else :
                fmap.surface_levels[0] = thr_at
                break

        #print "| %.2f/%.4f" % (thr_at, sms)
        return sms



    def OptShapeScoreRegs ( self, fmap, dmap, regs ) :

        thr_at = dmap.surface_levels[0]
        sms = self.ShapeMatchScore ( fmap, dmap, regs )

        print ", optz regs - %.2f/%.4f" % (thr_at, sms),

        while 1 :
            dmap.surface_levels[0] = thr_at + .01
            new_sms = self.ShapeMatchScore ( fmap, dmap, regs )
            if new_sms > sms :
                thr_at = dmap.surface_levels[0]
                sms = new_sms
                print "- %.2f/%.4f" % (thr_at, sms),
            else :
                break

            ro = VolumeViewer.volume.Rendering_Options()
            dmap.update_surface ( False, ro )
            for sp in dmap.surfacePieces :
                v, t = sp.geometry
                if len(v) == 8 and len(t) == 12 : sp.display = False

        while 1 :
            dmap.surface_levels[0] = thr_at - .01
            new_sms = self.ShapeMatchScore ( fmap, dmap, regs )
            if new_sms > sms :
                thr_at = dmap.surface_levels[0]
                sms = new_sms
                print "- %.2f/%.4f" % (thr_at, sms),
            else :
                break

            ro = VolumeViewer.volume.Rendering_Options()
            dmap.update_surface ( False, ro )
            for sp in dmap.surfacePieces :
                v, t = sp.geometry
                if len(v) == 8 and len(t) == 12 : sp.display = False

        dmap.surface_levels[0] = thr_at
        print "| %.2f/%.4f" % (thr_at, sms)
        return sms





    def ShapeMatchScore1 ( self, fmap, dmap, regs, bPrint=False ) :

        points = regs[0].rpoints.copy()
        for r in regs[1:] : points = numpy.concatenate ( [points, r.rpoints], axis=0 )
        _contour.affine_transform_vertices ( points, dmap.data.ijk_to_xyz_transform )

        dvals = dmap.interpolated_values ( points, dmap.openState.xform )
        ge = numpy.greater_equal ( dvals, dmap.surface_levels[0] )
        points = numpy.compress ( ge, points, 0 )

        # put the points in the coordinate frame of fmap

        f1 = Matrix.xform_matrix ( dmap.openState.xform )
        f2 = Matrix.xform_matrix ( fmap.openState.xform.inverse() )
        f3 = fmap.data.xyz_to_ijk_transform

        tf = Matrix.multiply_matrices( f2, f1 )
        tf = Matrix.multiply_matrices( f3, tf )

        _contour.affine_transform_vertices ( points, tf )


        mm = fmap.data.matrix()
        mmr = numpy.zeros_like ( mm )

        n1, n2, n3 = fmap.data.size[0], fmap.data.size[1], fmap.data.size[2]

        for fi, fj, fk in points :
            for i in [ int(numpy.floor(fi)), int(numpy.ceil(fi)) ] :
                for j in [ int(numpy.floor(fj)), int(numpy.ceil(fj)) ] :
                    for k in [ int(numpy.floor(fk)), int(numpy.ceil(fk)) ] :
                        #if i < 0 or i >= n1 : continue
                        #if j < 0 or j >= n2 : continue
                        #if k < 0 or k >= n3 : continue
                        mmr[k,j,i] = 1.0


        #mmrd = VolumeData.Array_Grid_Data (mmr, fmap.data.origin, fmap.data.step, fmap.data.cell_angles, fmap.data.rotation)
        #gv = VolumeViewer.volume.add_data_set ( mmrd, None )
        #gv.name = 'regions_mask'
        #gv.openState.xform = fmap.openState.xform
        
        thr = dmap.surface_levels[0]
        mmf = numpy.where ( mm > thr, numpy.ones_like(mm), numpy.zeros_like(mm) )

        nz = numpy.shape ( numpy.nonzero ( mmf ) )[1]
        tvol = float(nz) * fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]

        if bPrint : print " - values %s above %f, %d, volume %.2f" % ( fmap.name, thr, nz, tvol )

        nz_int =  numpy.shape ( (mmr * mmf).nonzero () )[1]
        nz_uni =  numpy.shape ( (mmr + mmf).nonzero () )[1]

        sm_score = float(nz_int) / float (nz_uni)

        if bPrint : print " - intersection %d, union %d" % (nz_int, nz_uni)

        if bPrint : print " * shape match score %.3f" % sm_score

        return sm_score




    def ShapeMatchScore ( self, atoms, dmap, regs, bPrint=False ) :

        #fmol = fmap.mol
        #print "atoms from", fmol.name
        #points = _multiscale.get_atom_coordinates ( fmol.atoms, transformed = True )

        print "shape match of %d atoms" % len(atoms)
        points = _multiscale.get_atom_coordinates ( atoms, transformed = True )

        tfd = Matrix.xform_matrix( dmap.openState.xform.inverse() )
        #tff = Matrix.xform_matrix( fmap.openState.xform.inverse() )
        #tf = Matrix.multiply_matrices( tff, tfd )
        _contour.affine_transform_vertices( points, tfd )

        sg_fmap = VolumeData.zone_masked_grid_data ( dmap.data, points, max(3.0,dmap.data.step[0]) )

        #try : gv = VolumeViewer.volume.add_data_set ( sg_fmap, None )
        #except : gv = VolumeViewer.volume.volume_from_grid_data ( sg_fmap )
        #gv.name = 'registered_structure_region'
        #gv.openState.xform = dmap.openState.xform


        points = regs[0].rpoints.copy()
        for r in regs[1:] : points = numpy.concatenate ( [points, r.rpoints], axis=0 )
        _contour.affine_transform_vertices ( points, dmap.data.ijk_to_xyz_transform )

        sg_dmap = VolumeData.zone_masked_grid_data ( dmap.data, points, dmap.data.step[0] / 2.0 )
        #try : gv = VolumeViewer.volume.add_data_set ( sg_dmap, None )
        #except : gv = VolumeViewer.volume.volume_from_grid_data ( sg_dmap )
        #gv.name = 'regions_mask'

        regs_m = sg_dmap.matrix()
        regs_f = sg_fmap.matrix()
        regs_mz = numpy.where ( regs_m > dmap.surface_levels[0], numpy.ones_like(regs_m), numpy.zeros_like(regs_m) )
        regs_fz = numpy.where ( regs_f > dmap.surface_levels[0], numpy.ones_like(regs_f), numpy.zeros_like(regs_f) )
        
        nz = numpy.shape ( numpy.nonzero ( regs_mz ) )[1]
        print "regions %d nonzero, %d points" % (nz, len(points))

        if nz != len(points) :
            print "mask failed - %d of %d pts nonzero" % (nz, len(points))

        nz = numpy.shape ( numpy.nonzero ( regs_fz ) )[1]
        print "struct. %d nonzero, %d atoms" % (nz, len(atoms))

        nz_int =  numpy.shape ( (regs_mz * regs_fz).nonzero () )[1]
        nz_uni =  numpy.shape ( (regs_mz + regs_fz).nonzero () )[1]

        sm_score = float(nz_int) / float (nz_uni)

        if 1 or bPrint : print " - intersection %d, union %d" % (nz_int, nz_uni)

        print " * shape match score %.3f" % sm_score

        return sm_score





    def SelRegsShapeScore ( self ) :


        if len ( self.dmap.get() ) == 0 : umsg ("Please select a map first" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        if len ( self.regions_file.get() ) == 0 : umsg ("Please select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        fmap = self.SimulatedMap()
        if fmap == None:
            return


        sel_sps = SelectedRegions ( smod )
        if len(sel_sps)==0 :
            umsg ( "Please select a region" );
            return

        print "Shape score of %s to %d regions:" % (fmap.name, len(sel_sps)),
        regs = []
        for sp in sel_sps :
            print sp.region.rid,
            regs.append ( sp.region )
        print ""

        fmap.sms = self.ShapeMatchScore ( fmap.fmol.atoms, dmap, regs )
        ov, fmap.fit_score = map_overlap_and_correlation ( fmap, dmap, True )

        #self.ShapeMaskedCorr ( fmap, dmap, regs )
        #self.OptShapeScore ( fmap, dmap, regs )

        umsg ( "Shape match score: %.3f, cross-correlation: %.3f" % (fmap.sms, fmap.fit_score) )

        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        log_file = path + smod.name + "_fits_sms.txt"
        print "SMS to", log_file

        fpl = open ( log_file, 'a' )
        fpl.write ( "%s %f %f\n" % (fmap.name, fmap.sms, fmap.fit_score) )
        fpl.close ()




    def SelRegsOptimizeShapeScore ( self ) :


        if len ( self.dmap.get() ) == 0 : umsg ("Please select a map first" ); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        if len ( self.regions_file.get() ) == 0 : umsg ("Please select a regions file first" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        fmap = self.SimulatedMap()
        if fmap == None:
            return


        sel_sps = SelectedRegions ( smod )
        if len(sel_sps)==0 :
            umsg ( "Please select a region" );
            return

        print "Shape score of %s to %d regions:" % (fmap.name, len(sel_sps)),
        regs = []
        for sp in sel_sps :
            print sp.region.rid,
            regs.append ( sp.region )
        print ""

        # fmap.sms = self.ShapeMatchScore ( fmap, dmap, regs )
        # self.ShapeMaskedCorr ( fmap, dmap, regs )
        # fmap.sms = self.OptShapeScore ( fmap, dmap, regs )

        umsg ( "Optimized shape match score: %.3f" % fmap.sms )


    def SimulatedMap ( self ):

        if len ( self.struc_map.get() ) == 0 :
            self.GenStrucMap( show = False )
            if len ( self.struc_map.get() ) == 0 :
                return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : umsg ( "%s is not open" % self.struc_map.get() ); return
        return fmap


    def MaskedMap ( self ) :

        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        if len(self.struc_map.get()) == 0 : print "Please select a structure map"; return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return


        imap, f_COM, f_bRad = self.MapMaskedMapIndexes ( dmap, fmap, True )


        

    def MapMaskedMapIndexes ( self, ref_map, mask_map, bAddMap=False ) :

        f1 = ref_map.data.ijk_to_xyz_transform
        f2 = Matrix.xform_matrix( ref_map.openState.xform )
        tf = Matrix.multiply_matrices( f2, f1 )

        n1, n2, n3 = ref_map.data.size[0], ref_map.data.size[1], ref_map.data.size[2]
        m_points = VolumeData.grid_indices( (n1,n2,n3), numpy.single )  # i,j,k indices
        _contour.affine_transform_vertices( m_points, tf )

        a_vals = mask_map.interpolated_values ( m_points, chimera.Xform() )
        mm = a_vals.reshape( (n3,n2,n1) )

        thr = mask_map.surface_levels[0]
        mm = numpy.where ( mm > thr, mm, numpy.zeros_like(mm) )
        print "- in masked map, %d above %.3f" % ( numpy.shape(mm.nonzero())[1], thr )

        if bAddMap :
            mmd = VolumeData.Array_Grid_Data (mm, ref_map.data.origin, ref_map.data.step, ref_map.data.cell_angles, ref_map.data.rotation)
            gv = VolumeViewer.volume.add_data_set ( mmd, None )
            gv.name = ref_map.name + "---" + mask_map.name
            #gv.openState.xform = dmap.openState.xform

        nze = numpy.nonzero ( mm )
        print " - index values of %d non-zero points" % len ( nze[0] )

        imap = {}
        for ei, i in enumerate ( nze[0] ) :
            j = nze[1][ei]
            k = nze[2][ei]

            try : mi = imap[i]
            except : mi = {}; imap[i] = mi

            try : mij = mi[j]
            except : mij = {}; mi[j] = mij

            mij[k] = 1


        nzs = numpy.array ( [nze[2], nze[1], nze[0]] )
        points = numpy.cast[numpy.float32] ( numpy.transpose (nzs) )
        _contour.affine_transform_vertices( points, f1 )

        com = numpy.sum(points, axis=0) / len(points)
        C = chimera.Vector ( com[0], com[1], com[2] )
        comv = numpy.ones_like ( points ) * com
        points = points - comv
        bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points), 1 ) ) )

        return imap, C, bRad



    def MapIndexesInMap ( self, ref_map, mask_map ) :

        thr = mask_map.surface_levels[0]
        mm = mask_map.data.matrix()
        mm = numpy.where ( mm > thr, mm, numpy.zeros_like(mm) )

        nze = numpy.nonzero ( mm )
        nzs = numpy.array ( [nze[2], nze[1], nze[0]] )

        # the copy is needed! otherwise the _contour.afine_transform does not work for some reason
        points = numpy.cast[numpy.float32] ( numpy.transpose ( nzs ) ).copy()

        #points = numpy.zeros ( ( len(nze[0]), 3 ), numpy.float32 )
        #for ei, i in enumerate ( nze[0] ) :
        #    j = nze[1][ei]
        #    k = nze[2][ei]
        #    points[ei][0], points[ei][1], points[ei][2] = float (k), float(j), float(i)

        #print points[0]

        print "Making map indices for %s in %s" % ( mask_map.name, ref_map.name )
        print " - %d points above %.3f" % ( len(points), thr )

        # transform to index reference frame of ref_map
        f1 = mask_map.data.ijk_to_xyz_transform
        f2 = Matrix.xform_matrix ( mask_map.openState.xform )
        f3 = Matrix.xform_matrix ( ref_map.openState.xform.inverse() )
        f4 = ref_map.data.xyz_to_ijk_transform

        tf = Matrix.multiply_matrices( f2, f1 )
        tf = Matrix.multiply_matrices( f3, tf )
        tf = Matrix.multiply_matrices( f4, tf )
        _contour.affine_transform_vertices ( points, tf )

        #_contour.affine_transform_vertices ( points, f1 )
        #_contour.affine_transform_vertices ( points, f2 )
        #_contour.affine_transform_vertices ( points, f3 )

        #print points[0]

        #com = numpy.sum (points, axis=0) / len(points)
        #C = chimera.Vector ( com[0], com[1], com[2] )
        #comv = numpy.ones_like ( points ) * com
        #points_v = points - comv
        #bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points_v), 1 ) ) )

        # transform points to indexes in reference map
        # _contour.affine_transform_vertices ( points, ref_map.data.xyz_to_ijk_transform )

        imap = {}
        for fk, fj, fi in points :

            for i in [ int(numpy.floor(fi)), int(numpy.ceil(fi)) ] :
                for j in [ int(numpy.floor(fj)), int(numpy.ceil(fj)) ] :
                    for k in [ int(numpy.floor(fk)), int(numpy.ceil(fk)) ] :

                        try : mi = imap[i]
                        except : mi = {}; imap[i] = mi
                        try : mij = mi[j]
                        except : mij = {}; mi[j] = mij
                        mij[k] = 1


        return imap #, C, bRad




    def OverlappingRegions ( self, dmap, fmap, smod, hide_others = True ) :

        #imap, f_COM, f_bRad = self.MapMaskedMapIndexes ( dmap, fmap ) 
        #print " - com, brad, old way 1", f_COM, f_bRad

        imap = self.MapIndexesInMap ( dmap, fmap ) 
        # print " - com, brad", f_COM, f_bRad

        try : fmap.COM
        except : fmap.COM = fmap.mol.COM; fmap.bRad = fmap.mol.BoundRad

        p = numpy.array ( [ [fmap.COM[0],fmap.COM[1],fmap.COM[2]] ], numpy.float32 )
        _contour.affine_transform_vertices ( p, Matrix.xform_matrix( fmap.openState.xform ) )
        _contour.affine_transform_vertices ( p, Matrix.xform_matrix( dmap.openState.xform.inverse() ) )
        f_COM = chimera.Vector ( p[0][0], p[0][1], p[0][2] )
        f_bRad = fmap.bRad
        print " - com, brad", f_COM, f_bRad

        jregs = []

        for rid, r in smod.regions.iteritems() :

            try :
                if r.placed == True :
                    continue
            except :
                pass

            #f_COM = chimera.Vector ( fmap.M[0,3], fmap.M[1,3], fmap.M[2,3] )
            #if (f_COM - r.COM).length > fmap.mol.BoundRad + r.bRad : continue

            #try : r.COM
            #except : r.COM = None

            #if r.COM != None :
            #    if (f_COM - r.COM).length > f_bRad + r.bRad : continue

            # print reg_i,

            #points = r.tpoints.copy ()
            #tfd = Matrix.xform_matrix ( dmap.openState.xform )
            #tff = Matrix.xform_matrix ( fmap.openState.xform.inverse() )
            #tf = Matrix.multiply_matrices ( tff, tfd )
            #_contour.affine_transform_vertices ( points, tf )

            #sg = VolumeData.zone_masked_grid_data ( fmap.data, points, fmap.data.step[0]/2.0 )
            #regsm = sg.matrix()
            #nz = numpy.shape ( numpy.nonzero ( regsm ) )[1]
            #ov = float(nz) / float ( len(points) )

            ipoints = numpy.cast[numpy.int] ( r.rpoints )
            noverlap = 0
            for i,j,k in ipoints :
                try :
                    imap[k][j][i]
                    noverlap = noverlap + 1
                except :
                    continue

            ov = float(noverlap) / float ( len(r.rpoints) )

            # print " region %d, overlap %.3f" % (rid, ov)

            if ov > .5 :
                #print " - %d/%d - %d - %d of %d overlap, %.3f" % ( reg_i+1, len(smod.surfacePieces), r.rid, noverlap, len(r.rpoints), ov )
                #r.display = True
                jregs.append ( r )
            # else : r.display = False

            try :
                if ov > r.max_ov :
                    r.max_ov = ov
                    r.max_ov_cid = fmap.chain_id
                    r.max_ov_bioM = smod.bio_mt_at
            except :
                pass


        oregs = jregs
        
        umsg ( "Found %d regions overlapping" % len(oregs) )
        sel_sps = []

#        chimera.viewer.keyLight.specular = chimera.MaterialColor(0,0,0,1)
        for sp in smod.surfacePieces :
            try : clr = sp.region.color
            except : continue
            if oregs.count ( sp.region ) > 0 :
                sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
                sp.display = True
                sel_sps.append ( sp )
            else :
                sp.color = ( clr[0], clr[1], clr[2], 1.0 )
                sp.display = not hide_others


        chimera.selection.clearCurrent ()
        chimera.selection.addCurrent ( sel_sps )
        return jregs



    def ZeroDMap_with_FMap ( self ) :
        
        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        if len(self.struc_map.get()) == 0 : print "Please select a structure map"; return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        print "Taking %s away from %s" % ( fmap.name, dmap.name )
        rname = dmap.name + '_-_' + fmap.name

        mmc = getMod ( rname )

        if mmc == None :
            mmc = dmap.writable_copy ()
            mmc.name = rname
            print " - cloned", dmap.name
        else :
            print " - found", mmc.name

        self.ZeroOverlappingRegion ( mmc, fmap )





    def ZeroOverlappingRegion ( self, ref_map, mask_map ) :

        thr = mask_map.surface_levels[0]
        mm = mask_map.data.matrix()
        mm = numpy.where ( mm > thr, mm, numpy.zeros_like(mm) )

        nze = numpy.nonzero ( mm )
        nzs = numpy.array ( [nze[2], nze[1], nze[0]] )

        # the copy is needed! otherwise the _contour.afine_transform does not work for some reason
        points = numpy.cast[numpy.float32] ( numpy.transpose ( nzs ) ).copy()

        print " - %s - %d points above %.3f" % ( mask_map.name, len(points), thr )

        # transform to index reference frame of ref_map
        f1 = mask_map.data.ijk_to_xyz_transform
        f2 = Matrix.xform_matrix ( mask_map.openState.xform )
        f3 = Matrix.xform_matrix ( ref_map.openState.xform.inverse() )
        f4 = ref_map.data.xyz_to_ijk_transform

        tf = Matrix.multiply_matrices( f2, f1 )
        tf = Matrix.multiply_matrices( f3, tf )
        tf = Matrix.multiply_matrices( f4, tf )

        _contour.affine_transform_vertices ( points, tf )


        mm = ref_map.data.matrix()
        n1, n2, n3 = ref_map.data.size[0], ref_map.data.size[1], ref_map.data.size[2]

        for fi, fj, fk in points :

            for i in [ int(numpy.floor(fi)), int(numpy.ceil(fi)) ] :
                for j in [ int(numpy.floor(fj)), int(numpy.ceil(fj)) ] :
                    for k in [ int(numpy.floor(fk)), int(numpy.ceil(fk)) ] :
                        if i < 0 or i >= n3 : continue
                        if j < 0 or j >= n2 : continue
                        if k < 0 or k >= n1 : continue
                        mm[k,j,i] = 0.0




    def TakeDMap_with_FMap ( self ) :

        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        if len(self.struc_map.get()) == 0 : print "Please select a structure map"; return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        print "Taking densities from %s with %s" % ( dmap.name, fmap.name )
        rname = fmap.name + '_-_' + dmap.name

        #mmc = fmap.writable_copy ( require_copy = True )
        #mmc.name = rname
        #print " - cloned", fmap.name


        n1, n2, n3 = fmap.data.size[0], fmap.data.size[1], fmap.data.size[2]
        f_points = VolumeData.grid_indices( (n1,n2,n3), numpy.single )  # i,j,k indices
        _contour.affine_transform_vertices( f_points, fmap.data.ijk_to_xyz_transform )

        d_vals = dmap.interpolated_values ( f_points, fmap.openState.xform )
        df_mat = d_vals.reshape( (n3,n2,n1) )

        f_mat = fmap.data.full_matrix()
        f_mask = numpy.where ( f_mat > fmap.surface_levels[0], numpy.ones_like(f_mat), numpy.zeros_like(f_mat) )

        df_mat = df_mat * f_mask

        df_data = VolumeData.Array_Grid_Data ( df_mat, fmap.data.origin, fmap.data.step, fmap.data.cell_angles )
        
        try : df_v = VolumeViewer.volume.add_data_set ( df_data, None )
        except : df_v = VolumeViewer.volume.volume_from_grid_data ( df_data )

        df_v.name = rname
        df_v.openState.xform = fmap.openState.xform



    def SelectNonPlacedRegions ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a density map"); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return
                                         
        if len(self.regions_file.get()) == 0 : umsg ("Please select a regions file"); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        nsel = 0
        chimera.selection.clearCurrent ()
        for sp in smod.surfacePieces :
            try :
                sp.region.placed
            except :
                chimera.selection.addCurrent ( sp )
                nsel = nsel + 1

        print "%d non-placed regions" % nsel

        

    def ShowOverlappingRegions ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a density map"); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return
                                         
        if len(self.regions_file.get()) == 0 : umsg ("Please select a regions file"); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        if len(self.struc_map.get()) == 0 : umsg ("Please select a structure map"); return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : umsg ( "%s is not open" % self.struc_map.get() ); return


        oregs = self.OverlappingRegions ( dmap, fmap, smod )

        return oregs




    def FitMapToGroupsAround ( self, fmap, smod, reg, dmap, bFirst = True ) :

        bestFitScore = -1e99
        bestFitM = None
        bestFitGroup = None
        fmap.fit_score = None

        tvol = self.MapVolume ( fmap )
        bRad = -1.0 # fmap.mol.BoundRad / float(dmap.data.step[0]); # self.MapBoundingRad ( fmap )
        bRad = fmap.mol.BoundRad / float(dmap.data.step[0]);

        print "\nMaking groups around region %d - target vol %.3f, b-Rad %.3f" % ( reg.rid, tvol, bRad )
        smod.rgroups, maxDepthReached = self.GroupAroundReg ( smod, reg, tvol, bRad )
        smod.rgroups.sort()
        print " - depth reached: %d" % maxDepthReached

        if len(smod.rgroups) == 0 : umsg ( "No groups found" ); return -1e99

        nsearchgrps = min ( len(smod.rgroups), SAF_LS_NGROUPS )
        print "________________________________________________________________________"
        umsg ( "Fitting %s to %d/%d groups..." % ( fmap.name, nsearchgrps, len(smod.rgroups) ) )
        print "________________________________________________________________________"


#        chimera.viewer.keyLight.specular = chimera.MaterialColor(0,0,0,1)

        for i, dv_regs in enumerate ( smod.rgroups[0:nsearchgrps] ) :

            dv, regs = dv_regs

            umsg ( "Fitting to group %d/%d, dVolume %.4f, %d regions" % (i+1, nsearchgrps, dv, len(regs) ) )
            print " - regions:",
            for r in regs : print r.rid,
            print ""

            for sp in smod.surfacePieces :
                if regs.count ( sp.region ) > 0 :
                    sp.display = True
                    clr = sp.region.color
                    sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )

                else : sp.display = False


            points = regs[0].rpoints.copy()
            for r in regs[1:] : points = numpy.concatenate ( [points, r.rpoints], axis=0 )
            _contour.affine_transform_vertices ( points, dmap.data.ijk_to_xyz_transform )
            self.saFitMapToPoints ( fmap, points, dmap )

            print ""
            if bFirst :
                if mdlg : mdlg.Step ( 20 );

            if fmap.fit_score > bestFitScore :
                bestFitScore = fmap.fit_score
                bestFitM = fmap.M

        umsg ( "Best cross-correlation: %.4f\n\n" % ( bestFitScore ) )

        fmap.fit_score = bestFitScore
        fmap.M = bestFitM

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA




    def FitMapToRegionsAroundSel ( self ) :

        if len(self.dmap.get()) == 0 : umsg ("Please select a density map"); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return
                                         
        if len(self.regions_file.get()) == 0 : umsg ("Please select a regions file"); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return


        surfs = SelectedRegions ( smod )
        regs = []
        for sp in surfs : regs.append ( sp.region )

        if len(regs) != 1 : umsg ( "Please select 1 region" ); return


        if len(self.struc_map.get()) == 0 : umsg ("Please select a structure map"); return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : umsg ( "%s is not open" % self.struc_map.get() ); return


        t0 = time.clock()
        self.FitMapToGroupsAround ( fmap, smod, regs[0], dmap )
        took_sec = time.clock() - t0
        print "Time: %.1f sec, %.1f min" % (took_sec, took_sec/60.0)


        oregs = self.OverlappingRegions ( dmap, fmap, smod, hide_others = False )

        if mdlg : mdlg.Step ( 30 )

        if len(oregs) > 1 :
            if mdlg : mdlg.Step ( 30 )
            #jreg = JoinRegions ( smod, oregs )
            #jreg.placed = True
            #self.ReportRegionCount(len(smod.regions))
        elif len(oregs) == 1 :
            oregs[0].placed = True

        if mdlg : mdlg.Step ( 30 )




    def FitMapsToRegionsAroundSel ( self ) :

        print "_______________________________________________________________"

        if len(self.dmap.get()) == 0 : umsg ("Please select a density map"); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        if len(self.regions_file.get()) == 0 : umsg ("Please select a regions file"); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return


        sregs = SelectedRegions ( smod )        
        if len(sregs) != 1 : print "please selected 1 region"; return

        t0 = time.clock()

        if mdlg : mdlg.Step ( 10 )

#        chimera.viewer.keyLight.specular = chimera.MaterialColor(0,0,0,1)
        for sp in smod.surfacePieces :
            if sp.region == sregs[0].region :
                sp.display = True
                clr = sp.region.color
                sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
            else :
                sp.display = False

        if mdlg : mdlg.Step ( 90 )


        self.FitOpenMapsToGroupsAround ( smod, sregs[0].region, dmap )

        took_sec = time.clock() - t0
        print "Time: %.1f sec, %.1f min" % (took_sec, took_sec/60.0)



    def FitOpenMapsToGroupsAround ( self, smod, reg, dmap, bFirst=True ) :

        bestFitScore = -1e99
        bestFitMap = None
        bestFitM = None

        fmaps = []

        # TODO: Don't use "centered" to decide what maps to fit.
        for fmap in OML() :
            try : fmap.mol.centered
            except : continue
            fmap.display = False
            fmap.mol.display = False
            fmaps.append ( fmap )


        for fmap in fmaps :

            print "\n****************************************************"
            print "Fitting: ", fmap.name
            print "****************************************************\n"

            fmap.mol.display = True

            self.FitMapToGroupsAround ( fmap, smod, reg, dmap, bFirst )

            if fmap.fit_score == None :
                print " - no fits for map", fmap.name

            elif fmap.fit_score > bestFitScore :
                bestFitScore = fmap.fit_score
                bestFitMap = fmap
                bestFitM = fmap.M

            fmap.mol.display = False


        if bestFitM == None :
            print "No best fit recorded, perhaps there were not groups to start with"
            return None

        fmap = bestFitMap
        fmap.mol.display = True

        print "\n**********************************************************"
        print "Best fit score was %.4f for %s" % (bestFitScore, fmap.name)
        print "**********************************************************\n"
        fmap.M = bestFitM

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA

        oregs = self.OverlappingRegions ( dmap, fmap, smod, hide_others = False  )

        if oregs.count ( reg ) == 0 :
            print "Overlapping regions not inclusive"
            return False

        if mdlg :
            if bFirst : mdlg.Step ( 90 )
            else : mdlg.Step ( 5 )

        if len(oregs) > 1 :
            jreg = JoinRegions ( smod, oregs )
            jreg.placed = True
            self.ReportRegionCount(len(smod.regions))

            jsp = smod.rid_sp_map[jreg.rid]
            clr = jreg.color
            jsp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )

            if mdlg :
                if bFirst : mdlg.Step ( 90 )
                else : mdlg.Step ( 5 )

            jsp.display = False

        elif len(oregs) == 1 :
            oregs[0].placed = True
            smod.rid_sp_map[oregs[0].rid].display = False

        else :
            fmap.mol.display = False
            return False

        fmap.mol.display = False

        if self.strucCopyAfterFit.get() :
            self.SaveFit ( fmap )

        return True



    def FitMapsToRGroups ( self ) :

        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        if len(self.regions_file.get()) == 0 : print "Please select a regions file"; return
        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return

#        chimera.viewer.keyLight.specular = chimera.MaterialColor(0,0,0,1)
        for sp in smod.surfacePieces :
            clr = sp.region.color
            sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
            sp.display = False


        t0 = time.clock()

        while 1 :

            sp = None
            for spi in smod.surfacePieces :

                try :
                    spi.region.placed.display = False

                    clr = spi.region.color
                    spi.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
                    spi.display = True
                    if mdlg : mdlg.Step ( 5 )
                    spi.region.placed.display = True
                    if mdlg : mdlg.Step ( 10 )
                    spi.display = False

                except :
                    pass

                if spi.region.placed == False :
                    sp = spi
                    break

            if sp == None :
                print "\nAll regions have maps placed"
                break

            clr = sp.region.color
            sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
            sp.display = True

            if mdlg : mdlg.Step ( 5 )

            if self.FitOpenMapsToGroupsAround ( smod, sp.region, dmap, False ) == False :
                print "___ No map fit for region! ___", sp.region.rid, sp.region.volume
                sp.failed_fit = True

            sp.region.placed = True


        took_sec = time.clock() - t0
        print "Time: %.1f sec, %.1f min" % (took_sec, took_sec/60.0)

        for sp in smod.surfacePieces :
            try : sp.failed_fit
            except : continue
            print "Region %d failed fit and still in model" % sp.region.rid







    def MapVolume ( self, fmap ) :

        thr = fmap.surface_levels[0]
        mm = fmap.data.matrix()
        mmab = numpy.where ( mm > thr, numpy.ones_like(mm), numpy.zeros_like(mm) )
        nz = numpy.shape ( numpy.nonzero ( mmab ) )[1]
        vvol = fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]
        tvol = vvol * float(nz)
        print "%s - %d above %f, VOLUME %.3f" % (fmap.name, nz, thr, tvol)
        return tvol


    def MapBoundingRad ( self, fmap ) :
        
        fmol = fmap.mol
        points = _multiscale.get_atom_coordinates ( fmol.atoms, transformed = False )

        print "%s (%s)\n - COM: %.3f %.3f %.3f, bounding rad %.3f" % ( fmap.name, fmol.name, C[0], C[1], C[2], bRad )

        return bRad




    def FitMapToRGroups ( self ) :

        print "_______________________________________________________________"

        if len(self.dmap.get()) == 0 : umsg ("Please select a density map"); return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : umsg ( "%s is not open" % self.dmap.get() ); return

        fmap = self.SimulatedMap()
        if fmap == None : return
        tvol = self.MapVolume ( fmap )
                                         
        if len(self.regions_file.get()) == 0 : umsg ("Please select a regions file"); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s is not open" % self.regions_file.get() ); return

        print "---"

        t0 = time.clock()

        bRad = fmap.mol.BoundRad / float(dmap.data.step[0]);

        smod.rgroups = self.GroupAllRegions ( smod, tvol, bRad )
        smod.rgroups.sort()

        bestFitScore = -1e99
        #bestFitMap = None
        bestFitM = None

        print "Got %d groups..." % (len(smod.rgroups) )

        # return

        dmap_name = os.path.splitext ( dmap.name )[0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        log_f = path + "%s_fits.txt" % (dmap_name)
        print "Log file: %s" % log_f
        ffp = open ( log_f,  'w' ) ; ffp.close()


        nsearchgrps = min ( MAX_NUM_GROUPS, len(smod.rgroups) )

        for i, dv_regs in enumerate ( smod.rgroups [0:nsearchgrps] ) :

            dv, regs = dv_regs

            umsg ( "Fitting to group %d/%d, dVolume %.4f, %d regions" % (i+1, nsearchgrps, dv, len(regs) ) )
            print " - regions:",
            for r in regs : print r.rid,
            print ""

            for sp in smod.surfacePieces :
                if regs.count ( sp.region ) > 0 :
                    sp.display = True
                    clr = sp.region.color
                    sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )

                else : sp.display = False


            points = regs[0].rpoints.copy()
            for r in regs[1:] : points = numpy.concatenate ( [points, r.rpoints], axis=0 )
            _contour.affine_transform_vertices ( points, dmap.data.ijk_to_xyz_transform )


            if self.rotaSearch.get () :
                self.saFitMapToPoints_byRot ( fmap, points, dmap )
            else :
                self.saFitMapToPoints ( fmap, points, dmap )

            fmap.sms = 0.0

            fp = open ( log_f, 'a' )
            fp.write ( "%s %f %f %f" % (fmap.mol.name, fmap.fit_score, fmap.sms, dv) )
            T = fmap.M ; fp.write ( " %f %f %f %f %f %f %f %f %f %f %f %f" % (T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3]) )
            T = fmap.MI; fp.write ( " %f %f %f %f %f %f %f %f %f %f %f %f" % (T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3]) )
            for r in regs : fp.write ( " %d" % r.rid )
            fp.write ( "\n" )
            fp.close()


            if fmap.fit_score > bestFitScore :
                bestFitScore = fmap.fit_score
                bestFitM = fmap.M

        umsg ( "Best cross-correlation: %.4f\n\n" % ( bestFitScore ) )

        fmap.fit_score = bestFitScore
        fmap.M = bestFitM

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA

        took_sec = time.clock() - t0
        print "Time: %.1f sec, %.1f min" % (took_sec, took_sec/60.0)

        oregs = self.OverlappingRegions ( dmap, fmap, smod, hide_others = False  )

        self.GetFits ()



    def GetFits ( self ) :

        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        #if len(self.regions_file.get()) == 0 : print "Please select a regions file"; return
        #smod = getMod ( self.regions_file.get() )
        #if smod == None : print self.regions_file.get(), "not open"; return

        dmap_name = os.path.splitext ( dmap.name )[0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        log_f = path + "%s_fits.txt" % (dmap_name)
        print "Log file: %s" % log_f


        class ClusterEntry :
            def __init__ (self, COM, o) :
                self.COM = chimera.Vector ( COM[0], COM[1], COM[2] )
                self.o = o

        class Cluster :
            def __init__ (self) :
                self.entries = []
                self.COM = chimera.Vector (0,0,0)

            def AddEntry ( self, new_e ) :
                self.entries.append  ( [new_e.o.sms,  new_e] )

                # compute the new COM
                self.COM = chimera.Vector (0,0,0)
                for o, e in self.entries :
                    self.COM = self.COM + e.COM
                self.COM = self.COM / float ( len(self.entries) )


        try : fp = open ( log_f, 'r' )
        except :
            print "Could not open log file:", log_f
            return

        self.clusters = []

        li = 0
        while 1 :
            li = li + 1
            line = fp.readline()
            if len(line) == 0 : break

            n = line.split()

            o = aClass ()
            o.mname, o.cor, o.sms, o.dv = n[0], float(n[1]), float(n[2]), float(n[3])

            # get M
            mi = 4
            M = numpy.matrix ( [ [ float(n[mi+ 0]), float(n[mi+ 1]), float(n[mi+ 2]), float(n[mi+ 3]) ],
                                 [ float(n[mi+ 4]), float(n[mi+ 5]), float(n[mi+ 6]), float(n[mi+ 7]) ],
                                 [ float(n[mi+ 8]), float(n[mi+ 9]), float(n[mi+10]), float(n[mi+11]) ],
                                 [ 0.0, 0.0, 0.0, 1.0 ]  ] )
            # orthogonalize M
            T = M ; xf = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3], True )
            mt, mr = xf_2_M ( xf )
            o.M = mt * mr

            COM = chimera.Vector ( M[0,3], M[1,3], M[2,3] )
            #print COM

            # get MI
            mi = mi+12
            MI = numpy.matrix ( [ [ float(n[mi+ 0]), float(n[mi+ 1]), float(n[mi+ 2]), float(n[mi+ 3]) ],
                                  [ float(n[mi+ 4]), float(n[mi+ 5]), float(n[mi+ 6]), float(n[mi+ 7]) ],
                                  [ float(n[mi+ 8]), float(n[mi+ 9]), float(n[mi+10]), float(n[mi+11]) ],
                                  [ 0.0, 0.0, 0.0, 1.0 ]  ] )
            # orthogonalize MI
            T = MI ; xf = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3], True )
            mt, mr = xf_2_M ( xf )
            MI = mt * mr

            mi = mi+12

            o.regs = []
            for rmi in range ( mi, len(n) ) :
                ri = int ( n[rmi] )
                #try : o.regs.append ( smod.rid_sp_map[ri] )
                #except : pass

            # print "%s - corr %.3f, shape score %.3f, regions:" % (o.mname, o.cor, o.sms), o.regs

            cluster = None
            for cl in self.clusters :
                v = cl.COM - COM
                if v.length < 6.0 :
                    cluster = cl

            if cluster == None :
                #print "+++ Creating new cluster: ",
                cluster = Cluster()
                self.clusters.append ( cluster )

            cluster.AddEntry ( ClusterEntry (COM, o) )
            #print "- cluster COM (%f, %f, %f)" % (cluster.COM[0], cluster.COM[1], cluster.COM[2])

            # self.fits.append ( [o.sms, o] )

        fp.close()

        print "____________ %d alignments, %d clusters ____________" % ( li, len(self.clusters) )

        self.fits = []

        for cl in self.clusters :
            cl.entries.sort()
            cl.entries.reverse()
            sms, cle = cl.entries[0]
            self.fits.append ( [cle.o.cor, cle.o] )

        self.fits.sort()
        self.fits.reverse()

        log_f = path + "%s_fits_sorted.txt" % (dmap_name)
        print "Writing fits to: %s" % log_f
        lfp = open ( log_f, "w" )

        for i, sms_o in enumerate ( self.fits ) :
            sms, o = sms_o

            if i < 20 :
                print "%d - %s - correlation %.3f, dv %.3f" % (
                    i+1, o.mname, o.cor, o.dv),
                #for r in o.regs : print r.rid,
                print ""

            lfp.write ( "%d - structure: %s, cross-correlation: %.3f, dVolume: %.3f\n" % (i+1, o.mname, o.cor, o.dv) )

        print ""

        lfp.close ()

        self.fit_at = -1



    def NextFit ( self ) :

        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        self.fit_at = self.fit_at + 1
        sms, o = self.fits[self.fit_at]

        print "Fit %d, %s, corr %.3f, sms %.3f, regions" % (self.fit_at, o.mname, o.cor, o.sms),

        if len(self.regions_file.get()) == 0 : print "Please select a regions file"; return
        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return

        for r in smod.surfacePieces : r.display = False
        for r in o.regs : r.display = True; print r.rid,

        print ""

        fmol = getMod ( o.mname )
        if fmol == None : print "Fitted mol %s not open" % o.mname; return

        fmap = None
        for m in chimera.openModels.list() :
            # TODO: Don't use "centered" to decide what maps to fit.
            try : m.mol.centered
            except : continue
            if m.mol.name == o.mname :
                fmap = m
                m.display, m.mol.display = True, True
            else :
                m.display, m.mol.display = False, False

        self.struc.set ( fmap.mol.name )
        self.struc_map.set ( fmap.name )

        if fmap == None :
            print "Fitted map corresponding to %s not open" % o.mname; return

        print "Found fitted map", fmap.name                

        fmap.M = o.M
        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * o.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA



    def BestFits0 ( self ) :

        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return
        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        print " - path:", path

        if len(self.regions_file.get()) == 0 : print "Please select a regions file"; return
        smod = getMod ( self.regions_file.get() )
        if smod == None : print self.regions_file.get(), "not open"; return

        for r in smod.surfacePieces : r.display = True

        self.GetFits ()
        regs_claimed = {}
        dmap.fitted_mols = []

        for i, sms_o in enumerate ( self.fits ) :

            sms, o = sms_o

            print "%d - %s - corr %.3f, sm_corr %.3f, regions, dv %.3f" % (
                i+1, o.mname, o.cor, o.sms, o.dv),

            bRegionsClaimed = False
            for r in o.regs :
                print r.rid,
                if regs_claimed.has_key ( r.rid ) :
                    bRegionsClaimed = True
                    print "*",
                else :
                    r.display = False
                    regs_claimed[r.rid] = True

            print ""

            if bRegionsClaimed :
                print " - one or more regions claimed"
                break

            nsp = JoinRegions ( smod, o.regs )
            self.nRegions.set ( len(smod.surfacePieces) )

            fmol = getMod ( o.mname )
            if fmol == None : print " - fitted mol %s not open" % o.mname; return

            fmap = fitMap ( o.mname )
            if fmap == None : print " - fitted map not open"

            fmap.M = o.M
            tXO, tXR = xf_2_M ( dmap.openState.xform )
            T = tXO * tXR * o.M
            xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
            fmap.openState.xform = xfA
            fmap.mol.openState.xform = xfA

            if 0 :
                self.saFitMapToPoints ( fmap, nsp.tpoints, dmap )
                fmap.sms = self.ShapeMatchScore ( fmap.fmol.atoms, dmap, [nsp] )


            if self.strucCopyAfterFit.get() : self.SaveFit ( fmap )



    def PlaceBestFits ( self ) :

        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return
        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        print " - path:", path

        map_name = os.path.splitext ( dmap.name )[0]

        if len(self.regions_file.get()) == 0 : umsg ( "No regions file selected" ); return
        smod = getMod ( self.regions_file.get() )
        if smod == None : umsg ( "%s not open", self.regions_file.get() ); return

        #if len(self.regions_file.get()) == 0 : print "Please select a regions file"; return
        #smod = getMod ( self.regions_file.get() )
        #if smod == None : print self.regions_file.get(), "not open"; return
        #for r in smod.surfacePieces : r.display = False

        self.GetFits ()

        dmap.fitted_mols = []

        scores = []

        for i, sms_o in enumerate ( self.fits ) :

            if i >= int ( self.numFits.get() ) : break

            sms, o = sms_o

            print "%d/%d - %s - corr %.3f, sm_corr %.3f, dv %.3f" % (
                i+1, len(self.fits), o.mname, o.cor, o.sms, o.dv)

            fmol = getMod ( o.mname )
            if fmol == None : print " - fitted mol %s not found" % o.mname; return

            fmap = fitMap ( o.mname )
            if fmap == None : print " - fitted map for %s not found" % fmol.name; return

            fmap.M = o.M
            tXO, tXR = xf_2_M ( dmap.openState.xform )
            T = tXO * tXR * o.M
            xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
            fmap.openState.xform = xfA
            fmap.mol.openState.xform = xfA

            #if self.strucCopyAfterFit.get() :
            self.SaveFit ( fmap )
            

            if 0 :
                oregs = self.OverlappingRegions ( dmap, fmap, smod )
                sms = self.ShapeMatchScore ( fmap.mol.atoms, dmap, oregs )
                print " ---- sms %.4f --- " % (sms)
                scores.append ( sms )

            # if joinRegs : JoinRegions ( smod, regs )


        if 0 :
            log_f = path + map_name + "_fits_acc.txt"
            print "\nAccuracies log file:", log_f

            try : fp = open ( log_f, "a" )
            except : print " - could not open log file"; return
            fp.write ( "%f %f %f " % ( min(scores), max(scores), sum(scores)/float(len(scores))) )
            for sm_score in scores : fp.write ( "%f " % sm_score )
            fp.write ( "\n" )
            fp.close ()
            




    def AlignToSel ( self ) :

        if len(self.struc.get()) == 0 : print "Please select a structure first"; return
        m = getMod ( self.struc.get() )
        if m == None : print self.struc.get(), "not open"; return

        m_cid = None
        to_cid = None
        m_to = None

        sela = chimera.selection.currentContents()[0]
        for sel_at in sela :
            try : cid = sel_at.residue.id.chainId
            except : pass

            if m_cid == None and sel_at.molecule == m :
                m_cid = cid
                print "Found selected atom in chain %s in %s" % (cid, m.name)
            elif m_to == None and sel_at.molecule != m :
                m_to = sel_at.molecule
                to_cid = cid
                print "Found selected atom in chain %s in %s" % (cid, m_to.name)

        if m_cid == None or m_to == None or to_cid == None :
            print "Please select at least two atoms in the corresponding structures to align"
            return
        
        print "Aligning %s chain %s to %s chain %s" % (m.name, m_cid, m_to.name, to_cid)

        m_seq = m.sequences ( asDict=True ) [m_cid]
        to_seq = m_to.sequences ( asDict=True ) [to_cid]

        print " -- %d residues -- %d residues" % ( len(m_seq.residues), len(to_seq.residues) )

        m_ats, to_ats = AlignChains ( m_seq, to_seq )
        xf, rmsd = chimera.match.matchAtoms ( to_ats, m_ats )

        m_points = _multiscale.get_atom_coordinates ( m_ats, transformed = True )
        f_points = _multiscale.get_atom_coordinates ( to_ats, transformed = True )
        vss = numpy.square ( numpy.subtract ( m_points, f_points ) )
        sums = numpy.sum ( numpy.sum ( vss, axis=1 ) )
        armsd = numpy.sqrt ( sums / float ( len(m_points) ) )

        print " - %d aligned - RMSD: %.4f, RMSD as placed: %.4f" % ( len(m_ats), rmsd, armsd )

        mxf = m_to.openState.xform
        mxf.multiply ( xf )
        m.openState.xform = mxf


        fmap = None
        for om in chimera.openModels.list() :
            try : om_mol = om.mol
            except : continue
            if om_mol == m : fmap = om; break

        if fmap == None :
            print " - no map found for struc"
            return

        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        print ""
        umsg ( "Fitting %s and %s into %s" % ( fmap.name, m.name, dmap.name ) )

        fxf = m.openState.xform
        dxf = dmap.openState.xform

        mxf = dxf.inverse()
        mxf.multiply ( fxf ) # fxf = dxf * mxf

        tXO, tXR = xf_2_M ( mxf )
        fmap.M = tXO * tXR

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA

        fmap.M0 = fmap.M



    def FitRMSD ( self ) :

        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        print "Map: %s" % dmap.name
        try : print " - %d fitted molecules" % len(dmap.fitted_mols)
        except : print " - no fitted molecules found"; return


        if len(self.struc.get()) == 0 : print "Please select a structure first"; return
        m = getMod ( self.struc.get() )
        if m == None : print self.struc.get(), "not open"; return

        # try : m.ch_colors
        # except : m.ch_colors = RandColorChains ( m )
        # print "%s - %d chains -" % ( m.name, len(m.ch_colors) ), m.ch_colors.keys()

        print " - aligning %s (%d chains) to %d fit mols" % (
            m.name, len(m.sequences()), len(dmap.fitted_mols) )


        mseqs = m.sequences()
        fmol_msi = {}
        m_atoms = []
        f_atoms = []

        for msi, m_seq in enumerate ( mseqs ) :

            min_fmi = None
            min_rmsd = 1e99
            min_m_atoms = None
            min_f_atoms = None

            print "Chain %s - %d residues" % ( m_seq.chain, len(m_seq.residues) )

            for fmi, fm in enumerate ( dmap.fitted_mols ) :

                try : fmol_msi[fmi]; continue
                except : pass

                f_seq = fm.sequences()[0]
                print " - %d/%d %s, %d residues" % (
                    fmi+1, len(dmap.fitted_mols), dmap.fitted_mols[fmi].name, len(f_seq.residues) ),

                m_ats, f_ats = AlignChains ( m_seq, f_seq )
                xf, rmsd = chimera.match.matchAtoms ( m_ats, f_ats )
                m_points = _multiscale.get_atom_coordinates ( m_ats, transformed = True )
                f_points = _multiscale.get_atom_coordinates ( f_ats, transformed = True )
                vss = numpy.square ( numpy.subtract ( m_points, f_points ) )
                sums = numpy.sum ( numpy.sum ( vss, axis=1 ) )
                armsd = numpy.sqrt ( sums / float ( len(m_points) ) )

                print " %d aligned - RMSD: %.4f, RMSD as placed: %.4f" % (
                    len(m_ats), rmsd, armsd )

                if armsd < min_rmsd :
                    min_rmsd = armsd
                    min_fmi = fmi
                    min_m_atoms = m_ats
                    min_f_atoms = f_ats

            if min_fmi != None :
                print " - min is %d - %s" % (min_fmi, dmap.fitted_mols[min_fmi].name)
                fmol_msi[min_fmi] = msi
                m_atoms = m_atoms + min_m_atoms
                f_atoms = f_atoms + min_f_atoms
            else :
                print " - no maps left to align!"

        print ""
        
        # m.chain_fitmol = {}
        for fmi, msi in fmol_msi.iteritems () :
            print "Chain %s -- fit mol %s" % ( mseqs[msi].chain, dmap.fitted_mols[fmi].name )
            # m.chain_fitmol [ mseqs[msi].chain ] = dmap.fitted_mols[fmi]


        print ""

        xf, rmsd = chimera.match.matchAtoms ( m_atoms, f_atoms )
        m_points = _multiscale.get_atom_coordinates ( m_atoms, transformed = True )
        f_points = _multiscale.get_atom_coordinates ( f_atoms, transformed = True )
        vss = numpy.square ( numpy.subtract ( m_points, f_points ) )
        sums = numpy.sum ( numpy.sum ( vss, axis=1 ) )
        armsd = numpy.sqrt ( sums / float ( len(m_points) ) )

        print "Total %d residues aligned - RMSD: %.4f, RMSD as placed: %.4f" % (
            len(m_atoms), rmsd, armsd )







# AlignChains ( f11.sequences()[0], mm.sequences()[4] )

def AlignChains ( ref, match ) :

    import MatchMaker
    import MatchMaker.prefs
    import chimera.misc

    gapOpen = MatchMaker.prefs.defaults[MatchMaker.prefs.GAP_OPEN]
    gapExtend = MatchMaker.prefs.defaults[MatchMaker.prefs.GAP_EXTEND]
    ksdsspCache = set()
    ssFraction = MatchMaker.prefs.defaults[MatchMaker.prefs.SS_MIXTURE]
    ssMatrix = MatchMaker.prefs.defaults[MatchMaker.prefs.SS_SCORES]
    alignKw = { 'ssFraction': ssFraction, 'ssMatrix': ssMatrix, 'computeSS': False }

    # print "Aligning %d res to %d res" % ( len(ref), len(match) )

    score, s1, s2 = MatchMaker.align( ref, match, 'BLOSUM-62', "nw", gapOpen, gapExtend, ksdsspCache, **alignKw)

    refAtoms, matchAtoms = [], []

    for i in range( len(s1) ) :
        if s1[i] == "." or s2[i] == ".": continue
        refRes = s1.residues[s1.gapped2ungapped(i)]
        matchRes = s2.residues[s2.gapped2ungapped(i)]
        refAtom = chimera.misc.principalAtom(refRes)
        if not refAtom: continue
        matchAtom = chimera.misc.principalAtom(matchRes)
        if not matchAtom: continue
        if refAtom.name != matchAtom.name:
            # nucleic P-only trace vs. full nucleic
            if refAtom.name != "P":
                try : refAtom = refAtom.residue.atomsMap["P"][0]
                except KeyError: continue
            else:
                try : matchAtom = matchAtom.residue.atomsMap["P"][0]
                except KeyError: continue
        refAtoms.append(refAtom)
        matchAtoms.append(matchAtom)

    # print " - aligned %d atoms" % len (refAtoms)

    if len ( refAtoms ) < 3 :
        print "too few atoms to perform 3D alignment"
        return None

    # return chimera.match.matchAtoms ( refAtoms, matchAtoms )

    return refAtoms, matchAtoms




    def FitRMSDNoRef ( self ) :

        if len(self.dmap.get()) == 0 : print "Please select a density map"; return
        dmap = getMod ( self.dmap.get() )
        if dmap == None : print self.dmap.get(), "not found"; return

        print "Map: %s" % dmap.name
        try : print " - %d fitted molecules" % len(dmap.fitted_mols)
        except : print " - no fitted molecules found"; return


        if len(self.struc.get()) == 0 : print "Please select a structure first"; return
        m = getMod ( self.struc.get() )
        if m == None : print self.struc.get(), "not open"; return

        # try : m.ch_colors
        # except : m.ch_colors = RandColorChains ( m )
        # print "%s - %d chains -" % ( m.name, len(m.ch_colors) ), m.ch_colors.keys()

        print " - aligning %s (%d chains) to %d fit mols" % (
            m.name, len(m.sequences()), len(dmap.fitted_mols) )


        mseqs = m.sequences()
        cmap = []
        for msi, m_seq in enumerate ( mseqs ) :

            min_fid = None
            min_rmsd = 1e99
            min_txf = None
            cmap.append ( '?' )

            for fmi, fm in enumerate ( dmap.fitted_mols ) :

                bIncluded = False
                for cfmi in cmap :
                    if cfmi == fmi : bIncluded = True; break
                if bIncluded : continue

                cmap[msi] = fmi
                # print cmap,

                mats,fats = [],[]
                for i in range ( len(cmap) ) :
                    cid = cmap[i]
                    # print cid, fmols[i].name,
                    for ca_posi, cat in ch_cas[cid].iteritems() :
                        mats.append ( cat.coord() )
                        fats.append ( f_ch_cas[i][ca_posi].coord() )

                # print "matching %d to %d ca atoms" % ( len(fats), len(mats) )
                # txf, rmsd = chimera.match.matchAtoms(fats, mats)
                txf, rmsd = chimera.match.matchPositions(fats, mats)
                # print "- rmsd:", rmsd
                if rmsd < min_rmsd :
                    min_rmsd = rmsd
                    min_cid = chains[i2]
                    min_txf = txf

            print "- pos %d is %s, rmsd %f" % (i1, min_cid, min_rmsd),
            cmap[i1] = min_cid
            print cmap
            # return

            for at in m.atoms :
                c = min_txf.apply ( at.coord() )
                at.setCoord ( c )

        print cmap







def volume_segmentation_dialog ( create=False ) :

  from chimera import dialogs
  return dialogs.find ( Volume_Segmentation_Dialog.name, create=create )


def show_volume_segmentation_dialog ():

    from chimera import dialogs
    d = volume_segmentation_dialog ( create = True )
    # Avoid transient dialog resizing when created and mapped for first time.
    d.toplevel_widget.update_idletasks ()
    d.enter()

    r = int(chimera.viewer.windowSize[0])
    d._toplevel.geometry( "+%d+20" % (r+70) )

    from Accelerators import add_accelerator
    add_accelerator('gg', 'Group regions', d.JoinSelRegs )
    add_accelerator('uu', 'Ungroup regions', d.UngroupSelRegs )
    add_accelerator('rh', 'Hide regions', d.RegSurfsHide )
    add_accelerator('rs', 'Hide regions', d.RegSurfsShow )
    add_accelerator('dd', 'Delete regions', d.DelSelRegs )
    print "gg - groups regions"
    print "uu - ungroup regions"

    return d



# -----------------------------------------------------------------------------
#
from chimera import dialogs
dialogs.register (Volume_Segmentation_Dialog.name, Volume_Segmentation_Dialog, replace = 1)




def RemoveRegionConnections ( smod, reg ) :

    if len(reg.cregs) == 0 :
        try : reg_cons = smod.rcons[reg]
        except : reg_cons = None

        if reg_cons :
            for rc, o in reg_cons.iteritems() :
                bcons = smod.rcons[rc]
                bcons.__delitem__ ( reg )

        try : smod.rcons.__delitem__ ( reg )
        except : pass

    else :
        for creg in reg.cregs :
            RemoveRegionConnections ( smod, creg )
        

def DeleteRegion ( smod, reg ) :

    del smod.regions [ reg.rid ]
    sp = smod.rid_sp_map [ reg.rid ]
    del smod.rid_sp_map [ reg.rid ]
    smod.removePiece ( sp )
    RemoveRegionConnections ( smod, reg )



def JoinRegions ( smod, regs ) :

    nreg = aClass ()
    nreg.rpoints = None
    nreg.volume = 0.0
    nreg.map_inds = []
    nreg.cregs = regs
    nreg.preg = None

    nreg.rid = max(smod.regions.keys()) + 1
    smod.regions[nreg.rid] = nreg

    for reg in regs :

        if nreg.rpoints == None : nreg.rpoints = reg.rpoints.copy()
        else : nreg.rpoints = numpy.concatenate ( (nreg.rpoints, reg.rpoints) )

        nreg.volume = nreg.volume + reg.volume
        nreg.map_ind = reg.map_ind
        nreg.map_inds = nreg.map_inds + reg.map_inds
        nreg.color = reg.color

        reg.preg = nreg

        rid = reg.rid

        del smod.regions[rid]
        if rid in smod.rid_sp_map:
            # Won't have surface piece if MaximumRegionsToDisplay() exceeded.
            sp = smod.rid_sp_map[rid]
            del smod.rid_sp_map[rid]
            smod.removePiece ( sp )

    tpoints = numpy.copy ( nreg.rpoints )
    _contour.affine_transform_vertices ( tpoints, smod.seg_map.data.ijk_to_xyz_transform )
    com = numpy.sum(tpoints, axis=0) / len(tpoints)
    nreg.COM = chimera.Vector ( com[0], com[1], com[2] )
    comv = numpy.ones_like ( tpoints ) * com
    points = tpoints - comv
    nreg.bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points), 1 ) ) )
    nreg.placed = False

    vt = MultiScale.surface.surface_points ( nreg.rpoints, 1.0, 0.1, .25, 5 )
    _contour.affine_transform_vertices ( vt[0], smod.seg_map.data.ijk_to_xyz_transform )
    nsp = makeRegionSurface ( smod, vt[0], vt[1], nreg )

    return nreg


def MakeRegionFromCRegs ( smod, nreg ) :

    nreg.volume = 0.0
    nreg.map_inds = []
    nreg.rpoints = None

    for reg in nreg.cregs :

        if nreg.rpoints == None : nreg.rpoints = reg.rpoints.copy()
        else : nreg.rpoints = numpy.concatenate ( (nreg.rpoints, reg.rpoints) )

        nreg.volume = nreg.volume + reg.volume
        nreg.map_ind = reg.map_ind
        nreg.map_inds = nreg.map_inds + reg.map_inds
        nreg.color = reg.color


    tpoints = numpy.copy ( nreg.rpoints )
    _contour.affine_transform_vertices ( tpoints, smod.seg_map.data.ijk_to_xyz_transform )
    com = numpy.sum(tpoints, axis=0) / len(tpoints)
    nreg.COM = chimera.Vector ( com[0], com[1], com[2] )
    comv = numpy.ones_like ( tpoints ) * com
    points = tpoints - comv
    nreg.bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points), 1 ) ) )
    nreg.placed = False






def ComputeConnections ( smod ) :

    print "Computing connections...",

    smod.cons = {}

    for r1, r1_cons_ in smod.rcons.iteritems() :

        r1_parent = r1
        while r1_parent.preg != None : r1_parent = r1_parent.preg

        for r2, o in r1_cons_.iteritems() :

            r2_parent = r2
            while r2_parent.preg != None : r2_parent = r2_parent.preg

            try : r1p_cons = smod.cons[r1_parent]
            except :
                r1p_cons = {}
                smod.cons[r1_parent] = r1p_cons

            try : no = r1p_cons[r2_parent]
            except :
                no = aClass()
                no.N, no.D = 0, 0.0
                r1p_cons[r2_parent] = no

            no.D, no.N = no.D + o.D, no.N + o.N

            try : r2p_cons = smod.cons[r2_parent]
            except :
                r2p_cons = {}
                smod.cons[r2_parent] = r2p_cons
            
            r2p_cons[r1_parent] = no

    print "done"
    


def getMods ( tag ) :
    if len(tag) == 0 : return None
    taggedMols = []
    mlist = chimera.openModels.list ()
    for mol in mlist :
        if mol.name[0:len(tag)] == tag :
            taggedMols.append ( mol )
    return taggedMols


def getMod ( name ) :
    mlist = chimera.openModels.list ()
    for mol in mlist :
        if mol.name == name :
            return mol
    return None


def fitMap(name):
    for m in chimera.openModels.list() :
        if hasattr(m, 'mol') and m.mol.name == name :
            if not hasattr(m.mol, 'centered'):
                centerMol ( m.mol )
            return m
    return None


def prAxes ( points ) :

    com = numpy.sum(points, axis=0) / len(points)
    C = chimera.Vector ( com[0], com[1], com[2] )

    comv = numpy.ones_like ( points ) * com
    points = points - comv

    i = numpy.matrix ( [[1,0,0], [0,1,0], [0,0,1]] )
    ii = i * numpy.sum ( numpy.multiply ( points, points ) )
    p_t = numpy.transpose(points)
    td = numpy.tensordot ( points, p_t, axes=[0,1] )

    I0 = ii - td

    try :
        U, S, V = numpy.linalg.svd( I0 )
    except :
        print "- error computing SVD - prob. singular matrix"
        return []

    #U[0,0] = U[0,0] * -1.0
    #U[1,0] = U[1,0] * -1.0
    #U[2,0] = U[2,0] * -1.0

    #U[0,2] = U[0,2] * -1.0
    #U[1,2] = U[1,2] * -1.0
    #U[2,2] = U[2,2] * -1.0

    return [C, U, S, V]


def molApplyT ( mol, T ) :

    xf = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3], True )
    # print xf

    mol.COM = chimera.Vector (0,0,0)

    for at in mol.atoms :
        c = xf.apply ( at.coord() )
        at.setCoord ( c )
        mol.COM = mol.COM + c.toVector()

    mol.COM = mol.COM / float ( len(mol.atoms) )

#
# Change atom coordinates so the center of mass is at the origin and
# the principal axes are x, y and z.  Make a compensating transformation
# of the molecule coordinate system so the molecule does not move in the
# graphics window.
#
def centerMol ( fmol ):

        umsg ( "Centering %s" % fmol.name )

#        fmol.ch_colors = RandColorChains ( fmol )
#        fmol.color = fmol.atoms[0].color

        # for at in fmol.atoms : at.mass = at.element.mass

        points = _multiscale.get_atom_coordinates ( fmol.atoms, transformed = False )
        fmol.COM, fmol.U, fmol.S, fmol.V = prAxes ( points )

        # move COM to origin and align pr. axes with XYZ
        tAO = numpy.matrix ( [
            [ 1, 0, 0, -fmol.COM[0] ],
            [ 0, 1, 0, -fmol.COM[1] ],
            [ 0, 0, 1, -fmol.COM[2] ],
            [ 0, 0, 0,           1 ]  ] )

        tAR = numpy.matrix ( [
            [ fmol.V[0,0], fmol.V[0,1], fmol.V[0,2], 0 ],
            [ fmol.V[1,0], fmol.V[1,1], fmol.V[1,2], 0 ],
            [ fmol.V[2,0], fmol.V[2,1], fmol.V[2,2], 0 ],
            [           0,           0,           0, 1 ]  ] )

        #print tAO
        #print tAR
        #print tAR * tAO

        molApplyT ( fmol, tAO )
        print " - new COM:", fmol.COM

        molApplyT ( fmol, tAR )
        print " - new COM:", fmol.COM

        # Adjust coordinate system so molecule does not appear to move.
        tf = Matrix.invert_matrix((tAR*tAO).tolist()[:3])
        fmol.openState.localXform(Matrix.chimera_xform(tf))

        fmol.mT = numpy.matrix ( [
            [ 1, 0, 0, 0 ],
            [ 0, 1, 0, 0 ],
            [ 0, 0, 1, 0 ],
            [ 0, 0, 0, 1 ]  ] )

        fmol.mR = numpy.matrix ( [
            [ 1, 0, 0, 0 ],
            [ 0, 1, 0, 0 ],
            [ 0, 0, 1, 0 ],
            [ 0, 0, 0, 1 ]  ] )

        fmol.M = fmol.mT * fmol.mR

        #com = numpy.sum(points, axis=0) / len(points)
        #comv = numpy.ones_like ( points ) * com
        #points0 = points - comv

        points = _multiscale.get_atom_coordinates ( fmol.atoms, transformed = False )
        fmol.COM, fmol.U, fmol.S, fmol.V = prAxes ( points )
        # print fmol.U

        ppoints = points * fmol.U
        fmol.BoundRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (ppoints), 1 ) ) )
        fmol.Extents = numpy.asarray ( numpy.max ( numpy.abs ( ppoints ), 0 ) )[0]

        fmol.Extents[0] = fmol.Extents[0] + 5.0
        fmol.Extents[1] = fmol.Extents[1] + 5.0
        fmol.Extents[2] = fmol.Extents[2] + 5.0

        #print " - b-rad %.3f," % (fmol.BoundRad)
        #print " - extents: %.2f %.2f %.2f" % (fmol.Extents[0], fmol.Extents[1], fmol.Extents[2])

        fmol.centered = True

        umsg ( "Centered %s (radius %.2fA, extents %.2fA %.2fA %.2fA)" % (
            fmol.name, fmol.BoundRad, fmol.Extents[0], fmol.Extents[1], fmol.Extents[2] ) )



def NumberFromName ( name, tag ) :

    #name = os.path.splitext ( name ) [0]
    if name.rfind('.mrc') >= 0 : name = name [ : name.rfind('.mrc') ]
    ts = name.split ("_")
    num = None

    for t in ts :
        if t[0:len(tag)] == tag :
            try :
                num = float ( t[len(tag):] )
                break
            except :
                continue

    return num


def SelectedRegions ( smod ) :

    sela = chimera.selection.currentContents()[0]
    sregs = []
    for sp in smod.surfacePieces :
        if sela.count ( sp ) > 0 :
            try : sp.region
            except : continue
            sregs.append ( sp )

    # print " - %d selected surfaces" % len ( selsurfs )
    status ( "%d selected regions in %s" % ( len(sregs), smod.name ) )

    return sregs


def xf_2_M (xf) :

    X = ( numpy.matrix (xf.getOpenGLMatrix()) ).reshape([4,4]).transpose()

    tXO = numpy.matrix ( [
        [ 1, 0, 0, X[0,3] ],
        [ 0, 1, 0, X[1,3] ],
        [ 0, 0, 1, X[2,3] ],
        [ 0, 0, 0,      1 ]  ] )

    tXR = numpy.matrix ( [
        [ X[0,0], X[0,1], X[0,2], 0 ],
        [ X[1,0], X[1,1], X[1,2], 0 ],
        [ X[2,0], X[2,1], X[2,2], 0 ],
        [      0,      0,      0, 1 ]  ] )

    return [tXO, tXR]


def am_2_M ( X ) :

    M = numpy.matrix ( [
        [ X[0,0], X[0,1], X[0,2], X[0,3] ],
        [ X[1,0], X[1,1], X[1,2], X[1,3] ],
        [ X[2,0], X[2,1], X[2,2], X[2,3] ],
        [      0,      0,      0,      1 ]  ] )

    return M


def dx_2_M ( X ) :

    M = numpy.matrix ( [
        [ X[0][0], X[0][1], X[0][2], X[0][3] ],
        [ X[1][0], X[1][1], X[1][2], X[1][3] ],
        [ X[2][0], X[2][1], X[2][2], X[2][3] ],
        [      0,      0,      0,      1 ]  ] )

    return M



def CopyMol ( mol ) :

    nmol = chimera.Molecule()
    nmol.name = mol.name

    aMap = dict()
    clr = ( rand(), rand(), rand() )

    for res in mol.residues :
        nres = nmol.newResidue (res.type, chimera.MolResId(res.id.chainId, res.id.position))
        # print "New res: %s %d" % (nres.id.chainId, nres.id.position)
        for at in res.atoms :
            nat = nmol.newAtom (at.name, chimera.Element(at.element.number))
            aMap[at] = nat
            nres.addAtom( nat )
            nat.setCoord ( at.coord() )
            nat.drawMode = nat.Sphere
            nat.color = chimera.MaterialColor( clr[0], clr[1], clr[2], 1.0 )
            nat.display = True

        nres.isHelix = res.isHelix
        nres.isHet = res.isHet
        nres.isSheet = res.isSheet
        nres.isStrand = res.isStrand
        nres.isTurn = res.isTurn
        nres.ribbonDisplay = True
        nres.ribbonDrawMode = 2
        nres.ribbonColor = chimera.MaterialColor( clr[0], clr[1], clr[2], 1.0 );

    for bond in mol.bonds :
        nb = nmol.newBond ( aMap[bond.atoms[0]], aMap[bond.atoms[1]] )
        nb.display = True

    return nmol



def copyMolChain ( nmol, mol, cid, new_cid, xf, clr ) :

    if nmol == None :
        nmol = chimera.Molecule()
        nmol.name = "complex"

    print "Copying chain %s (%s) to %s" % (cid, mol.name, new_cid)
    aMap = dict()
    for res in mol.residues :
        if res.id.chainId == cid :
            nres = nmol.newResidue(res.type,
                    chimera.MolResId(new_cid, res.id.position))
            # print "New res: %s %d" % (nres.id.chainId, nres.id.position)
            for at in res.atoms :
                nat = nmol.newAtom (at.name, chimera.Element(at.element.number))
                aMap[at] = nat
                nres.addAtom( nat )
                if xf : nat.setCoord ( xf.apply( at.coord() )  )
                else : nat.setCoord ( at.coord() )
                nat.drawMode = nat.Sphere
                nat.color = chimera.MaterialColor( clr[0], clr[1], clr[2], 1.0 )
                nat.display = True

            nres.isHelix = res.isHelix
            nres.isHet = res.isHet
            nres.isSheet = res.isSheet
            nres.isStrand = res.isStrand
            nres.isTurn = res.isTurn
            nres.ribbonDisplay = True
            nres.ribbonDrawMode = 2
            nres.ribbonColor = chimera.MaterialColor( clr[0], clr[1], clr[2], 1.0 );

    for bond in mol.bonds :
        if (bond.atoms[0].residue.id.chainId == cid and
            bond.atoms[1].residue.id.chainId == cid) :
            nb = nmol.newBond ( aMap[bond.atoms[0]], aMap[bond.atoms[1]] )
            nb.display = True

    return nmol



def AxesMod ( COM=[0,0,0], U=None, Extents=[30,30,30], rad=1.0, f=1.0,
              alignTo = None ) :

    mol = _surface.SurfaceModel()
    chimera.openModels.add([mol], sameAs = alignTo)

    axes = AddAxes ( rad, Extents[0]*f, Extents[1]*f, Extents[2]*f, 1.0, mol )
    axes.name = "Axes"

    if U != None :
        R = numpy.array([
            [  U[0,0], U[0,1], U[0,2], 0.0    ],
            [  U[1,0], U[1,1], U[1,2], 0.0    ],
            [  U[2,0], U[2,1], U[2,2], 0.0    ]  ] )

        T = numpy.array([
            [  1.0, 0.0, 0.0, COM[0]   ],
            [  0.0, 1.0, 0.0, COM[1]   ],
            [  0.0, 0.0, 1.0, COM[2]   ]  ] )

        M = Matrix.multiply_matrices ( T, R )

        ps = []
        for p in axes.surfacePieces :
            v, t = numpy.copy(p.geometry[0]), numpy.copy(p.geometry[1])
            ps.append ( [v,t,p.color] )
            axes.removePiece ( p )

        for p in ps :
            _contour.affine_transform_vertices( p[0], M )
            axes.addPiece ( p[0], p[1], p[2] )

        clr = ( rand()*.7, rand()*.7, rand()*.7, 1.0 )
        # for p in axes.surfacePieces : p.color = clr

    #for g in axes.surfacePieces :
    #    g.initial_v = numpy.copy ( g.geometry[0] )

    return axes


def AxesModCOM ( COM=[0,0,0], Extents=[10,10,10], rad=0.3, alignTo = None ) :

    mol = _surface.SurfaceModel()
    chimera.openModels.add([mol], sameAs = alignTo)

    axes = AddAxes ( rad, Extents[0], Extents[1], Extents[2], 1.0, mol )
    axes.name = "Axes"

    T = numpy.array([
        [  1.0, 0.0, 0.0, COM[0]   ],
        [  0.0, 1.0, 0.0, COM[1]   ],
        [  0.0, 0.0, 1.0, COM[2]   ]  ] )

    ps = []
    for p in axes.surfacePieces :
        v, t = numpy.copy(p.geometry[0]), numpy.copy(p.geometry[1])
        ps.append ( [v,t,p.color] )
        axes.removePiece ( p )

    for p in ps :
        _contour.affine_transform_vertices( p[0], T )
        axes.addPiece ( p[0], p[1], p[2] )

    return axes




def AddArrow2 ( pos, v, d, clr=(0,1,1,1), rad=0.2, mol=None ) :

    xf = AlignXf ( pos, v )
    mol = CylinderMesh2 (rad, rad, d-(rad*3), 10, clr, xf, mol )

    xf = AlignXf ( pos+(v*(d-(rad*3))), v )
    mol = CylinderMesh2 (rad*3, 0.01, rad*3, 10, clr, xf, mol )

    return mol


def AddAxes ( rad, lX, lY, lZ, cF, mol ) :

    pos = chimera.Vector(0,0,0)
    mol = AddArrow2 ( pos, chimera.Vector(1,0,0), lX, (cF,0,0,1), rad, mol )
    mol = AddArrow2 ( pos, chimera.Vector(0,1,0), lY, (0,cF,0,1), rad, mol )
    mol = AddArrow2 ( pos, chimera.Vector(0,0,1), lZ, (0,0,cF,1), rad, mol )
    mol.name = "XYZ (RGB) Axes"

    return mol



def AlignXf ( pos, v ) :
    Z = v
    Z.normalize()
    dZ = chimera.Vector( random.random(), random.random(), random.random() )
    dZ.normalize()
    X = chimera.cross ( Z, dZ )
    X.normalize ()
    Y = chimera.cross ( Z, X )
    Y.normalize ()

    xf = chimera.Xform.xform (
        X.x, Y.x, Z.x, pos[0],
        X.y, Y.y, Z.y, pos[1],
        X.z, Y.z, Z.z, pos[2] )

    #xf3 = chimera.Xform.xform (
    #    d, 0, 0, 0,
    #    0, d, 0, 0,
    #    0, 0, d, 0 )
    #print xf3

    return xf


def Quad2Tri ( vi ) :
    t1 = (vi[0], vi[1], vi[2])
    t2 = (vi[0], vi[2], vi[3])
    return t1, t2


def CylinderMesh2 (r1, r2, Length, div, color, xf, mol) :

    v = None
    vi = []

    # print "CylinderMesh:", div

    at = 0
    for psi_i in range(div) :

        psi = float(psi_i) * 360.0/float(div)

        #print "%.0f(%d)(%d)" % (psi, at, at-l),
        x1 = r1 * numpy.sin(psi * numpy.pi/180)
        y1 = r1 * numpy.cos(psi * numpy.pi/180)

        x2 = r2 * numpy.sin(psi * numpy.pi/180)
        y2 = r2 * numpy.cos(psi * numpy.pi/180)

        p = chimera.Point ( x1,y1,0 );
        if xf : p = xf.apply ( p )

        if psi_i == 0 :
            v = numpy.array( [ [p[0], p[1], p[2]], ], numpy.float32 )
        else :
            pt1 = numpy.array( [ [p[0], p[1], p[2]], ], numpy.float32 )
            v = numpy.concatenate ( [v, pt1] )

        p = chimera.Point ( x2,y2,Length );
        if xf : p = xf.apply ( p )
        pt2 = numpy.array( [ [p[0], p[1], p[2]], ], numpy.float32 )
        v = numpy.concatenate ( [v, pt2] )

        at = at + 2                

        if psi_i == 0 :
            pass
        else :
            tris = Quad2Tri ( [at-3, at-1, at-2, at-4] )
            vi.extend(tris)

        if psi_i == div-1 :
            tris = Quad2Tri ( [at-1, 1, 0, at-2] )
            vi.extend(tris)


    p = chimera.Point ( 0,0,0 );
    if xf : p = xf.apply ( p )
    pt1 = numpy.array( [ [p[0], p[1], p[2]], ], numpy.float32 )
    v = numpy.concatenate ( [v, pt1] )

    p = chimera.Point ( 0,0,Length );
    if xf : p = xf.apply ( p )
    pt1 = numpy.array( [ [p[0], p[1], p[2]], ], numpy.float32 )
    v = numpy.concatenate ( [v, pt1] )

    sph = mol.addPiece ( v, vi, color )
    return mol



def SphereMesh ( r, div, pos = chimera.Vector(0,0,0) ) :

    v = numpy.array( [ [0+pos.x,0+pos.y,r+pos.z], ], numpy.float32 )
    vi = ()

    at = 1
    l = int ( numpy.ceil (float(div)*3.0/2.0) )
    if div < 10 : l = div*2
    lat = 0

    for phi_i in range(div) :

        phi = 90.0 - ( float(phi_i+1) * 180.0/float(div+1) )
        #print "%.2f: " % phi,
        z = r * numpy.sin(phi * numpy.pi/180)
        s = r * numpy.cos(phi * numpy.pi/180)

        for psi_i in range (l) :
            psi = float(psi_i) * 360.0/float(l)

            #print "%.0f(%d)(%d)" % (psi, at, at-l),
            x = s * numpy.sin(psi * numpy.pi/180)
            y = s * numpy.cos(psi * numpy.pi/180)

            pt = numpy.array( [ [x+pos.x,y+pos.y,z+pos.z], ], numpy.float32 )
            v = numpy.concatenate ( [v, pt] )

            if phi_i == 0 :
                if psi_i > 0 :
                    vi = vi + ( (at-1, at, 0), )
                if psi_i == l-1 :
                    vi = vi + ( (at, 1, 0), )
            else :
                if psi_i > 0 :
                    tris = Quad2Tri ( [at-1, at, at-l, at-l-1] )
                    vi = vi + tris
                if psi_i == l-1 :
                    tris = Quad2Tri ( [at, at-l+1, at-l*2+1, at-l] )
                    vi = vi + tris

            if phi_i == div-1 :
                if psi_i > 0 :
                    vi = vi + ( (at, at-1, lat+l), )
                if psi_i == l-1 :
                    vi = vi + ( (at-l+1, at, lat+l), )

            at = at + 1                
                

        lat = len ( v )

    pt = numpy.array( [ [0+pos.x,0+pos.y,-r+pos.z], ], numpy.float32 )
    v = numpy.concatenate ( [v, pt] )

    #sph = m.add_group( v, vi, color )
    #sph.set_display_style(sph.Mesh)

    return [v, vi]




def Map2Map ( to, frm, retV=False, dAbove=None ) :

    # print "%s masking %s" % (frm.name, to.name)

    n1, n2, n3 = to.data.size[0], to.data.size[1], to.data.size[2]

    #mm = numpy.zeros( (n3,n2,n1), numpy.single)
    ma = frm.data.full_matrix()

    an1, an2, an3 = frm.data.size[0], frm.data.size[1], frm.data.size[2]
    # print an1,an2,an3, " --> ", n1,n2,n3

    f_i2s = to.data.ijk_to_xyz_transform
    f_t = Matrix.xform_matrix( to.openState.xform )
    tf = Matrix.multiply_matrices( f_t, f_i2s )

    m_points = VolumeData.grid_indices( (n1,n2,n3), numpy.single )  # i,j,k indices
    _contour.affine_transform_vertices( m_points, tf )

    a_vals = frm.interpolated_values ( m_points, chimera.Xform() )
    mm = a_vals.reshape( (n3,n2,n1) )

    if dAbove :
        mm = numpy.where ( mm > dAbove, mm, numpy.zeros_like(mm) )
        # print "- %d above %.3f" % ( numpy.shape(mm.nonzero())[1], dAbove )

    if retV :
        zgd = VolumeData.Array_Grid_Data (
            mm, to.data.origin, to.data.step, to.data.cell_angles,
            to.data.rotation)
        zgd.name = "_>_"
        return zgd
    else :
        return mm



def TransferChainMap ( a, m ) :

    print " - tr %s" % ( a.name ),

    new_v = m.writable_copy()
    new_v.name = a.name + "_tr"
    mm = new_v.data.full_matrix()

    n1, n2, n3 = m.data.size[0], m.data.size[1], m.data.size[2]
    an1, an2, an3 = a.data.size[0], a.data.size[1], a.data.size[2]

    f_i2s = m.data.ijk_to_xyz_transform
    f_t = Matrix.xform_matrix( m.openState.xform )
    tf = Matrix.multiply_matrices( f_t, f_i2s )

    m_points = VolumeData.grid_indices( (n1,n2,n3), numpy.single )  # i,j,k indices
    _contour.affine_transform_vertices( m_points, tf )

    a_vals = a.interpolated_values ( m_points, chimera.Xform() )
    mm[:,:,:] = a_vals.reshape( mm.shape )

    gv = new_v
    gv.region = ( gv.region[0], gv.region[1], [1,1,1] )

    gv.display = True
    #gv.surface_levels = [ dmap.surface_levels[0] ]
    gv.surface_levels = [ a.d_thr ]
    gv.surf_color = ( rand(), rand(), rand(), 1.0 )
    gv.surface_colors = [ gv.surf_color ]
    ro = VolumeViewer.volume.Rendering_Options()
    #ro.surface_smoothing = True
    #ro.smoothing_factor = .25
    #ro.smoothing_iterations = 10
    gv.update_surface ( False, ro )

    for sp in gv.surfacePieces :
        v, t = sp.geometry
        if len(v) == 8 and len(t) == 12 : sp.display = False


    mmN = numpy.where ( mm > a.d_thr, mm, numpy.zeros_like(mm) )
    nze = numpy.nonzero ( mmN )
    print "%d NZ" % len ( nze[0] )

    gv.imap = {}
    for ei, i in enumerate ( nze[0] ) :
        j = nze[1][ei]
        k = nze[2][ei]

        try : mi = gv.imap[i]
        except : mi = {}; gv.imap[i] = mi

        try : mij = mi[j]
        except : mij = {}; mi[j] = mij

        mij[k] = 1

    nzs = numpy.array ( [nze[2], nze[1], nze[0]] )
    points = numpy.cast[numpy.float32] ( numpy.transpose ( nzs ) ).copy()
    _contour.affine_transform_vertices ( points, m.data.ijk_to_xyz_transform )

    com = numpy.sum(points, axis=0) / len(points)
    gv.COM = chimera.Vector ( com[0], com[1], com[2] )
    comv = numpy.ones_like ( points ) * com
    points_v = points - comv
    gv.bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points_v), 1 ) ) )

    return gv



def MakeClearDir ( dname ) :

    files = []
    try :
        files = os.listdir ( dname )
    except :
        try :
            os.mkdir ( dname )
            print " - created", dname
        except :
            print " - error creating dir" ; return

    for f in files :
        try :
            os.remove ( dname + "/" + f )
            print " - removed " + dname + "/" + f
        except :
            pass



def matrix_to_map_correlation ( mt, mp, dAbove=False ) :

    # o, c = map_overlap_and_correlation(f, m, False)

    f_i2s = mp.data.ijk_to_xyz_transform
    f_t = Matrix.xform_matrix( mp.openState.xform )
    tf = Matrix.multiply_matrices(f_t, f_i2s)

    size = list(mt.shape)
    size.reverse()
    f_points = VolumeData.grid_indices(size, numpy.single)        # i,j,k indices
    _contour.affine_transform_vertices(f_points, tf)
    f_weights = numpy.ravel(mt).astype(numpy.single)

    # Eliminate points with zero weight.
    nz = numpy.nonzero(f_weights)[0]
    # print "%d non-zero" % len(nz)
    if len(nz) < len(f_weights):
        f_points = numpy.take(f_points, nz, axis=0)
        f_weights = numpy.take(f_weights, nz, axis=0)

    if dAbove :
        ge = numpy.greater_equal(f_weights, dAbove)
        f_points = numpy.compress(ge, f_points, 0)
        f_weights = numpy.compress(ge, f_weights)

    m_weights = mp.interpolated_values ( f_points, chimera.Xform() )

    o, c = overlap_and_correlation ( f_weights, m_weights )
    return o, c



def densityMov ( mm, dstart, dend, dd ) :

    dat = dstart

    while 1 :

        mm.surface_levels = [dat]
        mm.show()

        #ro = VolumeViewer.volume.Rendering_Options()
        #ro.surface_smoothing = True
        #ro.smoothing_factor = .2
        #ro.smoothing_iterations = 5
        #mm.update_surface ( False, ro )

        print "%.2f" % dat,

        dat = dat + dd

        if dd < 0 and dat < dend : break
        if dd > 0 and dat > dend : break

    print ""






def RandColorChains ( m ) :

    ct = {}
    for r in m.residues: ct[r.id.chainId] = 1
    clist = ct.keys()
    clist.sort()
    chains_clrs = {}
    cnames = ""

    for ci, cid in enumerate ( clist ) :
        clr = ( rand()*.7, rand()*.7, rand()*.7 )
        #print "- %s: clr(%.2f, %.2f, %.2f)" % (cid, clr[0], clr[1], clr[2])
        chains_clrs[cid] = chimera.MaterialColor ( clr[0], clr[1], clr[2], 1.0 )
        cnames = cnames + cid

    print "%s - color ribbon for %d chains -" % ( m.name, len(cnames) ), cnames

    # color atoms
    for r in m.residues :
        clr = chains_clrs[r.id.chainId]
        r.ribbonDrawMode = 2
        r.ribbonColor = clr
        r.ribbonDisplay = True
        for at in r.atoms :
            at.display = False
            at.color = clr

    return chains_clrs



def WriteAtoms ( m, fname ) :
    fp = open ( fname, 'w' )
    for at in m.atoms :
        c = at.coord()
        fp.write ( "%f %f %f %d\n" % ( c[0], c[1], c[2], at.element.number ) )
    fp.close()





def makeSymmetric ( f, symm, niter=3 ) :

    f1 = f.writable_copy()
    f1.name = f1.name [ : f1.name.rfind('.') ] + "_s.mrc"
    f2 = f.writable_copy()
    f2.name = 'f2'

    for i in range ( niter ) :
        rotateMapSym ( f1, f2, symm, "make_sym.txt" )
        f2.data.full_matrix()[:,:,:] = f1.data.full_matrix()[:,:,:]

    chimera.openModels.close ( f2 )

    #print "Writing symmetric map to:", f1.name
    #f1.write_file ( f1.name )



def rotateMapSym ( f, m, symm, fname ) :

    thr = f.surface_levels[0]
    mm = f.data.matrix()
    mm = numpy.where ( mm > thr, mm, numpy.zeros_like(mm) )
    nze = numpy.nonzero ( mm )
    nzs = numpy.array ( [nze[2], nze[1], nze[0]] )
    points = numpy.cast[numpy.float32] ( numpy.transpose ( nzs ) ).copy()

    _contour.affine_transform_vertices ( points, f.data.ijk_to_xyz_transform )

    print " - making axes for %s, %d above %.4f" % ( f.name, len(points), thr )
    f.COM, f.U, f.S, f.V = prAxes ( points )

    print f.COM
    print f.U

    #f = v.v
    ax = chimera.Vector ( f.U[2,0], f.U[2,1], f.U[2,2] )
    #ax = chimera.Vector ( 0, 0, 1 )

    o, c = map_overlap_and_correlation( f, m, False )
    print "%s - %s corr %.4f" % (f.name, m.name, c)

    for i in range ( symm ) :

        print i,

        x = f.openState.xform
        x.multiply ( chimera.Xform.translation ( f.COM ) )
        x.multiply ( chimera.Xform.rotation ( ax, 360.0 / float(symm) ) )
        x.multiply ( chimera.Xform.translation ( -f.COM ) )

        f.openState.xform = x
        #v.contour_model.openState.xform = x

        f_i2s = f.data.ijk_to_xyz_transform
        f_t = Matrix.xform_matrix( f.openState.xform )
        tf = Matrix.multiply_matrices(f_t, f_i2s)
        #print tf
        #mm, tf_inv = f.matrix_and_transform(chimera.Xform(), False, False)
        #tf0 = Matrix.invert_matrix(tf_inv)
        #print tf0

        f_m = f.full_matrix()
        size = list(f_m.shape)
        size.reverse()
        f_points = VolumeData.grid_indices(size, numpy.single)        # i,j,k indices
        _contour.affine_transform_vertices(f_points, tf)
        f_weights = numpy.ravel(f_m).astype(numpy.single)

        if ( 0 ) :
            # Eliminate points with zero weight.
            nz = numpy.nonzero(f_weights)[0]
            # print len(nz), len(f_weights)
            if len(nz) < len(f_weights):
                f_points = numpy.take(f_points, nz, axis=0)
                f_weights = numpy.take(f_weights, nz, axis=0)

        m_weights = m.interpolated_values ( f_points, chimera.Xform() )

        mm = f.data.full_matrix()
        mm[:,:,:] += m_weights.reshape( mm.shape )
        #mm[:,:,:] /= 2.0

        o, c = map_overlap_and_correlation( f, m, False )
        print " - ov: %e, cor: %f" % (o, c)

    mm = f.data.full_matrix()
    mm[:,:,:] /= float ( symm )

    print ""




def BioMatrices ( m ) :

    matrices = {}

    for rm in m.pdbHeaders['REMARK'] :
        s = rm.split()
        if len(s) == 8 and s[2].find('BIOMT') == 0 :
            # print s
            mi = int ( s[3] )
            ri = int ( s[2][5] ) - 1
            try : matrices[mi]
            except : matrices[mi] = numpy.zeros ( (3,4), numpy.float32 )

            matrices[mi][ri][0] = float ( s[4] )
            matrices[mi][ri][1] = float ( s[5] )
            matrices[mi][ri][2] = float ( s[6] )
            matrices[mi][ri][3] = float ( s[7] )

    return matrices                                       



def ReplicateMapSurf ( mm, matrices ) :

    print "Replicating surface of %s with %d matrices" % (
        mm.name, len(matrices.keys()) )

    verts, tris = None, None
    for sp in mm.surfacePieces :
        v, t = sp.geometry
        if len(v) == 8 and len(t) == 12 : sp.display = False
        else : verts, tris = v, t

    print " - %d verts, %d tris" % ( len(verts), len(tris) )


    smod = _surface.SurfaceModel()
    smod.name = mm.name[ : mm.name.rfind('.') ] + " surfaces"
    chimera.openModels.add( [smod] )

    for mi, M in matrices.iteritems() :
        v, t = verts.copy(), tris.copy()
        _contour.affine_transform_vertices ( v, M )
        clr = ( rand(), rand(), rand(), 1.0 )
        nsp = smod.addPiece ( v, t, clr )



def ZeroOutsideRad ( zmap, rr ) :

    print "Taking away from %s" % ( zmap.name )
    rname = zmap.name + '_-_rr'
    mmc = getMod ( rname )

    if mmc == None :
        mmc = zmap.writable_copy ()
        mmc.name = rname
        print " - cloned", zmap.name
    else :
        print " - found", mmc.name


    mm = mmc.data.matrix()
    n1, n2, n3 = mmc.data.size[0], mmc.data.size[1], mmc.data.size[2]
    C = chimera.Vector ( n1/2, n2/2, n3/2 );
    print " size %d,%d,%d, center" % (n1,n2,n3), C

    for i,j,k in VolumeData.grid_indices ( (n3,n2,n1), int ) :
        v = chimera.Vector ( i,j,k ) - C
        if v.length > float(n1/2) - rr :
            mm[k,j,i] = 0.0



def RegionsRads ( dmap, smod ) :

    print "Radii of %d regions in %s" % ( len(smod.surfacePieces), smod.name )

    n1, n2, n3 = dmap.data.size[0], dmap.data.size[1], dmap.data.size[2]

    ctr = numpy.array ( [ [float(n1)/2.0, float(n2)/2.0, float(n3)/2.0] ], numpy.float32 );
    _contour.affine_transform_vertices ( ctr, dmap.data.ijk_to_xyz_transform )
    C = chimera.Vector ( ctr[0][0], ctr[0][1], ctr[0][2] );
    print " size %d,%d,%d, center" % (n1,n2,n3), C

    dr_reg = []
    for r in smod.surfacePieces :
        dr = (r.COM - C).length
        dr_reg.append ( [dr, r] )

    dr_reg.sort()

    path = os.path.dirname ( dmap.data.path ) + os.path.sep
    mname = os.path.splitext ( dmap.name )[0]

    if 0 :
        print "Writing sorted distances from center"
        fp = open ( path + mname + "_radii.txt", "w" )
        for dr, r in dr_reg :
            fp.write ("%f\n" % dr)
        fp.close ()

    elif 0 :
        print "Joining regions by distance from center"
        levels = [1000, 220, 185, 158, 130, 70, 40]
        print levels

        sregs = []
        for dr, r in dr_reg :

            if dr < levels [ len(levels) - 1 ] :
                sregs.append ( r )

            else :
                nr = JoinRegions ( smod, sregs )
                nr.rad_level = levels.pop()
                print levels
                
                sregs = [r]

        nr = JoinRegions ( smod, sregs )
        nr.rad_level = levels.pop()

    else :
        print "Ordering regions by disance from center"

        smod.regs_by_d_to_ctr = []
        
        for r in smod.surfacePieces :

            smallest_d_to_ctr = 1e99
            for x,y,z in r.tpoints :
                d = ( chimera.Vector(x,y,z) - C ).length
                if d < smallest_d_to_ctr :
                    smallest_d_to_ctr = d

            print "- region %d - %f" % ( r.rid, smallest_d_to_ctr )

            smod.regs_by_d_to_ctr.append ( [smallest_d_to_ctr, r] )

        smod.regs_by_d_to_ctr.sort()
    

def MakeBioMGroups ( smod ) :

    smod.bioM_groups = {}

    for r in smod.surfacePieces :

        if r.max_ov < .4 : continue

        try  : smod.bioM_groups [r.max_ov_bioM]
        except : smod.bioM_groups [r.max_ov_bioM] = {}

        bmgs = smod.bioM_groups [r.max_ov_bioM]
        
        try : bmgs[r.max_ov_cid]
        except : bmgs[r.max_ov_cid] = []
        
        try : bmgs['all']
        except : bmgs['all'] = []

        bmgs['all'].append ( r )
        bmgs[r.max_ov_cid].append ( r )

    print "done"


def bioMGroups ( smod ) :

    clrs = {}
    clrs['T'] =  ( 1, 1, .3, 1.0 )

    clrs['S'] = clrs['J'] = clrs['I'] = ( 1, 0, 0, 1.0 )
    clrs['C'] = clrs['D'] = clrs['P'] = ( 0, 1, 0, 1.0 )
    clrs['Q'] = clrs['F'] = clrs['E'] = ( 0, 0, 1, 1.0 )
    clrs['G'] = clrs['H'] = clrs['R'] = ( .7, .5, .2, 1.0 )

    clrs['A'] = ( 1, 0, 0, 1.0 )
    clrs['B'] = ( 0, 1, 0, 1.0 )

    for mi, rgroups in smod.bioM_groups.iteritems() :

        print mi,

        #c = chimera.Vector ( .5*rand(), .5*rand(), .5*rand() )
        #c.normalize()
        #c = c * .7
        #clr = ( c[0], c[1], c[2], 1.0 )

        clr = ( rand(), rand(), rand(), 1.0 )

        for cid, regs in rgroups.iteritems() :
            #print cid,


            if cid == 'all' :

                pass

            elif cid == 'A' or cid == 'B' :

                for r in regs :
                    r.color = clr
                    #r.color = clrs[cid]
                    r.display = True

                #nr = JoinRegions ( smod, regs )
                #rgroups[cid] = [nr]

                pass

            else :

                for r in regs :
                    r.display = True
                    #r.color = clrs[cid]
                    r.color = clr

                #nr = JoinRegions ( smod, regs )
                #rgroups[cid] = [nr]
                

                pass

        #print ""

    print ""



def icosa ( smod ) :

    p_ctrs = []
    for c in smod.p_ctrs : p_ctrs.append ( c )

    ico_verts = []
    print "%d p centers" % len(p_ctrs)

    for i in range ( 12 ) :

        c = p_ctrs[0]
        print "%d - %.2f %.2f %.2f" % (i, c[0], c[1], c[2]),

        # sort all (remaining points ) by distance to c
        d_p = []
        for coi, co in enumerate ( p_ctrs ) :
            d_p.append ( [ (co - c).length, coi ] )

        d_p.sort ()

        v = chimera.Vector(0,0,0)
        popinds = []
        for pi in range ( 5 ) :
            print "%.3f %d," % (d_p[pi][0], d_p[pi][1]),
            v = v + p_ctrs [ d_p[pi][1] ]
            popinds.append ( d_p[pi][1] )

        ico_verts.append ( v / 5.0 )

        print "%d left" % len(p_ctrs)
        popinds.sort()
        popinds.reverse()
        for pi in popinds : p_ctrs.pop ( pi )


    print "%d vertices" % len ( ico_verts )
    ico_ctr = chimera.Vector (0,0,0)
    for vc in ico_verts :
        ico_ctr = ico_ctr + vc

    ico_ctr = ico_ctr / float ( len ( ico_verts ) )
    
    print "Icosahedron center:", ico_ctr


    gmod = getMod ( 'icosahedron' )
    if gmod : chimera.openModels.close ( gmod )

    nmol = chimera.Molecule()
    chimera.openModels.add ( [nmol] )
    nmol.openState.xform = smod.openState.xform
    nmol.name = 'icosahedron'
    aMap = dict()
    nres = nmol.newResidue ( 'ICO', chimera.MolResId('IcoR', 1) )

    for vc in ico_verts :
        vtoc = vc - ico_ctr
        vtoc.normalize()
        apos = vc + vtoc * -10.0 #60.0

        nat = nmol.newAtom ( 'icov', chimera.Element(1))
        nres.addAtom ( nat )
        nat.setCoord ( chimera.Point(apos[0], apos[1], apos[2]) )
        nat.drawMode = nat.Sphere
        nat.radius = 10.0
        nat.color = chimera.MaterialColor ( .7, .7, .7, 1.0 )
        nat.display = True


    for at in nmol.atoms :

        d_ats = []
        for at2 in nmol.atoms :
            d_ats.append ( [ (at2.coord()-at.coord()).length, at2 ] )

        d_ats.sort ()

        for ai in range ( 1, 6 ) :
            try : nb = nmol.newBond ( at, d_ats[ai][1] )
            except : continue
            nb.display = True
            nb.drawMode = nb.Stick
            nb.halfbond = False
            nb.radius = 5.0
            nb.color = chimera.MaterialColor ( .7, .7, .7, 1 )



def rdv_ani () :


    mmap = getMod ( 'rdv_am12.mrc' )
    smod = getMod ( 'rdv_am12_g14.0_D0.53_sg0.0_regions' )

    mmap.display = True
    smod.display = True

    mclr = mmap.surface_colors[0]

    print "Fading out %s, fading in %s" % ( mmap.name, smod.name )

    nsteps = 20
    for i in range (nsteps) :
        print i
        f = float(i) / float(nsteps-1)
        mmap.surface_colors[0] = ( mclr[0], mclr[1], mclr[2], 1.0-f )
        for r in smod.surfacePieces : r.color = ( r.color[0], r.color[1], r.color[2], f )


    print ""



def rdv_ani () :


    smic = getMod ( 'rdv_am12_dna_g4.0_g10.0_D0.01_2009-05-07_22-32-16_sg0.0_regions' )
    smod = getMod ( 'rdv_am12_g14.0_D0.53_2009-05-08_21-24-38_regions' )

    for r in smod.surfacePieces :
        r.display = False

    for r in smic.surfacePieces :
        r.display = False

    for d, rat in smic.regs_by_d_to_ctr :

        print "%.2f" % d,
        rat.display = True

        # rat.color = ( rand(), rand(), rand(), 0.0 )

        clrb = numpy.array ( [0,0,0,0], numpy.float32 )
        clre = numpy.array ( [rat.color[0], rat.color[1], rat.color[2], 1.0], numpy.float32 )
        
        #rat.transparencyBlendMode == rat.SRC_ALPHA_DST_1_MINUS_ALPHA
        rat.transparencyBlendMode == rat.SRC_1_DST_1_MINUS_ALPHA
        
        nsteps = 3
        for i in range (nsteps) :
            print i,
            f = float(i) / float(nsteps-1)

            clr = clrb * (1.0-f) + clre * f
            rat.color = ( clr[0], clr[1], clr[2], f )


        print ""


    i_asus = {}
    p_asus = {}
    for mi, rgroups in smod.bioM_groups.iteritems() :
        i_asus[mi] = []
        p_asus[mi] = []
        for cid, regs in rgroups.iteritems() :
            if cid == 'all' : pass
            elif cid == 'A' or cid == 'B' :
                for r in regs : i_asus[mi].append (r)
            else :
                for r in regs : p_asus[mi].append (r)


    print "- inner shell"
    mi_at = 5
    for i in range ( 60 ) :
        print mi_at,
        regs = i_asus.pop ( mi_at )
        for r in regs : r.display = True

        closest_mi = None
        closest_d = 1e99
        for mi, mregs in i_asus.iteritems() :
            for r in regs :
                for mr in mregs :
                    d = (r.COM - mr.COM).length
                    if d < closest_d :
                        closest_d = d
                        closest_mi = mi

        if closest_mi == None : break
        mi_at = closest_mi



    print ""


    print "- outer shell"
    mi_at = 5
    for i in range ( 60 ) :
        print mi_at,
        regs = p_asus.pop ( mi_at )
        for r in regs : r.display = True

        closest_mi = None
        closest_d = 1e99
        for mi, mregs in p_asus.iteritems() :
            for r in regs :
                for mr in mregs :
                    d = (r.COM - mr.COM).length
                    if d < closest_d :
                        closest_d = d
                        closest_mi = mi

        if closest_mi == None : break
        mi_at = closest_mi


    print ""


def laplacian ( m ) :

    lm = m.astype(numpy.float32)        # Copy array

    numpy.multiply(lm, -6.0, lm)

    numpy.add(lm[:-1,:,:], m[1:,:,:], lm[:-1,:,:])
    numpy.add(lm[1:,:,:], m[:-1,:,:], lm[1:,:,:])

    numpy.add(lm[:,:-1,:], m[:,1:,:], lm[:,:-1,:])
    numpy.add(lm[:,1:,:], m[:,:-1,:], lm[:,1:,:])

    numpy.add(lm[:,:,:-1], m[:,:,1:], lm[:,:,:-1])
    numpy.add(lm[:,:,1:], m[:,:,:-1], lm[:,:,1:])

    lm[0,:,:] = 0
    lm[-1,:,:] = 0
    lm[:,0,:] = 0
    lm[:,-1,:] = 0
    lm[:,:,0] = 0
    lm[:,:,-1] = 0

    return lm


def lsmooth ( v, numit ) :

    m = v.data.matrix()

    for i in range ( numit ) :
        print i,
        lm = laplacian ( m )
        numpy.multiply(lm, 0.1, lm)
        m = m + lm


    print "done"

    origin, step = v.data.origin, v.data.step
    d = v.data
    ld = VolumeData.Array_Grid_Data(m, origin, step, d.cell_angles, d.rotation,
                       name = v.name + ' Laplacian')
    ld.polar_values = True

    lv = VolumeViewer.volume_from_grid_data( ld, show_data = False )
    lv.copy_settings_from(v, copy_thresholds_and_colors = False)
    lv.set_parameters(cap_faces = False)
    lv.initialize_thresholds()
    lv.show()
    v.unshow()          # Hide original map



def tseg ( v, out_file ) :

    m = v.data.matrix()

    thr = v.surface_levels[0]
    inds = numpy.where ( m > thr )
    iss = numpy.array ( [inds[2], inds[1], inds[0]] )
    ist = numpy.transpose (iss)

    umsg ( "Making list with %d voxels, threshold %.3f" % (len(ist), thr) )
    d_ind = []
    for p in ist :
        dval = m [ p[2], p[1], p[0] ]
        if dval < thr :
            print "error - indices rev"
            return
        d_ind.append ( [ dval, p[0], p[1], p[2] ] )

    # for d, i, j, k in d_ind[0:10] : print "%.3f - %d %d %d" % (d, k, j, i)


    umsg ( "Sorting list with %d voxels, threshold %.3f" % (len(ist), thr) )

    d_ind.sort()
    d_ind.reverse()

    # for d, i, j, k in d_ind[0:10] : print "%.3f - %d %d %d" % (d, k, j, i)

    print "Segmenting (1-10) :",

    rid_m = numpy.zeros ( m.shape, numpy.int )
    regions = []
    regions.append ( [] )

    rid_at = 0
    num_ind_m = len(d_ind) / 10
    ind_at = 0
    last_p = 0.0

    cons = {}
    
    for d, i, j, k in d_ind :

        if ind_at % num_ind_m == 0 :
            print "%d" % (1+ind_at/num_ind_m),

        p_at = 1.0 + ( 100.0*float(1+ind_at)/float(len(d_ind)) )
        if ( p_at >= 1.0 + last_p ) :
            status ( "Segmenting: %.0f%%" % p_at )
            last_p = p_at

        ind_at = ind_at + 1

        adj_regs = {}
        for ii in [i-1,i,i+1] :
            for jj in [j-1,j,j+1] :
                for kk in [k-1,k,k+1] :
                    rid = rid_m [ kk, jj, ii ]
                    if rid == 0 : continue
                    try : adj_regs[rid] = adj_regs[rid] + 1
                    except : adj_regs[rid] = 1

        regs = adj_regs.keys()
        num_regs = len ( regs )

        if num_regs == 0 :
            rid_at = rid_at + 1
            rid_m [ k, j, i ] = rid_at

            regions.append ( [] )
            regions [ rid_at ].append ( [i,j,k] )

        elif num_regs == 1 :
            rid = regs[0]
            rid_m [ k, j, i ] = rid

            regions [ rid ].append ( [i,j,k] )

        else :
            max_count = 0
            max_count_reg = None
            for reg, count in adj_regs.iteritems() :
                if count > max_count :
                    max_count = count
                    max_count_reg = reg

            rid_m [ k, j, i ] = max_count_reg
            regions [ max_count_reg ].append ( [i,j,k] )

        if num_regs == 2 :
            rid1 = regs[0]
            rid2 = regs[1]

            try : c1 = cons[rid1]
            except : c1 = {}; cons[rid1] = c1
            try : o = c1[rid2]
            except : o = aClass(); o.N, o.D = 0, 0.0; c1[rid2] = o

            o.N, o.D = o.N + 1, o.D + d

            try : c2 = cons[rid2]
            except : c2 = {}; cons[rid2] = c2
            c2[rid1] = o


    consa = []
    for r1, cr1 in cons.iteritems () :
        for r2, o in cr1.iteritems () :
            if r1 < r2 :
                consa.append ( [r1, r2, o] )

    print ""
    print " - found %d regions, %d connections" % ( len(regions)-1, len(consa) )

    e = None

    if 0 :
        print " - serializing regions"

        e = numpy.array ( [ len(regions)-1 ], numpy.float32 )

        for rinds in regions[1:] :

            np = numpy.array ( [ len(rinds) ], numpy.float32 )
            rpoints = numpy.array ( rinds, numpy.float32 )
            pts = numpy.resize ( rpoints, ( 3*len(rpoints), ) )
            e = numpy.concatenate ( [e, np] )
            e = numpy.concatenate ( [e, pts] )


        print " - serializing connections"

        nc = numpy.array ( [ len(consa) ], numpy.float32 )
        e = numpy.concatenate ( [e, nc] )

        for r1, r2, o in consa :
            nc = numpy.array ( [ r1, r2, o.N, o.D ], numpy.float32 )
            e = numpy.concatenate ( [e, nc] )

    else :

        size = 1
        for rinds in regions[1:] : size = size + 1 + 3 * len(rinds)
        size = size + 1 + 4 * len(consa)

        # print " - size:", size

        e = numpy.zeros ( [size], numpy.float32 )
        e[0] = float ( len(regions)-1 )
        at = 1

        print " - writing %d regions..." % ( len(regions)-1 )

        for rinds in regions[1:] :
            e[at] = len(rinds)
            at = at + 1
            for rp in rinds :
                e[at], e[at+1], e[at+2] = rp[0], rp[1], rp[2]
                at = at + 3

        e[at] = float ( len(consa) )
        at = at + 1

        print " - writing %d connections" % ( len(consa) )

        for rid1, rid2, o in consa :
            e[at+0] = rid1
            e[at+1] = rid2
            e[at+2] = o.N
            e[at+3] = o.D
            at = at + 4

    # e = numpy.cast[numpy.float32] ( e )
    # print e
    print " - writing to", out_file
    e.tofile ( out_file )


def dToMol ( pos, mol ) :

    min_d_to_mol = 1e99

    for r in mol.residues :
        cpos = r.atomsMap["C"][0].xformCoord()
        v = pos.toVector() - cpos.toVector()
        if v.length < min_d_to_mol :
            min_d_to_mol = v.length

    return min_d_to_mol


def sphDensities ( mm, pos, rad ) :

    pp = Matrix.apply_matrix ( mm.data.xyz_to_ijk_transform, [ [pos[0], pos[1], pos[2]] ] )[0]
    i,j,k = int (pp[0]), int(pp[1]), int(pp[2])
    dd = int ( numpy.ceil ( rad / mm.data.step[0] ) )
    # print pos, " -> ", i, j, k, "window", dd

    densities = []
    M = mm.data.full_matrix()
    
    for ii in range ( i-dd, i+dd+1 ) :
        for jj in range ( j-dd, j+dd+1 ) :
            for kk in range ( k-dd, k+dd+1 ) :
                vd = chimera.Point (i,j,k) - chimera.Point (ii,jj,kk)
                if int(numpy.round(vd.length)) <= dd :
                    densities.append ( M[kk,jj,ii] )

    return densities


def sphAvgD ( mm, pos, rad ) :
    ds = sphDensities (mm,pos,rad)
    return sum(ds) / float ( len(ds) )



def clearSphD ( mm, pos, rad ) :

    pp = Matrix.apply_matrix ( mm.data.xyz_to_ijk_transform, [ [pos[0], pos[1], pos[2]] ] )[0]
    i,j,k = int (pp[0]), int(pp[1]), int(pp[2])
    dd = int ( numpy.ceil ( rad / mm.data.step[0] ) )
    # print pos, " -> ", i, j, k, "window", dd

    M = mm.data.full_matrix()
    
    for ii in range ( i-dd, i+dd+1 ) :
        for jj in range ( j-dd, j+dd+1 ) :
            for kk in range ( k-dd, k+dd+1 ) :
                vd = chimera.Point (i,j,k) - chimera.Point (ii,jj,kk)
                if int(numpy.round(vd.length)) <= dd :
                    M[kk,jj,ii] = 0.0

    mm.matrix_changed ()

    mm.update_surface ( False, VolumeViewer.volume.Rendering_Options() )
    for sp in mm.surfacePieces :
        v, t = sp.geometry
        if len(v) == 8 and len(t) == 12 : sp.display = False



def randPlaceSphereOnD ( mm, mol, rad ) :

    m = mm.data.matrix()
    thr = mm.surface_levels[0]
    inds = numpy.where ( m > thr )
    iss = numpy.array ( [inds[2], inds[1], inds[0]] )
    ist = numpy.transpose (iss)

    for i in range ( 1000 ) :

        vi = int ( rand() * float((len(ist)-1)) )
        p = ist[vi]
        pp = Matrix.apply_matrix ( mm.data.ijk_to_xyz_transform, [ [p[0], p[1], p[2]] ] )[0]

        cp = chimera.Point ( pp[0], pp[1], pp[2] )
        tcp = mm.openState.xform.apply ( cp )
        d = dToMol ( tcp, mol )

        if d > rad :
            return cp



def placeSphere ( mm, mol, rad, steps, init=False ) :

    sm = getMod ( "es_mod" )
    if sm == None :
        sm = _surface.SurfaceModel()
        chimera.openModels.add( [sm] )
        [v,vi] = SphereMesh ( rad, 30, chimera.Vector(0,0,0) )
        sm.name = "es_mod"
        nsp = sm.addPiece ( v, vi, (1,.6,.8,.5) )

    if init :
        print "Setting initial sphere pos"
        spos = randPlaceSphereOnD ( mm, mol, rad )
        sm.spos = spos

        oxf = mm.openState.xform
        oxf.multiply ( chimera.Xform.translation ( spos.toVector() ) )
        sm.openState.xform = oxf


    for i in range ( steps ) :

        bMoved = False

        for i in range ( 100 ) :

            sd = chimera.Vector ( rand()-.5, rand()-.5, rand()-.5 )
            sd.normalize ()
            #try : sd = sd + 0.3 * sm.posv
            #except : pass
            spos = sm.spos + 10.0 * sd

            ppos = chimera.Point ( spos[0], spos[1], spos[2] )
            d = dToMol ( mm.openState.xform.apply(ppos), mol )

            #if d > rad+2 and d < rad*2.0 :
            if d > rad+2 :

                ds = sphDensities ( mm, spos, rad-1 )
                #avgd = sum(ds) / float(len(ds))

                if max(ds) >= 1.3 :

                    clearSphD ( mm, spos, rad-1 )

                    #print " moved to pos :", spos, " d :", d
                    sm.spos = spos
                    sm.posv = sd

                    oxf = mm.openState.xform
                    oxf.multiply ( chimera.Xform.translation ( spos.toVector() ) )
                    sm.openState.xform = oxf
                    print "-",

                    if mdlg : mdlg.Step ( 1 )

                    bMoved = True
                    break

        if bMoved == False :
            print "*"

            spos = randPlaceSphereOnD ( mm, mol, rad )
            sm.spos = spos

            oxf = mm.openState.xform
            oxf.multiply ( chimera.Xform.translation ( spos.toVector() ) )
            sm.openState.xform = oxf

            # break

    return sm



def ExhaustiveSearch ( fmap, dmap ) :

    mmat = dmap.data.matrix()
    thr = dmap.surface_levels[0]
    inds = numpy.where ( mmat > thr )
    iss = numpy.array ( [inds[2], inds[1], inds[0]] )
    ist = numpy.transpose (iss)

    ppos = numpy.cast[numpy.float32] ( ist )

    _contour.affine_transform_vertices ( ppos, dmap.data.ijk_to_xyz_transform )
    

    highest_fit, highest_M, highest_MI, opt_af = 0.0, None, None, None

    last_pos = None
    num_it = 0

    # for pi in range ( 0, len(ppos) ) :

    for ii in range ( 0, dmap.data.size[0], 10 ) :
        for jj in range ( 0, dmap.data.size[1], 10 ) :
            for kk in range ( 0, dmap.data.size[2], 10 ) :

                if mmat[kk][jj][ii] < dmap.surface_levels[0] :
                    continue

                parray = numpy.array ( [ [ii,jj,kk] ], numpy.float32 )
                _contour.affine_transform_vertices ( parray, dmap.data.ijk_to_xyz_transform )

                p = parray [0]

                print "[%.2f,%.2f,%.2f]" % (p[0], p[1], p[2]),

                comT = numpy.matrix ( [
                    [ 1, 0, 0, p[0] ],
                    [ 0, 1, 0, p[1] ],
                    [ 0, 0, 1, p[2] ],
                    [ 0, 0, 0,    1 ]  ] )

                num_it = num_it + 1
                if num_it > 100 : return


                tXO, tXR = xf_2_M ( dmap.openState.xform )

                N, M = 2, 10
                thetas, phis = [], []
                # http://www.math.niu.edu/~rusin/known-math/97/spherefaq
                for k in range ( 1, N+1 ) :
                    if 1 :
                        thetas.append ( rand() * 2.0*numpy.pi  )
                    else :
                        h = -1.0 + ( 2.0*float(k-1)/float(N-1) )
                        thetas.append ( numpy.arccos(h) )
                    phi = 0.0
                    if k == 1 or k == N : phi = 0
                    else : phi = (phis[k-2] + 3.6/numpy.sqrt(float(N)*(1.0-h**2.0)))
                    while phi > 2.0*numpy.pi : phi -= 2.0*numpy.pi
                    phis.append ( phi )

                for ti, thet in enumerate ( thetas ) :
                    phi, theta = phis[ti], thet - (numpy.pi / 2.0)
                    v = chimera.Vector (0,0,0)
                    v[2] = numpy.sin(theta)
                    v[1] = numpy.cos(theta) * numpy.sin (phi)
                    v[0] = numpy.cos(theta) * numpy.cos (phi)

                    # print ":",

                    for m in range ( M ) :
                        ang = 360.0 * float(m)/float(M)

                        xfR = chimera.Xform.rotation ( v, ang )
                        rrO, rrR = xf_2_M ( xfR )

                        fmap.M = comT * rrR
                        T = tXO * tXR * fmap.M
                        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
                        fmap.openState.xform = xfA
                        fmap.mol.openState.xform = xfA

                        #ov, corr = self.FitMap_T ( fmap, dmap, num_tries=1 )

                        o, corr = map_overlap_and_correlation( fmap, dmap, True )
                        # print "%.1f" % corr,

                        if corr > highest_fit :
                            highest_fit = corr
                            highest_M = fmap.M
                            highest_MI = comT * rrR

                        status ( "%d/%d : %.3f" % (ti*M+m+1, N*M, corr) )

                        mdlg.Step ( 1 )

                # return

    mdlg.Step ( 10 )



def writeSurf ( mod ) :

    fp = open ( "groel_surf.txt", "w" )

    for sp in mod.surfacePieces :

        verts, tris = sp.geometry

        fp.write ( "TRI\n\n%d\n\n%d\n\n" % ( len(verts), len(tris) ) )

        if len(verts) == 8 and len(tris) == 12 :
            sp.display = False
        else :
            print "%d verts, %d tris" % (len(verts), len(tris))
            for v in verts :
                fp.write ( "%f %f %f\n" % ( v[0], v[1], v[2] ) )
            fp.write ( "\n" );
            for t in tris :
                fp.write ( "%d %d %d\n" % ( t[0], t[1], t[2] ) )
                
    fp.close()

    

def modRF (mol, mol2) :

    #mol = chimera.openModels.list()[0]
    #mol2 = chimera.openModels.list()[1]

    for ai, at in enumerate ( mol.atoms ) :
        for at2 in mol2.atoms :
            if at.name == at2.name :
                at2.setCoord ( at.coord() )

    mol2.openState.xform = mol.openState.xform

    atoms = chimera.selection.OSLSelection( "#0" ).atoms()

    points = _multiscale.get_atom_coordinates ( atoms, transformed = False )
    COM, U, S, V = prAxes ( points )

    COM = COM + chimera.Vector ( -5, -5, -5 )

    print COM
    print U

    try : mol.axes
    except :
        mol.axes = AxesModCOM ( COM, alignTo = mol )
        mol.axes.name = "axes_1"

    try : mol2.axes
    except :
        mol2.axes = AxesModCOM ( COM, alignTo = mol2 )
        mol2.axes.name = "axes_2"

    mol2.axes.openState.xform = mol.openState.xform


    Z = chimera.Vector( random.random(), random.random(), random.random() )
    Z.normalize()
    dZ = chimera.Vector( random.random(), random.random(), random.random() )
    dZ.normalize()
    X = chimera.cross ( Z, dZ )
    X.normalize ()
    Y = chimera.cross ( Z, X )
    Y.normalize ()

    O = numpy.matrix ( [
        [ 1, 0, 0, -COM[0] ],
        [ 0, 1, 0, -COM[1] ],
        [ 0, 0, 1, -COM[2] ],
        [    0,     0,   0, 1 ] ] )

    R = numpy.matrix ( [
        [ X[0], Y[0], Z[0], 0 ],
        [ X[1], Y[1], Z[1], 0 ],
        [ X[2], Y[2], Z[2], 0 ],
        [    0,     0,   0, 1 ] ] )

    C = numpy.matrix ( [
        [ 1, 0, 0, COM[0]+16 ],
        [ 0, 1, 0, COM[1]+16 ],
        [ 0, 0, 1, COM[2]+5 ],
        [    0,     0,   0, 1 ] ] )

    T = C * R * O

    xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3], True )

    xf = mol.axes.openState.xform
    xf.multiply ( xfA )
    mol2.axes.openState.xform = xf
    mol2.openState.xform = xf



def map_overlap_and_correlation (map1, map2, above_threshold):

    olap, cor = FitMap.map_overlap_and_correlation ( v1, v2, above_threshold )[:2]
    return olap, cor



def overlap_and_correlation ( v1, v2 ):

    olap, cor = FitMap.overlap_and_correlation ( v1, v2 )[:2]
    return olap, cor



def unusedFile ( path_format ):

    i = 1
    exists = True
    while exists:
        path = path_format % (i,)
        exists = os.path.exists ( path )
        i += 1
    return path


def makeRegionSurface ( smod, vertices, triangles, region ):

    nsp = smod.addPiece ( vertices, triangles, region.color )
    nsp.oslName = str(region.rid)
    nsp.region = region
    smod.rid_sp_map [ region.rid ] = nsp
    return nsp


def reportFitRegions(map_name, regs):

    r = ', '.join(str(reg.rid) for reg in regs[:5])
    if len(regs) > 5:
        r += '...'
    umsg ( "Fitting %s to %d regions (%s)" % ( map_name, len(regs), r ) )


#
# Number regions having no children in depth first order since that is the
# order they will be written to the file.  Renumber nodes with children
# using higher numbers to keep all region numbers distinct.
#
def renumberRegions(smod):

    newrid = {}
    parents = []
    rlist = smod.regions.values()
    rlist.sort(lambda r1, r2: cmp(r1.rid, r2.rid))
    for r in rlist:
        renumberRegion(r, newrid, parents)
    for r in parents:
        next_id = len(newrid) + 1
        newrid[r] = next_id

    for r, rid in newrid.items():
        r.rid = rid

    smod.regions.clear()
    for r in rlist:
        smod.regions[r.rid] = r

    splist = smod.rid_sp_map.values()
    smod.rid_sp_map.clear()
    for sp in splist:
        smod.rid_sp_map[sp.region.rid] = sp

    return rlist

def renumberRegion(r, rid, parents):

    if len(r.cregs) == 0:
        next_id = len(rid) + 1
        rid[r] = next_id
    else:
        parents.append(r)
        for c in r.cregs:
            renumberRegion(c, rid, parents)


# "C:\Program Files\Chimera\bin\chimera.exe" --geometry 800x800+0+0 --bgopacity
#chimera.viewer.windowSize = (500, 500)
#chimera.viewer.background = chimera.MaterialColor ( 1,1,1,0 )
# chimera.tkgui.app.winfo_toplevel().geometry('600x600+0+20')
# chimera.printer.saveImage ( "", format = 'jpeg', supersample=3 )
# _surface.closest_geometry_intercept
# SurfaceDistance.surfdist
# SurfaceDistance.surfdist.distance_to_surface ( ((0,0,0),(100,0,0)), g )
# d = %.5g, surface point (%.5g, %.5g, %.5g), side %.0f


